/* 
 *   Fetcher, Copyright (C) 1996, 1997 James Bailie
 * ==================================================================
 *
 * Fetcher is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * Fetcher is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 *
 * Although Fetcher is licensed under the Free Software Foundation's GNU
 * General Public License, Fetcher is not produced by, nor is it endorsed
 * by the Free Software Foundation. The Free Software Foundation is
 * not responsible for developing, distributing, or supporting Fetcher
 * in any way. Anyone may place software they own the copyright to,
 * under the GNU General Public License.
 *
 * The GNU General Public License is included in the Mail Peruser
 * distribution archive in a file called COPYING. If you do not have a
 * copy of the license, you can download one from
 * ftp://prep.ai.mit.edu, or you can write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * =====================================================================
 */


#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<signal.h>
#include<sys/socket.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/wait.h>
#include<fcntl.h>
#include<ctype.h>
#include<netinet/in.h>
#include<netdb.h>
#include<unistd.h>

#define BUFFER_SIZE 1024


int pop_socket;  

FILE *pop_server;
FILE *spool_file;
FILE *read_file;
FILE *output, *client;

pid_t pid;

const char *name = "fetcher";
const char *home_directory;

void graceful_exit( int status )
{
    fclose( pop_server );
    close( pop_socket );
    
    fflush( output );
    fclose( output );
  
    exit( status );
}

int main( int argc, char **argv[] )
{
  char line_buffer[ BUFFER_SIZE ];

  struct protoent *protocol;
  struct hostent *host;
  struct servent *service;
  struct sockaddr_in socket_address;

  int *messages;
  int index;
  int second_index;

  home_directory = getenv( "HOME" );
  
  if ( strstr( ( char *)argv[ 0 ], "cfetcher" ))
  {
      sprintf( line_buffer, "%s/.peruser_spool/client_pipe",
               home_directory );
      if (( output = fopen( line_buffer, "w" )) == NULL )
      {
          perror( name );
          graceful_exit( EXIT_FAILURE );
      }
  }
  else
      output = stdout;
  
  
    if ( argc < 4 )
    {
      fprintf( output, "fetcher: You must supply the POP3 server's name,\n" ); 
      fprintf( output, "fetcher: followed by your POP3 user name\n" );
      fprintf( output, "fetcher: and password as the first, second,\n" );
      fprintf( output, 
	       "fetcher: and third command line arguments, respectively.\n" );
      graceful_exit( EXIT_FAILURE );
    }

  /*
   * Lookup the protocol number assigned to TCP. 
   */

  if ( !( protocol = getprotobyname( "tcp" )))
    {
      fprintf( output, "fetcher: Couldn't find protocol number for \"tcp\"\n" );
      fprintf( output, "fetcher: in /etc/protocols.\n" );
      graceful_exit( EXIT_FAILURE );
    }

  /*
   * Lookup the service number assigned to POP3.
   */

  if ( !( service = getservbyname( "pop3", "tcp" )))
    {
        if ( !( service = getservbyname( "pop-3", "tcp" )))
      {
          fprintf( output,
                   "fetcher: Couldn't find service number for \"pop3\"\n" );
          fprintf( output, "fetcher: or \"pop-3\"in /etc/services.\n" );
          graceful_exit( EXIT_FAILURE );
      }
    }

  /*
   * Lookup the address of the POP3 server named on the command line.
   */

  if ( !( host = gethostbyname( ( char *) argv[ 1 ] )))
     {
       herror( name );
       graceful_exit( EXIT_FAILURE );
     }

  /*
   * Create a socket for communication with the POP3 server.
   */

  if (( pop_socket = socket( AF_INET, SOCK_STREAM, protocol->p_proto ))
      == -1 )
    {
        perror( name );
        graceful_exit( EXIT_FAILURE );
    }

  /*
   * Zero and initialize the socket_address structure.
   */

  memset( &socket_address, 0, sizeof socket_address );
  socket_address.sin_family = AF_INET;
  memmove( (char *)&socket_address.sin_addr, host->h_addr, host->h_length );
  socket_address.sin_port = service->s_port;

/*
 * Connect to the POP3 server through the socket.
 */

  if ( connect( pop_socket, ( struct sockaddr *)&socket_address, 
	        sizeof socket_address ) == -1 )
    {
      perror( name );
      graceful_exit( EXIT_FAILURE );
    }

  /*
   * Create a stream and associate it with the socket.
   */

  if ( !( pop_server = fdopen( pop_socket, "r+" )))
    {
      perror( name );
      graceful_exit( EXIT_FAILURE );
    }

  /*
   * Get the welcome message from the pop server.
   */
  fgets( line_buffer, sizeof line_buffer, pop_server );
  fprintf( output, "%s\n", strtok( line_buffer, "\r" ));
  fflush( output );
  
  if ( line_buffer[ 0 ] != '+' )
    {
      fprintf( output, 
	       "fetcher: POP3 server responded with a non-positive greeting.\n" );
      graceful_exit( EXIT_FAILURE );
    }
  
  /*
   * Give the user name to the pop server.
   */


  sprintf( line_buffer, "USER %s\r\n", ( char *)argv[ 2 ] );
  fputs( line_buffer, pop_server );
  fprintf( output, "fetcher: %s\n", strtok( line_buffer, "\r" ));
  fflush( output );
  
  fgets( line_buffer, BUFFER_SIZE, pop_server );
  fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
  fflush( output );
  
  if ( line_buffer[ 0 ] != '+' )
    {
      fprintf( output, "fetcher: POP3 server does not recognize user name.\n" );
      graceful_exit( EXIT_FAILURE );
    }

  /*
   * ...password
   */

  fprintf( pop_server, "PASS %s\r\n", ( char *)argv[ 3 ] );
  fputs( "fetcher: PASS <password>\n", output );

  fgets( line_buffer, BUFFER_SIZE, pop_server );
  fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
  fflush( output );
  
  if ( line_buffer[ 0 ] != '+' )
    {
      fprintf( output, "fetcher: POP3 server does not recognize password.\n" );
      graceful_exit( EXIT_FAILURE );
    }

  /*
   * Ask for stats..
   */

  sprintf( line_buffer, "STAT\r\n" );
  fputs( line_buffer, pop_server );
  fprintf( output, "fetcher: %s\n", strtok( line_buffer, "\r" ));
  fflush( output );
  
  fgets( line_buffer, BUFFER_SIZE, pop_server );
  fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
  fflush( output );
  
  if ( line_buffer[ 0 ] != '+' )
    {
      fprintf( output,
               "fetcher: POP3 server responded with a non-positive acknowledgment to STAT.\n" );
      graceful_exit( EXIT_FAILURE );
    }

  strtok( line_buffer, " " );
  if ( !( index = atoi( strtok( NULL, " " ))))
    {
      fprintf( output, "fetcher: (No messages waiting in mail drop.)\n" );
      fflush( output );
      
      sprintf( line_buffer, "QUIT\r\n" );
      fputs( line_buffer, pop_server );
      fprintf( output, "fetcher: %s\n", strtok( line_buffer, "\r" ));
      fflush( output );
      
      fgets( line_buffer, BUFFER_SIZE, pop_server );
      fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
      fflush( output );
      
      graceful_exit( EXIT_SUCCESS );
    }

  /*
   * If there are waiting messages, ask for list...
   */

  strcpy( line_buffer, "LIST\r\n" );
  fputs( line_buffer, pop_server );
  fprintf( output, "fetcher: %s\n", strtok( line_buffer, "\r" ));
  fflush( output );
  
  fgets( line_buffer, BUFFER_SIZE, pop_server );
  fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
  fflush( output );
  
  if( line_buffer[ 0 ] != '+' )
    {
      fprintf( output,
               "fetcher: POP3 server responded with a non-positive acknowledgment to LIST.\n" );
      graceful_exit( EXIT_FAILURE );
    }

  if( !( messages = calloc( index, sizeof( int ))))
    {
      perror( name );
      graceful_exit( EXIT_FAILURE );
    }

  for( second_index = 0; second_index < index; second_index++ )
    {
      fgets( line_buffer, BUFFER_SIZE, pop_server );
      fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
      fflush( output );
      
      messages[ second_index ] = atoi( strtok( line_buffer, " " )); 
    }

  /*
   * Get the terminating "."
   */

  fgets( line_buffer, BUFFER_SIZE, pop_server );
  fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
  fflush( output );

  if ( line_buffer[ 0 ] != '.' )
    {
      fprintf( output,
               "fetcher: server did not send terminating .CRLF, as expected" );
      graceful_exit( EXIT_FAILURE );
    }

  /*
   *   Create the .peruser_spool directory if it doesn't exist.
   */

  sprintf( line_buffer, "%s/.peruser_spool", home_directory );
  if ( mkdir( line_buffer, 0755 ))
    {
      if ( errno != EEXIST )
	{
	  perror( name );
	  graceful_exit( SIGINT );
	}
    }

  /*
   * Open spool.
   */

  sprintf( line_buffer, "%s/.peruser_spool/INBOX", home_directory );
  if (( spool_file = fopen( line_buffer, "a" )) == NULL )
    {
      perror( name );
      graceful_exit( EXIT_FAILURE );
    }
  
  /*
   * Open read file.
   */

  strcat( line_buffer, ".read" );
  if (( read_file = fopen( line_buffer, "a" )) == NULL )
    {
      perror( name );
      graceful_exit( EXIT_FAILURE );
    }

  /*
   * Cycle through the list, retrieving messages and deleting them from the
   * server.
   */

  for( second_index = 0; second_index < index; second_index++ )
    {
      sprintf( line_buffer, "RETR %d\r\n", messages[ second_index ] );
      fputs( line_buffer, pop_server );
      fprintf( output, "fetcher: %s\n", strtok( line_buffer, "\r" ));
      fflush( output );
      
      fgets( line_buffer, BUFFER_SIZE, pop_server );
      fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
      fflush( output );
      
      if ( line_buffer[ 0 ] != '+' )
	{
	  fprintf( output, "fetcher: POP3 server responded with a non-positive acknowledgment to RETR.\n" );
	  fflush( spool_file );
	  fclose( spool_file );
	  graceful_exit( EXIT_FAILURE );
	}

      fprintf( output, "       (retrieving message %d...)\n", 
               messages[ second_index ] );
      fflush( output );
      
      do
	{
	  fgets( line_buffer, BUFFER_SIZE, pop_server );
	  fputs( line_buffer, spool_file );
	}
      while( strncmp( line_buffer, ".\r\n", 3 ));

      fputs( "u\n", read_file );

      sprintf( line_buffer, "DELE %d\r\n", messages[ second_index ] );
      fputs( line_buffer, pop_server );
      fprintf( output, "fetcher: %s\n", strtok( line_buffer, "\r" ));
      fflush( output );
            
      fgets( line_buffer, BUFFER_SIZE, pop_server );
      fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
      fflush( output );
      
      if ( line_buffer[ 0 ] != '+' )
	{
	  fprintf( output,
                   "fetcher: POP3 server responded with a non-positive acknowledgment to DELE.\n" );
	  continue;
	}
      fprintf( output, "\n" );
      fflush( output );
    }

  /*
   * Goodbye now, goodbye! Write down all I said...
   */

  fflush( spool_file );
  fclose( spool_file );

  fflush( read_file );
  fclose( read_file );

  sprintf( line_buffer, "QUIT\r\n" );
  fputs( line_buffer, pop_server );
  fprintf( output, "fetcher: %s", line_buffer );
  fflush( output );

  fgets( line_buffer, BUFFER_SIZE, pop_server );
  fprintf( output, "server:  %s\n", strtok( line_buffer, "\r" ));
  fputs( "\n", output );
  fflush( output );
    
  fclose( pop_server );
  close( pop_socket );

  if ( output == stdout )
      execlp( "filterer", "filterer", NULL );
  else
  {
      sprintf( line_buffer, "%s/.peruser_spool/sorter_pipe",
               home_directory );
      if ( mkfifo( line_buffer, 0700 ))
      {
          if ( errno != EEXIST )
          {
              fclose( output );
              perror( name );
              exit( EXIT_FAILURE );
          }
      }

      pid = 0;
      switch( pid = fork() )
      {
          case -1:
              fclose( output );
              perror( name );
              exit( EXIT_FAILURE );
              break;
          case 0:
              execlp( "cfilterer", "cfilterer", NULL );

              perror( name );
              exit( EXIT_FAILURE );
              break;
          default:
              if (( client = fopen( line_buffer, "r" )) == NULL )
              {
                  fclose( output );
                  perror( name );
                  exit( EXIT_FAILURE );
              }

              for( ; ; )
              {
                  fgets( line_buffer, BUFFER_SIZE, client );
                  if ( feof( client ))
                      break;
                  
                  fputs( line_buffer, output );
                  fflush( output );
              }

              fclose( client );
              fclose( output );

              waitpid( pid, NULL, WNOHANG | WUNTRACED );
              pid = 0;
              exit( EXIT_SUCCESS );
              break;
      }
  }
  
  fclose( output );
  perror( name );
  exit( EXIT_FAILURE );
}
