 /*
    News Peruser Copyright (c) 1996, 1997 James Bailie
    ==================================================================

    News Peruser is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2, or (at
    your option) any later version.

    News Peruser is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    Although News Peruser is licensed under the Free Software
    Foundation's GNU General Public License, Peruser is not produced
    by, nor is it endorsed by the Free Software Foundation. The Free
    Software Foundation is not responsible for developing,
    distributing, or supporting Peruser in any way. Anyone may place
    software they own the copyright to, under the GNU General Public
    License.

    The GNU General Public License is included in the News Peruser 
    distribution archive in a file called COPYING. If you do
    not have a copy of the license, you can download one from
    ftp://prep.ai.mit.edu, or you can write to the Free Software
    Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    =====================================================================
*/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<signal.h>

#include "regex.h"

#define BUFFER_SIZE 1024

const char *name = "filterer";
const char *home;
const char *result;

char read_status[ 3 ];

char line_buffer[ BUFFER_SIZE ], folder_name[ BUFFER_SIZE ],
    folder_filter[ BUFFER_SIZE ];

FILE *output, *spool, *tmp_spool, *read, *tmp_read,
    *user_folders, *folder, *folder_read, *temp;

enum{ no, yes }found;

int return_code;

struct re_pattern_buffer buffer;
struct re_registers registers;

void signal_handler( int signal )
{
    if ( output )
        fclose( output );
    if ( spool )
        fclose( spool );
    if ( user_folders )
        fclose( user_folders );
    if ( folder )
        fclose( folder );

    exit( signal );
}


int main ( int argc, char **argv[] )
{
    signal( SIGINT, signal_handler );
    signal( SIGTERM, signal_handler );
    signal( SIGHUP, signal_handler );


    output = spool = folder = NULL;

    home = getenv( "HOME" );
    
    if ( strstr( ( char *)argv[ 0 ], "cfilterer" ))
    {
        sprintf( line_buffer, "%s/.peruser_spool/sorter_pipe",
                 home );
        if (( output = fopen( line_buffer, "w" )) == NULL )
        {
            perror( name );
            signal_handler( EXIT_FAILURE );
        }
    }
    else
        output = stdout;
    
    sprintf( line_buffer, "%s/.peruser_spool/.user-folders",
             home );
    if (( user_folders = fopen( line_buffer, "r" )) == NULL )
    {
        if ( errno == 2 )
            signal_handler( EXIT_SUCCESS );
        else
        {
            perror( name );
            signal_handler( EXIT_FAILURE );
        }
    }

    sprintf( line_buffer, "%s/.peruser_spool/INBOX", home );
    if (( spool = fopen( line_buffer, "r" )) == NULL )
    {
        if ( errno == 2 )
            signal_handler( EXIT_SUCCESS );
        else
        {
            perror( name );
            signal_handler( EXIT_FAILURE );
        }
    }

    sprintf( line_buffer, "%s/.peruser_spool/tmp-INBOX",
             home );
    if (( tmp_spool = fopen( line_buffer, "w" )) == NULL )
    {
        perror( name );
        exit( EXIT_FAILURE );
    }

    sprintf( line_buffer, "%s/.peruser_spool/INBOX.read",
             home );
    if (( read = fopen( line_buffer, "r" )) == NULL )
    {
        perror( name );
        exit( EXIT_FAILURE );
    }

    sprintf( line_buffer, "%s/.peruser_spool/tmp-read",
             home );
    if (( tmp_read = fopen( line_buffer, "w" )) == NULL )
    {
        perror( name );
        exit( EXIT_FAILURE );
    }
    
    for( ; ; )
    {
        fgets( line_buffer, BUFFER_SIZE, user_folders );
        if( feof( user_folders ))
            break;

        
        strcpy( folder_name, strtok( line_buffer, ";" ));
        strcpy( folder_filter, strtok( NULL, "\n" ));

        fprintf( output, "filterer: scanning for folder: %s...\n", folder_name );
        fflush( output );

        re_syntax_options = RE_NO_BK_VBAR | RE_NO_BK_PARENS;
        
        buffer.translate = NULL;
        buffer.fastmap = ( char *)1;
        buffer.buffer = NULL;
        buffer.allocated = 0;

        if ( ( result = re_compile_pattern( folder_filter,
					    strlen( folder_filter ),
					    &buffer )) )
        {
            fprintf( output, "filterer: error compiling regex: %s\n", result );
            fprintf( output, "filterer: for folder %s.\n", folder_name );
            fflush( output );
            continue;
        }

        sprintf( line_buffer, "%s/.peruser_spool/%s", home,
                 folder_name );
        if (( folder = fopen( line_buffer, "a" )) == NULL )
        {
            perror( name );
            signal_handler( EXIT_FAILURE );
        }

        strcat( line_buffer, ".read" );
        if (( folder_read = fopen( line_buffer, "a" )) == NULL )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }

        for( ; ; )
        {
            sprintf( line_buffer, "%s/.peruser_spool/.tmp",
                     home );
            if (( temp = fopen( line_buffer, "w+" )) == NULL )
            {
                perror( name );
                signal_handler( EXIT_FAILURE );
            }
            
            do
            {
                fgets( line_buffer, BUFFER_SIZE, spool );
                if ( feof( spool ))
                    break;
                fputs( line_buffer, temp );
            }
            while( strncmp( line_buffer, ".\r\n", 3 ));
            
            if ( feof( spool ))
            {
                fclose( temp );
                break;
            }

            fgets( read_status, sizeof read_status, read );
            
            rewind( temp );
            found = no;
            
            for( ; ; )
            {
                fgets( line_buffer, BUFFER_SIZE, temp );
                if ( line_buffer[ 0 ] == '\r' || feof( temp ))
                    break;

                return_code = re_search( &buffer, line_buffer,
                                         strlen( line_buffer ), 0,
                                         strlen( line_buffer ),
                                         &registers );
            
                switch( return_code )
                {
                    case -2:
                        fprintf( output,
                                 "filterer: regex library internal error\n" );
                        signal_handler( EXIT_FAILURE );
                        break;

                    case 0:
                        found = yes;
                        break;

                    default:
                        continue;
                        break;
                }

                if ( found )
                    break;
            }
            
            rewind( temp );
            
            if ( found )
            {
                fprintf( output, "filterer: filtering message into folder: %s\n",
                         folder_name );
                fflush( output );

                do
                {
                    fgets( line_buffer, BUFFER_SIZE, temp );
                    fputs( line_buffer, folder );
                }
                while( strncmp( line_buffer, ".\r\n", 3 ));
                    
                fputs( read_status, folder_read );
            }
            else
            {
                do
                {
                    fgets( line_buffer, BUFFER_SIZE, temp );
                    fputs( line_buffer, tmp_spool );
                }
                while( strncmp( line_buffer, ".\r\n", 3 ));

                fputs( read_status, tmp_read );
            }

            fclose( temp );
        }

        fclose( folder );
        fclose( folder_read );

        fclose( spool );
        fclose( tmp_spool );
        
        sprintf( line_buffer, "%s/.peruser_spool/tmp-INBOX",
                 home );
        sprintf( folder_name, "%s/.peruser_spool/INBOX",
                 home );

        rename( line_buffer, folder_name );

        if (( spool = fopen( folder_name, "r" )) == NULL )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }

        if (( tmp_spool = fopen( line_buffer, "w" )) == NULL )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }
        
        fclose( read );
        fclose( tmp_read );

        sprintf( line_buffer, "%s/.peruser_spool/tmp-read",
                 home );
        sprintf( folder_name, "%s/.peruser_spool/INBOX.read",
                 home );

        rename( line_buffer, folder_name );

        if (( read = fopen( folder_name, "r" )) == NULL )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }

        if (( tmp_read = fopen( line_buffer, "w" )) == NULL )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }
    }

    fclose( user_folders );

    fclose( spool );
    fclose( tmp_spool );
    fclose( read );
    fclose( tmp_read );

    return 0;
}



