/* 
   Hearer, Copyright (C) 1995, 1996, 1997 James Bailie
   ==================================================================

   Hearer is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Hearer is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
   License for more details.

   Although Hearer is licensed under the Free Software Foundation's GNU
   General Public License, Hearer is not produced by, nor is it endorsed
   by the Free Software Foundation. The Free Software Foundation is
   not responsible for developing, distributing, or supporting Hearer
   in any way. Anyone may place software they own the copyright to,
   under the GNU General Public License.

   The GNU General Public License is included in the News Peruser
   distribution archive in a file called COPYING. If you do not have a
   copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   =====================================================================
*/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<signal.h>
#include<sys/socket.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/wait.h>
#include<fcntl.h>
#include<ctype.h>
#include<netinet/in.h>
#include<netdb.h>
#include<unistd.h>

#define BUFFER_SIZE 1024


pid_t pid;

const char *name = "hearer";

int news_socket;  

FILE *news_server, *newsrc, *updated_newsrc, 
    *spool_file, *temp_spool_file, *read_file, 
    *temp_read_file, *requests_file, *read_pointer,
    *spool_pointer, *output, *client;

char line_buffer[ BUFFER_SIZE ], copy_buffer[ BUFFER_SIZE ],
     spool_path[ BUFFER_SIZE ], read_path[ BUFFER_SIZE ],
     temp_spool_path[ BUFFER_SIZE ], temp_read_path[ BUFFER_SIZE ],
     requests_path[ BUFFER_SIZE ], reference_id[ BUFFER_SIZE ],
     news_group[ BUFFER_SIZE ];

char *home_directory;

typedef enum { stream_closed, stream_open } stream_state;

enum { no, yes } headers, requests, found, authenticating, reference_request;

int newsrc_line, response_code, next_position, next_option;

struct stream_record 
{
  stream_state news_socket_fd;
  stream_state news_server_stream;
  stream_state newsrc_stream;
  stream_state updated_newsrc_stream;
  stream_state spool_file_stream;
  stream_state temp_spool_file_stream;
  stream_state read_file_stream;
  stream_state temp_read_file_stream;
  stream_state requests_file_stream;
  stream_state output_stream;
  stream_state client_stream;
} open_streams;

void close_files( void )
{
  if ( open_streams.spool_file_stream )
    {
      fflush( spool_file );
      fclose( spool_file );
    }
  if ( open_streams.temp_spool_file_stream )
    {
      fflush( temp_spool_file );
      fclose( temp_spool_file );
    }
  if ( open_streams.read_file_stream )
    {
      fflush( read_file );
      fclose( read_file );
    }
  if ( open_streams.temp_read_file_stream )
    {
      fflush( temp_read_file );
      fclose( temp_read_file );
    }
  if ( open_streams.updated_newsrc_stream )
    {
      fflush( updated_newsrc );
      fclose( updated_newsrc );
    }
  if ( open_streams.newsrc_stream )
    {
      fflush( newsrc );
      fclose( newsrc );
    }

  if ( open_streams.news_server_stream )
      fclose( news_server );

  if ( open_streams.news_socket_fd )
    {
      shutdown( news_socket, 2 );
      close( news_socket );
    }

  if ( open_streams.requests_file_stream )
    fclose( requests_file );

  if ( open_streams.output_stream )
      {
          fflush( output );
          fclose( output );
      }

  if ( open_streams.client_stream )
      fclose( client );
  
  return;
}

void graceful_exit( int signal_number )
{

/*
   Linux resets a signal's handler to its default once it has been raised,
   so we shut off the the signal we're handling to prevent interruption.
*/

  signal( signal_number, SIG_IGN );

  close_files();

/* 
   If a spool file or read file is open we remove the current spool
   and read files. Draconian, but it saves me from having to deal
   with possibly empty articles and mismatched spool and read files.
*/


  if (( open_streams.spool_file_stream ) || ( open_streams.read_file_stream ))
    {
      remove( spool_path );
      strcat( spool_path, ".read" );
      remove( spool_path );
    }

  if ( pid )
      {
          if ( kill( pid, SIGINT ))
	    kill( pid, SIGKILL );

          waitpid( pid, NULL, WNOHANG | WUNTRACED );
      }

  exit( signal_number );
}

int has_xover()
{
  enum{ no, yes }found;

  found = no;

  fputs( "help\r\n", news_server );
  while( strncmp( fgets( line_buffer, BUFFER_SIZE, news_server ), ".\r\n", 3 ))
    if ( strstr( line_buffer, "xover" ))
      found = yes;

  return found;
}

void unrecoverable_error( char *server, char* message )
{
  fprintf( output, "hearer: %s responded with a code\n", server );
  fprintf( output, "hearer: I don't know how to handle:\n" );
  fprintf( output, "hearer: %s\n", message );

  graceful_exit( SIGINT );
}

void get_headers( int requested, int last, char *name )
{
  char *line_pointer;
  int total, i;

  total = last - requested + 1;
  i = 0;

  fprintf( news_server, "xover %d-%d\r\n", requested, last );
  fprintf( output, "\nhearer: xover %d-%d\n", requested, last );

  fgets( line_buffer, BUFFER_SIZE, news_server );
  fprintf( output, "server: %s", line_buffer );

  if ( strncmp( line_buffer, "224", 3 ))
    unrecoverable_error( name, line_buffer );

  while( strncmp( fgets( line_buffer, BUFFER_SIZE, news_server ),
		  ".\r\n", 3 ))
    {
      line_pointer = strtok( line_buffer, "\t" );
      fprintf( spool_file, "@C22Header of article number %s:\r\n", 
	       line_pointer );

      fprintf( output, "        (receiving header %d of %d)\n",
	       ++i, total );
      fflush( output );

      fputs( "u\n", read_file );

      line_pointer = strtok( NULL, "\t" );
      fprintf( spool_file, "Subject: %s\r\n", line_pointer );

      line_pointer = strtok( NULL, "\t" );
      fprintf( spool_file, "From: %s\r\n", line_pointer );

      line_pointer = strtok( NULL, "\t" );
      fprintf( spool_file, "Date: %s\r\n", line_pointer );

      line_pointer = strtok( NULL, "\t" );
      fprintf( spool_file, "Message-ID: %s\r\n", line_pointer );

      line_pointer = strtok( NULL, "\t" );

      if ( ( line_pointer = NULL ) )
	{
	  fputs( ".\r\n", spool_file );
	  continue;
	}

      if ( *line_pointer == '<' )
	{
	  fprintf( spool_file, "References: %s\r\n", line_pointer );
	  line_pointer = strtok( NULL, "\t" );
	}

      line_pointer = strtok( NULL, "\t" );
      fprintf( spool_file, "Lines: %s\r\n", line_pointer );
      fputs( ".\r\n", spool_file );
    }
    
  return;
}

int check_response( char *server )
{	    
/* 
   220 indicates success, while 430 indicates no such article found,
   and 423 indicates a bad article number. Hearer jumps to the next article
   upon receiving either of these codes.
*/

  strcpy( copy_buffer, line_buffer );
  response_code = atoi( strtok( line_buffer, " " ));
  switch ( response_code )
    {
    case 220:
    case 221:
      break;
    case 423:
    case 430:
      return 1;
      break;
    default:
      unrecoverable_error( server, copy_buffer );
      break;
    }
  return 0;
}

void get_list( char *server )
{
    enum{ no, yes }new;
    char timestamp[ 64 ];
    
    new = yes;
    
    sprintf( line_buffer, "%s/.peruser_spool/LIST", 
             home_directory );
    if (( spool_file = fopen( line_buffer, "r" )) == NULL )
    {
        if ( errno == 2)
        {
            new = no;
            if (( spool_file = fopen( line_buffer, "w" )) == NULL )
            {
                perror( name );
                graceful_exit( SIGINT );
            }
            open_streams.spool_file_stream = stream_open;
        }
        else
        {
            perror( name );
            exit( EXIT_FAILURE );
        }
    }
    else
    {
        sprintf( line_buffer, "%s/.peruser_spool/.timestamp",
                 home_directory );
        if (( temp_spool_file = fopen( line_buffer, "r" )) == NULL )
        {
            if ( errno != 2 )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }
            else
                new = no;
        }
        
        if ( new )
        {
            fgets( timestamp, 64, temp_spool_file );
            fclose( temp_spool_file );
            
            sprintf( line_buffer, "%s/.peruser_spool/LIST",
                     home_directory );
            fclose( spool_file );
            if (( spool_file = fopen( line_buffer, "a" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }
            open_streams.spool_file_stream = stream_open;
        }
    }

    if ( new )
    {
        sprintf( line_buffer, "newgroups %s GMT\r\n", strtok( timestamp, "\n" ));
        fputs( line_buffer, news_server );
        fprintf( output, "hearer: newgroups %s GMT\n", timestamp );
        fflush( output );
    }
    else
    {
            fputs( "list\r\n", news_server );
            fprintf( output, "hearer: list\n" );
            fflush( output );
    }
  
    fgets( line_buffer, sizeof line_buffer, news_server );
    fprintf( output, "server: %s", line_buffer );
    fflush( output );
  		
    if ( new )
    {
        if ( strncmp( line_buffer, "231", 3 ))
            unrecoverable_error( server, line_buffer );
    }
    else
        if ( strncmp( line_buffer, "215", 3 ))
            unrecoverable_error( server, line_buffer );
		
    fprintf( output,
             "        (retrieving %s. This might just take a while...)\n",
             new ? "NEWGROUPS" : "LIST ");
    fflush( output );
  
    do
    {
        fgets( line_buffer, sizeof line_buffer, 
               news_server );
        strcpy( copy_buffer, line_buffer );

        if ( line_buffer[ 0 ] != '.' &&
             line_buffer[ 0 ] != '\n' )
            fprintf( spool_file, "%s\n", 
                     strtok( copy_buffer, " " ));
      
        if ( strncmp( copy_buffer, ".\r\n", 3 ))
            fprintf( output, "(%s)\n", copy_buffer );
        else
            fprintf( output, ".\n" );
        
        fflush( output );
    }
    while( strncmp( line_buffer, ".\r\n", 3 ));
		
    fflush( spool_file );	
    fclose( spool_file );		
    open_streams.spool_file_stream = stream_closed;
		
    fputs( "quit\r\n", news_server );
    fprintf( output, "hearer: quit\n" );
    fflush( output );
  
    fgets( line_buffer, sizeof line_buffer, news_server );
    fprintf( output, "server: %s", line_buffer );
    fflush( output );

    graceful_exit( 0 );
}

void close_request_files()
{
  fclose( spool_file );
  open_streams.spool_file_stream = stream_closed;

  fclose( read_file );
  open_streams.read_file_stream = stream_closed;

  if ( !reference_request )
    {
      fflush( temp_spool_file );
      fclose( temp_spool_file );
      open_streams.temp_spool_file_stream = stream_closed;

      fflush( temp_read_file );
      fclose( temp_read_file );
      open_streams.temp_read_file_stream = stream_closed;

      rename( temp_spool_path, spool_path );
      rename( temp_read_path, read_path );
    }
}
void get_requests( char *server )
{
  for( ; ; )
    {
      fgets( line_buffer, BUFFER_SIZE, requests_file );
      if ( feof( requests_file ))
	break;

      if ( line_buffer[ 0 ] == '@' )
	reference_request = no;
      else
	reference_request = yes;

      sprintf( spool_path, "%s/.peruser_spool/%s", home_directory, 
	       news_group );
      if (( spool_file = fopen( spool_path, "a+" )) == NULL )
	{
	  perror( name );
	  graceful_exit( SIGINT );
	}
      open_streams.spool_file_stream = stream_open;      

      sprintf( read_path, "%s.read", spool_path );
      if (( read_file = fopen( read_path, "a+" )) == NULL )
	{
	  perror( name );
	  graceful_exit( SIGINT );
	}
      open_streams.read_file_stream = stream_open;

      if ( !reference_request )
	{
	  sprintf( temp_spool_path, "%s/.peruser_spool/%s.temp",
		   home_directory, news_group );
	  if (( temp_spool_file = fopen( temp_spool_path, "w" )) == NULL )
	    {
	      perror( name );
	      graceful_exit( EXIT_FAILURE );
	    }
	  open_streams.temp_spool_file_stream = stream_open;

	  sprintf( temp_read_path, "%s/.peruser_spool/%s.read.temp",
		   home_directory, news_group );
	  if (( temp_read_file = fopen( temp_read_path, "w" )) == NULL )
	    {
	      perror( name );
	      graceful_exit( EXIT_FAILURE );
	    }
	  open_streams.temp_read_file_stream = stream_open;

	  rewind( spool_file );
	  rewind( read_file );
	  for( ; ; )
	    {
	      fgets( copy_buffer, BUFFER_SIZE, spool_file );
	      if ( feof( spool_file ))
		break;

	      if ( !strncmp( copy_buffer, ".\r\n", 3 ))
		{
		  fputs( copy_buffer, temp_spool_file );
		  fgets( copy_buffer, BUFFER_SIZE, read_file );
		  fputs( copy_buffer, temp_read_file );
		  continue;
		}
	      if ( !strcmp( copy_buffer, line_buffer ))
		{
		  found = yes;
		  break;
		}
	      fputs( copy_buffer, temp_spool_file );
	    }
	}

      if ( reference_request )
	{
	  fprintf( news_server, "article %s\r\n", 
		   strcpy( reference_id, strtok( line_buffer, ":" )));
	  fprintf( output, "hearer: article %s\n", line_buffer );
          fflush( output );
	}
      else
	{
	  strtok( line_buffer, " " );
	  strtok( NULL, " " );
	  strtok( NULL, " " );
	  strtok( NULL, " " );

	  sprintf( copy_buffer, "article %d\r\n", atoi( strtok( NULL, ":" )));
	  strcpy( line_buffer, copy_buffer );

	  fprintf( news_server, line_buffer );
	  fprintf( output, "hearer: %s\n", strtok( line_buffer, "\r\n" ));
          fflush( output );
	}
	      
      fgets( line_buffer, BUFFER_SIZE, news_server );
      fprintf( output, "server: %s", line_buffer );
      fflush( output );
      
      if ( check_response( server ) )
	{
	  close_request_files();
	  continue;
	}

      if ( reference_request )
	{
	  spool_pointer = spool_file;
	  read_pointer = read_file;
	}
      else
	{
	  spool_pointer = temp_spool_file;
	  read_pointer = temp_read_file;
	}

      fprintf( output, "        (receiving article...)\n" );
      fflush( output );
      
      do
	{
	  fgets( line_buffer, BUFFER_SIZE, news_server );
	  fputs( line_buffer, spool_pointer );
	}
      while( strncmp( line_buffer, ".\r\n", 3 ));
      fputs( "u\n", read_pointer );

      if ( found  )
	{
	  found = no;
	  fprintf( output,
                   "        (replacing header with full article text.)\n" );
          fflush( output );
          
	  do
	    {
	      fgets( copy_buffer, BUFFER_SIZE, spool_file );
	    }
	  while( strncmp( copy_buffer, ".\r\n", 3 ));

	  fgets( copy_buffer, BUFFER_SIZE, read_file );

	  for ( ; ; )
	    {
	      fgets( copy_buffer, BUFFER_SIZE, spool_file );
	      if ( feof( spool_file ))
		break;
	      if ( !strncmp( copy_buffer, ".\r\n", 3 ))
		{
		  fgets( line_buffer, BUFFER_SIZE, read_file );
		  fputs( line_buffer, temp_read_file );
		}
	      fputs( copy_buffer, temp_spool_file );
	    }
	}
      close_request_files();
    }

  fclose( requests_file );
  truncate( requests_path, 0 );
  open_streams.requests_file_stream = stream_closed;
}

void authenticate( char *server, char *username, char *password)
{
  fprintf( news_server, "authinfo user %s\r\n", username );
  fprintf( output, "hearer: authinfo user %s\n", username );
  fflush( output );
  
  fgets( line_buffer, BUFFER_SIZE, news_server ); 
  strcpy( copy_buffer, line_buffer );
  sprintf( line_buffer, "server: %s", copy_buffer );
  fprintf( output, line_buffer );
  fflush( output );
  
  if ( atoi( strtok( copy_buffer, " " )) != 381 )
    unrecoverable_error( server, copy_buffer );

  fprintf( news_server, "authinfo pass %s\r\n", password );
  fprintf( output, "hearer: authinfo pass <password>\n" );
  fflush( output );
  
  fgets( line_buffer, BUFFER_SIZE, news_server );
  strcpy( copy_buffer, line_buffer );
  sprintf( line_buffer, "server: %s", copy_buffer );
  fprintf( output, line_buffer );
  fflush( output );
  
  if ( atoi( strtok( copy_buffer, " " )) != 281 )
    unrecoverable_error( server, copy_buffer );

  return;
}

int main( int argc, char *argv[] )
{
    unsigned int last_read, total_articles, first_article,
               last_article, requested_article;

  struct protoent *protocol;
  struct hostent *host;
  struct servent *service;
  struct sockaddr_in socket_address;

  headers = requests = found = authenticating = no;

/*
   Establish signal handler.
*/

  signal( SIGTERM, graceful_exit );
  signal( SIGINT, graceful_exit );
  signal( SIGHUP, graceful_exit );
  
/*
   Keep track of the state of streams.
*/

  open_streams.news_socket_fd = stream_closed;
  open_streams.news_server_stream = stream_closed;
  open_streams.newsrc_stream = stream_closed;
  open_streams.updated_newsrc_stream = stream_closed;
  open_streams.spool_file_stream = stream_closed;
  open_streams.temp_spool_file_stream = stream_closed;
  open_streams.read_file_stream = stream_closed;
  open_streams.temp_read_file_stream = stream_closed;
  open_streams.requests_file_stream = stream_closed;
  open_streams.output_stream = stream_closed;
  open_streams.client_stream = stream_closed;
  
/*
  Figure out if we've been execvp'ed from Peruser
*/

  home_directory = getenv( "HOME" );

  
  if ( strstr( argv[ 0 ], "chearer" ))
      {
          sprintf( line_buffer, "%s/.peruser_spool/client_pipe",
                   home_directory );
          if (( output = fopen( line_buffer, "w" )) == NULL )
          {
              if ( errno != EEXIST )
              {
                  perror( name );
                  graceful_exit( EXIT_FAILURE );
              }
          }
          open_streams.output_stream = stream_open;
      }
  else 
      output = stdout;

/*
   The NNTP server's name is passed as a command line argument.
*/
  
  if ( argc == 1 )
    {
      fprintf( output, "hearer: You must supply the NNTP server's name\n" );
      fprintf( output, "hearer: as a command line argument to Hearer.\n" );
      graceful_exit( SIGINT );
    }

/*
   Lookup the protocol number assigned to TCP. 
*/

  if ( !( protocol = getprotobyname( "tcp" )))
    {
      fprintf( output, "hearer: Couldn't find a protocol number for \"tcp\"\n" );
      fprintf( output, "hearer: in /etc/protocols.\n" );
      graceful_exit( SIGINT );
    }

/*
   Lookup the service number assigned to NNTP.
*/

  if ( !( service = getservbyname( "nntp", "tcp" )))
    {
      fprintf( output, "hearer: Couldn't find a service number for \"nntp\"\n" );
      fprintf( output, "hearer: in /etc/services.\n" );
      graceful_exit( SIGINT );
    }

/*
   Lookup the address of the NNTP server named on the command line.
*/

  if ( !( host = gethostbyname( argv[ 1 ] )))
     {
       herror( name );
       graceful_exit( SIGINT );
     }

/*
   Create a socket for communication with the NNTP server.
*/

  if (( news_socket = socket( AF_INET, SOCK_STREAM, protocol->p_proto ))
      == -1 )
    {
      perror( name );
      graceful_exit( SIGINT );
    }

/*
   Zero and initialize the socket_address structure.
*/

  memset( &socket_address, 0, sizeof socket_address );
  socket_address.sin_family = AF_INET;
  memmove( (char *)&socket_address.sin_addr, host->h_addr, host->h_length );
  socket_address.sin_port = service->s_port;

/*
   Connect to the NNTP server through the socket.
*/

  if ( connect( news_socket, ( struct sockaddr *)&socket_address, 
	        sizeof socket_address ) == -1 )
    {
      perror( name );
      graceful_exit( SIGINT );
    }
  open_streams.news_socket_fd = stream_open;

/*
   Create a stream and associate it with the socket.
*/

  if ( !( news_server = fdopen( news_socket, "r+" )))
    {
      fprintf( output, "hearer: fdopen( news_socket, \"r+\" ) failed!\n" );
      graceful_exit( SIGINT );
    }
  open_streams.news_server_stream = stream_open;

/*
   Get the welcome message from the news server.
*/

  fgets( line_buffer, sizeof line_buffer, news_server );
  fprintf( output, line_buffer );

/* 
   A three-digit response code starts all NNTP server replies. The
   three digits are represented as three characters of ASCII,
   separated from any message text by a single space.

   There are two acceptable response codes after a connection has been
   made to a NNTP server: "200" or "201". The "2" indicates a command
   succeeded, the zero the command that succeeded, in this case the
   connect. The last digit indicates either posting allowed: "0", or
   posting not allowed: "1". Since it only downloads news, Hearer
   doesn't care about the last digit.
*/

  if ( strncmp( line_buffer, "20", 2 ))
    unrecoverable_error( argv[ 1 ], line_buffer );

/*
   Create the .peruser_spool directory if it doesn't exist.
*/

  sprintf( line_buffer, "%s/.peruser_spool", home_directory );
  if ( mkdir( line_buffer, 0755 ))
    {
      if ( errno != EEXIST )
	{
	  perror( name );
	  graceful_exit( SIGINT );
	}
    }

/*
   If user wants list, headers, requests...
*/

  if ( argc > 2 )
    {
      if ( !strncmp( argv[ 2 ], "-a", 2 ))
	{
	  if ( argc < 5 )
	    {
	      fprintf( output, "hearer: You must supply your NNTP server\n" );
	      fprintf( output, "hearer: login name and password,\n" );
	      fprintf( output, "hearer: respectively, after the -a option.\n" );

	      graceful_exit( SIGINT );
	    }

	  authenticating = yes;
	  next_position = 5;
	}
      else
	next_position = 2;

      if ( argc > next_position )
	{
	  if ( isdigit( argv[ next_position ][ 0 ] ))
	    next_option = next_position + 1;
	  else
	    next_option = next_position;

	  if ( argc > next_option )
	    {
	      if ( !strncmp( argv[ next_option ], "requests", 8 ))
		requests = yes;
	      else
		if ( !strncmp( argv[ next_option ], "headers", 7 ))
		  headers = yes;
		else
		  if ( !strncmp( argv[ next_option ], "list", 4 ))
		    {
		      if ( authenticating )
			authenticate( argv[ 1 ], argv[ 3 ], argv[ 4 ] );
		      get_list( argv[ 1 ] );
		    }
	    }
	}
    }

  if ( authenticating )
    authenticate( argv[ 1 ], argv[ 3 ], argv[ 4 ] );


  /*
    Find and open the user's .peruser-newsrc
   */

  sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );
  if (( newsrc = fopen( line_buffer, "r" )) == NULL )
    if ( errno == 2 )
      {
	fprintf( output, "hearer: cannot find ~/.peruser-newsrc.\n" );
	fprintf( output, "hearer: you must create this file before\n" );
	fprintf( output, "hearer: you can run hearer. See the file\n" );
	fprintf( output, "hearer: INSTALL in the peruser source archive.\n" );
	graceful_exit( EXIT_FAILURE );
      }
    else
      {
	perror( name );
	graceful_exit( EXIT_FAILURE );
      }
  open_streams.newsrc_stream = stream_open;

/*
  Check for any lines in ~/.peruser-newsrc without colons.
*/

  for( newsrc_line = 1; ; newsrc_line++ )
    {
      fgets( line_buffer, BUFFER_SIZE, newsrc );
      if ( feof( newsrc ))
	break;
      if ( line_buffer[ 0 ] == '\n' )
	continue;

      if ( !strchr( line_buffer, ':' ))
	{
	  fprintf( output, 
		   "hearer: Line %d in ~/.peruser-newsrc does not\n",
		   newsrc_line );
	  fprintf( output, 
		   "hearer: contain a colon. I cannot parse the\n" );
	  fprintf( output, 
		   "hearer: newsgroup name! If you've added lines\n" );
	  fprintf( output, 
		   "hearer: to your ~/.peruser-newsrc, you need to\n" );
	  fprintf( output,
		   "hearer: terminate each new line with a colon!\n" );
	  graceful_exit( SIGINT );
	}
    }
  rewind( newsrc );


  /*
   Create the file to hold the updated .newsrc.
*/

  if ( !requests )
    {
      sprintf( line_buffer, "%s/.updated-newsrc", home_directory );
      if (( updated_newsrc = fopen( line_buffer, "w" )) == NULL )
	{
	  perror( name );
	  graceful_exit( EXIT_FAILURE );
	}
      open_streams.updated_newsrc_stream = stream_open;
    }

  /*
    This is the main loop.
   */

  for( ; ; )
    {

/* 
   Read a line of .peruser-newsrc, parsing for news group name, first and last
   read article.
*/

      fgets( line_buffer, sizeof line_buffer, newsrc );
      if ( feof( newsrc ))
	break;
      if ( line_buffer[ 0 ] == '\n' )
	  continue;
      strcpy( news_group, strtok( line_buffer, ":" ));
      last_read = atoi( strtok( NULL, "-" ));
      if ( last_read )
	last_read = atoi( strtok( NULL, "\n" ));

/*
  If the user wants to request full text of selected headers, check
  to see if there are any requests on file for the current newsgroup.
*/

      if ( requests )
	{
	  sprintf( requests_path, "%s/.peruser_spool/%s.requests",
		   home_directory, news_group );
	  if (( requests_file = fopen( requests_path, "r" )) == NULL )
	    {
	      if ( errno == 2 )
		continue;
	      else
		{
		  perror( name );
		  graceful_exit( SIGINT );
		}
	    }
	  open_streams.requests_file_stream = stream_open;
	  line_buffer[ 0 ] = fgetc( requests_file );
	  if ( feof( requests_file ))
	    {
	      fclose( requests_file );
	      open_streams.requests_file_stream = stream_closed;
	      continue;
	    }
	  ungetc( line_buffer[ 0 ], requests_file );
	}
	
/*
   Now query server for group stats.
*/

      fprintf( news_server, "group %s\r\n", news_group );
      fprintf( output, "\nhearer: group %s\n", news_group );
      fgets( line_buffer, sizeof line_buffer, news_server );
      fprintf( output, "server: %s", line_buffer );

/* 
   Make a copy of the server's reply, in case there's an error.
*/

      strcpy( copy_buffer, line_buffer );

/*
   A match on "211" in the first case statement below indicates the
   group selection succeeded. "411" indicates the server does not
   carry the requested news group.

   All other replies cause termination.
*/

      response_code = atoi( strtok( line_buffer, " " ));
      switch ( response_code )
	{
	case 411:
	  continue;
	case 211:
	  break;
	default:
	  unrecoverable_error( argv[ 1 ], copy_buffer );
	}
      total_articles = atoi( strtok( NULL, " " ));
      first_article = atoi( strtok( NULL, " " ));
      last_article = atoi( strtok( NULL, " " ));

/* 
   Open or create the group's spool file. Note that an empty spool
   file will be created for empty groups.
*/

      if ( !requests )
	{
	  sprintf( line_buffer, "%s/.peruser_spool/%s", home_directory, 
		   news_group );
	  strcpy( spool_path, line_buffer );
	  if (( spool_file = fopen( line_buffer, "a" )) == NULL )
	    {
	      perror( name );
	      graceful_exit( SIGINT );
	    }
	  open_streams.spool_file_stream = stream_open;      

/*
   Open or create the group's read file. This text file will have
   written to it, a one character line terminated by a newline,
   consisting of a "u" ( for "unread" ) for every article downloaded
   to the corresponding spool file.
*/

	  strcat( line_buffer, ".read" );
	  if (( read_file = fopen( line_buffer, "a" )) == NULL )
	    {
	      perror( name );
	      graceful_exit( SIGINT );
	    }
	  open_streams.read_file_stream = stream_open;
	}

/*
  If user is requesting full text of selected headers then read
  the selected article numbers from requests_file and request them from
  server.
*/

      if ( requests )
	{
	  get_requests( argv[ 1 ] );
	  continue;
	}

/*
   Otherwise, find out at what number unread articles start at.
*/

      if ( !last_article )
	{
	  fprintf( updated_newsrc, "%s:\n", news_group );
	  fprintf( output, "        (no articles in %s)\n", news_group );
	  fclose( spool_file );
	  open_streams.spool_file_stream = stream_closed;
	  fclose( read_file );
	  open_streams.read_file_stream = stream_closed;
	  continue;
	}

      if ( argc > 2 && isdigit( argv[ next_position  ][ 0 ] )) 
	{
	  if ( argv[ next_position ][ 0 ] == '0' )
	    {
	      fprintf( updated_newsrc, "%s: 1-%d\n", news_group, 
		       last_article );
	      fprintf( output, "        (updating .peruser-newsrc)\n" );
	      fclose( spool_file );
	      open_streams.spool_file_stream = stream_closed;
	      fclose( read_file );
	      open_streams.read_file_stream = stream_closed;
	      continue;
	    }
 	  if ( total_articles >= atoi( argv[ next_position ] ))
	    {
	      requested_article = last_article - atoi( argv[ next_position ] );
	      requested_article++;
	    }
	  else
	    {
	      requested_article = last_article - total_articles;
	      requested_article++;
	    }
	}
      else
	if ( last_read == last_article )
	  {
	    fprintf( updated_newsrc, "%s: 1-%d\n", news_group, last_article );
	    fprintf( output, "        (no new articles in %s)\n", news_group );
	    fclose( spool_file );
	    open_streams.spool_file_stream = stream_closed;
	    fclose( read_file );
	    open_streams.read_file_stream = stream_closed;
	    continue;
	  }
	else
	  if (( last_read > last_article ) || ( last_read < first_article ))
	    {
	      requested_article = last_article - total_articles;
	      requested_article++;
	    }
	  else
	    requested_article = last_read + 1;


      if ( has_xover())
	get_headers( requested_article, last_article, argv[ 1 ] );
      else
	{

	  /*
	    Start requesting articles.
	  */

	  for( ; requested_article <= last_article; requested_article++ )
	    {
	      if ( headers )
		{
		  fprintf( news_server, "head %d\r\n", requested_article );
		  fprintf( output, "hearer: head %d\n", requested_article );
		}
	      else
		{
		  fprintf( news_server, "article %d\r\n", requested_article );
		  fprintf( output, "hearer: article %d\n", requested_article );
		}
	      fgets( line_buffer, sizeof line_buffer, news_server );
	      fprintf( output, "server: %s", line_buffer );

	    
	      /* 
		 Check the server's reply for an error message. 
	      */

	      if ( check_response( argv[ 1 ] ) )
		continue;

	      /*
		Retrieve the article, leaving the last line, consisting of a single
		period, intact, for Peruser to use as a delimiter. If a line of the
		article starts with a period, the period will have been doubled by the
		server to avoid ambiguity.
	      */

	      if ( headers )
		fprintf( output, "        (receiving header...)\n" );
	      else
		fprintf( output, "        (receiving article...)\n" );

	      fflush( output );
                    
	      if ( headers )
		{
		  sprintf( line_buffer, "@C22Header of article number %d:\r\n", 
			   requested_article );
		  fputs( line_buffer, spool_file );
		}

	      do 
		{
		  fgets( line_buffer, sizeof line_buffer, news_server );
		  fputs( line_buffer, spool_file );
		} 
	      while ( strncmp( line_buffer, ".\r\n", 3 ));

	      /* 
		 Add line to read file.
	      */

	      fputs( "u\n", read_file );

	    }  /* end of article for-loop.  */
	}
      
      fflush( spool_file );
      fclose( spool_file );
      open_streams.spool_file_stream = stream_closed;

      fflush( read_file );
      fclose( read_file );
      open_streams.read_file_stream = stream_closed;

      /*
   write a line to .updated_newsrc
*/

      fprintf( updated_newsrc, "%s: 1-%d\n", news_group, last_article );

    }  /* end of main loop. */

  /*
   Sign off, clean-up, and we're outta here.
*/

  fputs( "quit\r\n", news_server );
  fprintf( output, "\nhearer: quit\n" );
  fflush( output );
  
  fgets( line_buffer, sizeof line_buffer, news_server );
  fprintf( output, "server: %s", line_buffer );
  fflush( output );

  fprintf( output, "hearer: I have heard the news.\n" );
  fprintf( output, "\n" );
  fflush( output );

  fflush( updated_newsrc );
  fclose( updated_newsrc );
  open_streams.updated_newsrc_stream = stream_closed;

  fclose( newsrc );
  open_streams.newsrc_stream = stream_closed;
  
  if ( !requests )
    {
      sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );
      sprintf( copy_buffer, "%s/.updated-newsrc", home_directory );
      rename( copy_buffer, line_buffer );
    }

  if ( output == stdout )
    {
      if ( !requests )
	execlp( "sorter", "sorter", NULL );
      else
	execlp( "threader", "threader", NULL );
    }
  else
    {
      if ( requests )
	{
          sprintf( line_buffer, "%s/.peruser_spool/threader_pipe",
                   home_directory );
          if ( mkfifo( line_buffer, 0700 ))
	    {
              if ( errno != EEXIST )
		{
                  perror( name );
                  graceful_exit( EXIT_FAILURE );
		}
	    }
	}
      else
	{
          sprintf( line_buffer, "%s/.peruser_spool/sorter_pipe",
                   home_directory );
          if ( mkfifo( line_buffer, 0700 ))
	    {
              if ( errno != EEXIST )
		{
                  perror( name );
                  graceful_exit( EXIT_FAILURE );
		}
	    }
	}
      
      pid = 0;
      switch( pid = fork() )
	{
	case -1:
	  perror( name );
	  graceful_exit( EXIT_FAILURE );
	  break;
	case 0:
	  if ( requests )
	    execlp( "cthreader", "cthreader", NULL );
	  else
	    execlp( "csorter", "csorter", NULL );

	  perror( name );
	  graceful_exit( EXIT_FAILURE );
	  break;
	default:
	  if (( client = fopen( line_buffer, "r" )) == NULL )
	    {
	      perror( name );
	      graceful_exit( EXIT_FAILURE );
	    }
	  open_streams.client_stream = stream_open;

	  for( ; ; )
	    {
	      fgets( line_buffer, BUFFER_SIZE, client );
	      if ( feof( client ))
		break;
                  
	      fputs( line_buffer, output );
	      fflush( output );
	    }

	  waitpid( pid, NULL, WNOHANG ^ WUNTRACED );
	  pid = 0;
	  graceful_exit( EXIT_SUCCESS );
	  break;
	}
    }
  
  perror( name );
  graceful_exit( EXIT_FAILURE );

  return -1;
}
