/*
   News Peruser Copyright (c) 1996, 1997 James Bailie
   ==================================================================
   
   News Peruser is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   News Peruser is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   Although News Peruser is licensed under the Free Software
   Foundation's GNU General Public License, Peruser is not produced
   by, nor is it endorsed by the Free Software Foundation. The Free
   Software Foundation is not responsible for developing,
   distributing, or supporting Peruser in any way. Anyone may place
   software they own the copyright to, under the GNU General Public
   License.

   The GNU General Public License is included in the News Peruser 
   distribution archive in a file called COPYING. If you do
   not have a copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   =====================================================================
*/

#include<ctype.h>

#include "peruser_includes.h"

void update_read_file();
void refresh_newsgroups();
void toggle_italic( int line );

void export_callback( FL_OBJECT *obj, long caller )
{
    int maximum;
    int index;
    char *filename;

    if ( !fl_get_browser( fd_peruser->subject_browser ))
        return;

    if ( caller )
    {
        sprintf( line_buffer, "%s/.peruser_decode", spool_path );
        filename = line_buffer;
    }
    else
        ( const)filename = fl_show_fselector( "File to hold message:", 
                                              home_directory, "*", 
                                              "exported.txt" );

    if ( !filename ) 
        return;

    if (( temp_file = fopen( filename, "w" )) == NULL )
    {
	perror( name );
	exit( EXIT_FAILURE );
    }

/*
  Get article lines.
*/

    maximum = fl_get_browser_maxline( fd_peruser->message_browser );

    for( index = 1; index <= maximum; index++ )
    {
        strcpy( line_buffer, 
                fl_get_browser_line( fd_peruser->message_browser, index ));

        if ( !strncmp( line_buffer, "@-", 2 ))
            continue;

        if ( !strncmp( line_buffer, "@C1Subject: ", 12 ) ||
             !strncmp( line_buffer, "@C4From: ", 9 ) ||
             !strncmp( line_buffer, "@C5", 3 ))
	{
            strcpy( second_line_buffer, line_buffer + 3 );
            strcpy( line_buffer, second_line_buffer );
	}
        else
            if ( !strncmp( line_buffer, "@C", 2 ))
            {
                strcpy( second_line_buffer, line_buffer + 4 );
                strcpy( line_buffer, second_line_buffer );
            }
            else
                if ( !strncmp( line_buffer, "@i", 2 ))
                    {
                        strcpy( second_line_buffer, line_buffer + 2 );
                        strcpy( line_buffer, second_line_buffer );
                    }
      
        line_pointer = strtok( line_buffer, "\r" );
        if( line_pointer )
	{
            strcpy( second_line_buffer, line_pointer );
            strcat( second_line_buffer, "\n" );
            fputs( second_line_buffer, temp_file );
	}
        else
            fputs( "\n", temp_file );
    }
    fflush( temp_file );
    fclose( temp_file );

/*
  Get the subject line and strip it of initial boldface attribute prefix,
  if present, and trailing carriage return.
*/

    strcpy( line_buffer,
            fl_get_browser_line( fd_peruser->subject_browser,
                                 fl_get_browser( fd_peruser->subject_browser )));

    strcpy( second_line_buffer, strtok( line_buffer, "\r" ));
    if ( second_line_buffer[ 0 ] == '@' )
    {
        line_pointer = second_line_buffer;
        line_pointer++;
        line_pointer++;
        strcpy( line_buffer, line_pointer );
        memcpy( second_line_buffer, line_buffer, BUFFER_SIZE );
    }

    fl_show_message( "messages:", second_line_buffer, 
                     "exported successfully" );
    return;
}

void follow_up_callback( FL_OBJECT *obj, long caller )
{
    char attribution_line[ BUFFER_SIZE ] = "";

    int line;
    int index;

    if ( !article_count ) 
        return;

    fl_set_choices_shortcut( "y", "n", "c" );
    index = fl_show_choice( "Quote original article?", "", "", 3, "yes", "no",
                            "cancel", 2 );
    switch( index )
    {

/*
  fd_article->follow_up_button.
*/

        case 1:
            fl_clear_browser( fd_quoter->quoter_source_browser );
            fl_clear_browser( fd_quoter->quoter_destination_browser );

            line = fl_get_browser_maxline( fd_peruser->message_browser );

/*
  Create attribution line.
*/

            for( index =1; strncmp( line_buffer, "@C4From: ", 6 ); index++ )
                strcpy( line_buffer,
                        fl_get_browser_line( fd_peruser->message_browser, index ));

            line_pointer = strstr( line_buffer, " " );
            sprintf( attribution_line, "%s wrote:\r", 
                     strtok( line_pointer, "\r" ));

            fl_add_browser_line( fd_quoter->quoter_destination_browser,
                                 attribution_line );
            fl_add_browser_line( fd_quoter->quoter_destination_browser,
                                 "\r" );

/*
  Add fd_peruser->message_browser
  to fd_quoter->quoter_source_browser.
*/

            for( index = 1; index <= line; index++ )
                fl_add_browser_line( fd_quoter->quoter_source_browser,
                                     fl_get_browser_line( fd_peruser->message_browser,
                                                          index ));

            switch( caller )
            {
                case 21:
                    reply_by_mail = yes;
                    break;
                case 30:
                    reply_by_mail = both;
                    break;
	  
                default:
                    reply_by_mail = no;
            }      

            fl_show_form( fd_quoter->quoter, FL_PLACE_MOUSE, FL_FULLBORDER, 
                          "Quoter" );

            break;
        case 2:
            switch( caller )
            {
                case 21:
                    reply_by_mail = yes;
                    run_editor( obj, caller );
                    break;
                case 30:
                    reply_by_mail = both;
                    run_editor( obj, caller );
                    break;

                default:
                    reply_by_mail = no;
                    run_editor( obj, 1 );
                    break;
            }

            break;
        case 3:
            break;
    }
    return;
}

char *get_subject( void )
{
    int index;
    int maxline;
    enum{ no, yes } found;

    found = no;

    maxline = fl_get_browser_maxline( fd_peruser->message_browser );

    for( index = 1; index <= maxline; index++ )
    {
        strcpy( line_buffer, fl_get_browser_line( fd_peruser->message_browser,
                                                  index ));

        if ( !strncmp( line_buffer, "@C1Subject: ", 12 ))
	{
            found = yes;
            break;
	}
    }

    if ( !found )
        strcpy( second_line_buffer, "Subject: (no subject)\r\n" );
    else
        if ( strstr( line_buffer, " Re:" ) == NULL && 
             strstr( line_buffer, " re:") == NULL )
        {
            strcpy( second_line_buffer, "Subject: Re: " );
            line_pointer = strstr( line_buffer, " " );
            line_pointer++;
            strcat( second_line_buffer, line_pointer );
	
        }
        else
            strcpy( second_line_buffer, line_buffer + 3 );

    if ( ( line_pointer = strchr( second_line_buffer, '\r' )) != NULL )
        *line_pointer = '\0';

    return( second_line_buffer );
}

char *wrap_header( char *header, char *delimiter, int padding );

void run_editor( FL_OBJECT *obj, long caller )
{
    char local_buffer[ BUFFER_SIZE ];

/*
  variable to hold the values of the usenet header about to be created.
*/

    char from[ BUFFER_SIZE ];
    char date[ 50 ];
    char newsgroups[ BUFFER_SIZE ];
    char subject[ BUFFER_SIZE ];
    char references[ BUFFER_SIZE ];
    char message_id[ BUFFER_SIZE ];
    char user_id[ BUFFER_SIZE ];
    char nntp_domain[ BUFFER_SIZE ]; 
    char path[ BUFFER_SIZE ];
    char to[ BUFFER_SIZE ];
  
/*
  Variable used to convert unix time to acceptable format for header.
*/

    char *month;
    char *day;
    char *hours_minutes_seconds;
    char *year;

    time_t unix_time;
    struct tm *gmt;

    enum{ no, yes } found;

/*
  Counter, and line number currently selected in fd_peruser->group_browser.
*/

    int index;
    int group_line;
    int subject_line;
    int maxline;

/*
  If a newsgroup isn't selected, or a folder is selected,
  return.
*/

    group_line = fl_get_browser( fd_peruser->group_browser );

    if (( !group_line || folder || !strncmp( newsgroup, "INBOX", 5 )) 
        && ( caller < 20 || caller == 30 ))
    {
        fl_show_alert( "Error:",
                       "No newsgroup is selected.", "", 0 );
        return;
    }
  
    subject_line = fl_get_browser( fd_peruser->subject_browser );

/*
  Create From: header line. */

    sprintf( from, "From: %s <%s>", user_name, user_email_address );

/*
  Create Date: header line.
*/

    unix_time = time( NULL );
    gmt = gmtime( &unix_time );
    strcpy( line_buffer, asctime( gmt ));
    month = strtok( line_buffer, " " );
    month = strtok( NULL, " " );
    day = strtok( NULL, " " );
    hours_minutes_seconds = strtok( NULL, " " );
    year = strtok( NULL, "\n" );
  
    sprintf( date, "Date: %s %s %s %s GMT", day, month, year,
             hours_minutes_seconds );

/*
  Create Newsgroups: header line. If we're creating a follow-up, the
  "Newsgroups: " header is taken from the original article, with a multiline
  "Newsgroups: " header being concatenated into a single line.
*/

    switch( caller )
    {
        case 0:
            strcpy( line_buffer, 
                    fl_get_browser_line( fd_peruser->group_browser,
                                         fl_get_browser( 
                                                        fd_peruser->group_browser )));
            line_pointer = line_buffer;
            if ( !strncmp( line_buffer, "@b", 2 ))
            {
                line_pointer++;
                line_pointer++;
            }
            break;
        case 1:
        case 2:
        case 30:
        case 32:
            for( index = 1; strncmp( line_buffer, "Newsgroups: ", 12 ); index++ )
            {
                strcpy( line_buffer,
                        fl_get_browser_line( fd_peruser->message_browser, 
                                             index ));
            }
            line_pointer = line_buffer + 12;
            strcpy( local_buffer, strtok( line_pointer, "\r" ));

            for( ; ; )
            {
                strcpy( line_buffer, 
                        fl_get_browser_line( fd_peruser->message_browser,
                                             index++ ));

                if ( line_buffer[ 0 ] == '\r' )
                    break;

                if ( isspace( ( unsigned char)line_buffer[ 0 ] ))
                {
                    line_pointer = line_buffer + 1;
                    strcat( local_buffer, strtok( line_pointer, "\r" ));
                }
                else
                    break;
            }
            strcat( local_buffer, "\r\n" );
            line_pointer = local_buffer;
            break;
    }

    if ( caller < 20 || caller >= 30 )
        sprintf( newsgroups, "Newsgroups: %s", strtok( line_pointer, "\r" ));

  

/*
  Create Subject: header line. If the subject line of an article being
  replied to does not start with "Re: ...", then these characters
  are prepended to the article to be composed's "Subject: ..." header line. 
*/

    switch( caller )
    {
/* 
   Article to be composed is a follow-up, or a reply, with or without
   quotes.
*/

        case 1:
        case 2:
        case 21:
        case 22:
        case 30:
        case 32:
            strcpy( subject, get_subject() );
            break;

        default:
            strcpy( subject, "Subject: (no subject)" );
            break;
    }

/*
  Create Message-ID header line.
*/

    strcpy( line_buffer, user_email_address );
    strcpy( user_id, strtok( line_buffer, "@" ));
    strcpy( nntp_domain, strtok( NULL, "\0" ));
    gethostname( line_buffer, BUFFER_SIZE );
    srand( ( unsigned int)getpid() );
    sprintf( message_id, "Message-ID: <%s-%lu-%u-%s@%s>", user_id,
             unix_time, ( int )10000.0 * rand(),
             line_buffer, nntp_domain );

/*
  Create Path: header line.
*/

    sprintf( path, "Path: %s", user_id );

/*
  Create References: header line, if we're composing a follow-up.
  Notice we concatenate a possible multi-line "References: " header
  using a different method than what was used for the "Newsgroup: "
  header line. This is because the article being followed-up may or
  may not have a "References: " header line. 
*/

    if ( caller == 1 || caller == 2 || caller == 21 || caller == 22 || caller >= 30 )
    {
	strcpy( references, "References: " );

	maxline = fl_get_browser_maxline( fd_peruser->message_browser);

	for( index = 1; index <= maxline; index++ )
        {
	    strcpy( line_buffer,
		    fl_get_browser_line( fd_peruser->message_browser, index ));

	    if ( !strncmp( line_buffer, "References: ", 12 ))
            {
		line_pointer = line_buffer + 12;
		strcat( references, strtok( line_pointer, "\r" ));

		for( ; ; )
                {
		    strcpy( line_buffer, 
			    fl_get_browser_line( fd_peruser->message_browser,
                                                 ++index ));

		    if ( line_buffer[ 0 ] == '\r' )
                        break;

		    if ( isspace( ( unsigned char)line_buffer[ 0 ] ))
                    {
			line_pointer = line_buffer + 1;
			strcat( references, " " );
			strcat( references, strtok( line_pointer, "\r" ));
                    }
		    else
                        break;
                }
		strcat( references, " " );
		break;
            }
        }

        index = 1;
        
        while( index++ <= maxline )
        {
            strcpy( line_buffer, fl_get_browser_line( fd_peruser->message_browser,
                                                      index ));

            if ( line_buffer[ 0 ] == '\r' )
                break;
            
            if ( !strncmp( line_buffer, "Message-ID: ", 12 ) ||
                 !strncmp( line_buffer, "Message-Id: ", 12 ))
            {
                line_pointer = line_buffer + 12;
                strcat( references, strtok( line_pointer, "\r" ));

                strcpy( line_buffer, references );
                strcpy( references, wrap_header( line_buffer, ">", 12 ));
                break;
            }
        }
    }
  
/*
  Open ~/.peruser_spool/.peruser_follow_up and write header.
*/

    if (( follow_up_file = fopen( follow_up_path_and_name, "w" )) == NULL )
    {
        perror( name );
        exit( EXIT_FAILURE );
    }
 
        /* To: from address list */

    if ( caller == 23 )
    {
        sprintf( to, "To: %s", address_buffer );
        sprintf( line_buffer, "%s,", to );
        strcpy( to, wrap_header( line_buffer, ",", 4 ));
        to[ strlen( to ) - 1 ] = '\0';
      
        fprintf( follow_up_file, "%s\n%s\n%s\n", message_id, 
                 date, to );
    }
  
    switch( caller )
    {
        case 0:
        case 1:
        case 2:
            fprintf( follow_up_file, "%s\n%s\n%s\n%s\n",
                     path, date, newsgroups, message_id );
            break;
        case 20:
            fprintf( follow_up_file, "%s\n", message_id );
            strcpy( to, "To: (replace with address)" );
            break;
        case 21:
        case 22:
        case 30:
        case 32:
            if ( caller >= 30 )
                fprintf( follow_up_file, "%s\n%s\n",
                         path, newsgroups );

            fprintf( follow_up_file, "%s\n", message_id );

            strcpy( to, "To: " );
            found = no;

            maxline = fl_get_browser_maxline( fd_peruser->message_browser);

            for( index = 1; index <= maxline; index++ )
            {
                strcpy( line_buffer, 
                        fl_get_browser_line( fd_peruser->message_browser, index ));

                if ( line_buffer[ 0 ] == '\r' ) 
                {
                    index = maxline + 1;
                    continue;
                }

                if ( !strncmp( line_buffer, "Reply-To: ", 10 ))
                {
                    found = yes;
                    index = maxline + 1;
                }
            }

            if ( found )
                strcat( to, strtok( line_buffer + 10, "\r" ));
            else
            {
                for( index = 1; strncmp( line_buffer, "@C4From: ", 9 ); index++ )
                {
                    strcpy( line_buffer,
                            fl_get_browser_line( fd_peruser->message_browser, index ));
                }
                strcat( to, strtok( line_buffer + 9, "\r" ));
            }

            break;
    }


    switch( caller )
    {
        case 1:
        case 2:
            fprintf( follow_up_file, "%s\n", references );
            break;
        case 20:
        case 21:
        case 22:
        case 30:
        case 32:
            if ( caller >= 21 )
                fprintf( follow_up_file, "%s\n", references );

            fprintf( follow_up_file, "%s\n%s\n", date, to );
            break;
    }

    fprintf( follow_up_file, "%s\n%s\n%s\n", from, subject,
             caller > 2  ? "X-Mailer: News Peruser 2.37":
             "X-Newsreader: News Peruser 2.37" );

    if ( !strcmp( user_name, "James Bailie" ))
      fprintf( follow_up_file, "X-Greatest-Romantic-Poet: William Butler Yeats\n" );
    
/*
  Insert blank line between header and body of article.
*/

    fputs( "\n", follow_up_file );

/*
  If there are quotes, then add them to .peruser_follow_up.
*/

    if ( caller == 2 || caller == 22 || caller == 32 )
    {
        if (( quote_file = fopen( quote_path_and_name, "r" )) == NULL )
	{
            fflush( follow_up_file );
            fclose( follow_up_file );
            perror( name );
            exit( EXIT_FAILURE );
	}

        for( ; ; )
	{
            fgets( line_buffer, BUFFER_SIZE, quote_file );
            if ( feof( quote_file ))
                break;

            fputs( line_buffer, follow_up_file );
	}
        fclose( quote_file );
    }

/*
  Add ~/.signature, if present.
*/

    sprintf( line_buffer, "%s/.signature", home_directory );
    if (( signature_file = fopen( line_buffer, "r" )) != NULL )
    {
        for( ; ; )
	{
            fgets( line_buffer, BUFFER_SIZE, signature_file );
            if ( feof( signature_file ))
	    {
                fclose( signature_file );
                break;
	    }
            fputs( line_buffer, follow_up_file );
	}
    }
    fflush( follow_up_file );
    fclose( follow_up_file );

/*
  Run editor 
*/

    update_read_file();
    sprintf( line_buffer, "%s %s", editor_command, follow_up_path_and_name );
    system( line_buffer );

/*
  Spool returned article.
*/

    if (( follow_up_file = fopen( follow_up_path_and_name, "r" )) == NULL )
    {
        perror( name );
        exit( EXIT_FAILURE );
    }

    switch( caller )
    {
        case 0:
        case 1:
        case 2:
            sprintf( follow_ups_spool_path_and_name,
                     "%s/.peruser_spool/FOLLOW-UPS", home_directory );
            break;
        case 20:
        case 21:
        case 22:
        case 23:
        case 30:
        case 32:
            sprintf( follow_ups_spool_path_and_name,
                     "%s/.peruser_spool/OUTBOX", home_directory );
            break;
    }
     
    if (( follow_ups_spool_file = fopen( follow_ups_spool_path_and_name, "a" )) == NULL )
    {
        perror( name );
        exit( EXIT_FAILURE );
    }

    if ( caller >= 30 )
    {
        sprintf( line_buffer, "%s/FOLLOW-UPS", spool_path );
        if (( temp_file = fopen( line_buffer, "a" )) == NULL )
	{
            perror( name );
            exit( EXIT_FAILURE );
	}
    }

/*
   If the last line of the follow-up wasn't terminated with a newline
   by the user in their editor, then fgets will return the line, but
   set feof and the loop will break and the line won't get spooled. To
   catch an unterminated last line, a copy of line_buffer is put into
   second_line_buffer, every time a non-blank line is encountered,
   then after the loop, a test compares line_buffer with
   second_line_buffer to see if line_buffer has changed. If so, we
   have an unterminated last line to add to the spool.

   Newlines are stripped-off all lines, and carraige return/newline
   pairs added. NNTP servers expect lines to be terminated this way.
*/

  found = no;

  for( ; ; )
    {
      fgets( line_buffer, BUFFER_SIZE, follow_up_file );
      if ( feof( follow_up_file ))
	break;

      line_pointer = strtok( line_buffer, "\n" );
      if ( line_pointer )
	{
	  if ( caller < 30 )
	    fprintf( follow_ups_spool_file, "%s\r\n", line_pointer );
	  else
	    {
	      fprintf( temp_file, "%s\r\n", line_pointer );

	      if ( found || strncmp( line_buffer, "Path: ", 6 ))
		fprintf( follow_ups_spool_file, "%s\r\n", line_pointer );
	      else 
		{
		  found = yes;
		  continue;
		}
	    }

	  strcpy( second_line_buffer, line_buffer );
	}
      else
	{
	  fputs( "\r\n", follow_ups_spool_file );

	  if ( caller >= 30 )
	    fputs( "\r\n", temp_file );
	}
    }

  if ( strcmp( line_buffer, second_line_buffer ))
    {
      fprintf( follow_ups_spool_file, "%s\r\n", line_buffer );

      if ( caller >= 30 )
	fprintf( temp_file, "%s\r\n", line_buffer );
    }

  fputs( ".\r\n", follow_ups_spool_file );
  fflush( follow_ups_spool_file );
  fclose( follow_ups_spool_file );

  if ( caller >= 30 )
    {
      fputs( ".\r\n", temp_file );
      fflush( temp_file );
      fclose( temp_file );
    }

  fclose( follow_up_file );
  strcpy( line_buffer, follow_up_path_and_name );
  strcat( line_buffer, "*" );
  remove( line_buffer );

  refresh_newsgroups();

  return;
}

void font_callback( FL_OBJECT *obj, long caller )
{
    switch( caller )
    {
        case 0:
            switch( fl_get_menu( fd_peruser->font_size_menu ))
            {
                case 1:
                    font_size = FL_SMALL_SIZE;
                    break;
                case 2:
                    font_size = FL_MEDIUM_SIZE;
                    break;
                case 3:
                    font_size = 16;
                    break;
                case 4:
                    font_size = FL_LARGE_SIZE;
                    break;
            }
            break;
        case 1:
            switch( fl_get_menu( fd_peruser->font_style_menu ))
            {
                case 1:
                    if ( font_style == FL_NORMAL_STYLE )
                        font_style = FL_FIXED_STYLE;
                    else
                        if ( font_style == FL_BOLD_STYLE )
                            font_style = FL_FIXEDBOLD_STYLE;
                    break;
                case 2:
                    if ( font_style == FL_FIXED_STYLE )
                        font_style = FL_NORMAL_STYLE;
                    else 
                        if ( font_style == FL_FIXEDBOLD_STYLE )
                            font_style = FL_BOLD_STYLE;
                    break;
            }
            break;
        case 3:
            switch( font_style )
            {
                case FL_NORMAL_STYLE:
                    font_style = FL_BOLD_STYLE;
                    break;
                case FL_FIXED_STYLE:
                    font_style = FL_FIXEDBOLD_STYLE;
                    break;
                case FL_BOLD_STYLE:
                    font_style = FL_NORMAL_STYLE;
                    break;
                case FL_FIXEDBOLD_STYLE:
                    font_style = FL_FIXED_STYLE;
                    break;
            }
            break;
    }

  fl_set_browser_fontstyle( fd_peruser->message_browser, font_style );
  fl_set_browser_fontsize( fd_peruser->message_browser, font_size );
     
  return;
}

/*
  Change_group_callback allows the user to change the current newsgroup from
  the form fd_article.
*/

void adjust_browser( FL_OBJECT *browser )
{
    int index, max_line, screen_lines, top_line, new_line;

    index = fl_get_browser( browser );
    max_line = fl_get_browser_maxline( browser );
    screen_lines = fl_get_browser_screenlines( browser );
    top_line= fl_get_browser_topline( browser);

    if ( max_line > screen_lines )
    {
        if ( index < top_line )
            if (( top_line - index ) >= screen_lines )
                fl_set_browser_topline( browser, index );
            else
                if (( new_line = top_line - screen_lines ) > 0 )
                    fl_set_browser_topline( browser, new_line );
                else
                    fl_set_browser_topline( browser, 1 );

        if ( index > ( top_line + screen_lines - 1 ))
            if (( index - top_line + screen_lines - 1 ) > screen_lines )
                fl_set_browser_topline( browser, index );
            else
                if (( new_line = top_line + screen_lines ) <= max_line )
                    fl_set_browser_topline( browser, new_line );
    }

  return;
}

void change_group_callback( FL_OBJECT *obj, long caller )
{
  int index;

  index = fl_get_browser( fd_peruser->group_browser );
  
  switch ( caller )
    {
    case 0:
      if ( index == 1 )
	return;
      else
	--index;
      break;
    case 1: 
      if ( index == fl_get_browser_maxline( fd_peruser->group_browser ))
	return;
      else
	++index;
      break;
    }

  fl_select_browser_line( fd_peruser->group_browser, index );
  adjust_browser( fd_peruser->group_browser );
  fl_call_object_callback( fd_peruser->group_browser );

  if ( article_count )
    {
      fl_select_browser_line( fd_peruser->subject_browser, 1 );
      fl_call_object_callback( fd_peruser->subject_browser );
    }
  return;
}

void save_callback( FL_OBJECT *obj, long caller )
{
  int index, article_line;

  enum { no, yes } found;

  if( !article_count || !strcmp( newsgroup, "SAVED" ))
    return;
  
  sprintf( line_buffer, "%s/.peruser_spool/.temp", home_directory );
  if (( temp_file = fopen( line_buffer, "w+" )) == NULL )
    {
      perror( name );
      exit( EXIT_FAILURE );
    }

  sprintf( line_buffer, "%s/.peruser_spool/%s", home_directory, newsgroup );
  if (( spool_file = fopen( line_buffer, "r" )) == NULL )
    {
      perror( name );
      exit( EXIT_FAILURE );
    }

  index = fl_get_browser( fd_peruser->subject_browser );
  index--;

  while( index )
    {
      do
	{
	  fgets( line_buffer, BUFFER_SIZE, spool_file );
	}
      while( strncmp( line_buffer, ".\r\n", 3 ));
      index--;
    }

  found = no;
  do
    {
      fgets( line_buffer, BUFFER_SIZE, spool_file );
      if ( !strncmp( line_buffer, "@C18", 4 ) ||
           !strncmp( line_buffer, "@C22", 4 ))
          continue;
      
      if ( !strncmp( line_buffer, "Subject: ", 9 ) ||
	   !strncmp( line_buffer, "SUBJECT: ", 9 ))
	found = yes;
      fputs( line_buffer, temp_file );
    }
  while( strncmp( line_buffer, ".\r\n", 3 ));

  fclose( spool_file );

  sprintf( line_buffer, "%s/SAVED", spool_path );
  if (( spool_file = fopen( line_buffer, "a" )) == NULL )
    {
      perror( name );
      exit( EXIT_FAILURE );
    }

  rewind( temp_file );

  fgets( line_buffer, BUFFER_SIZE, temp_file );
  fputs( line_buffer, spool_file );

  if ( !found )
    fputs( "Subject: (no subject)\r\n", spool_file );

  do
    {
      fgets( line_buffer, BUFFER_SIZE, temp_file );
      fputs( line_buffer, spool_file );
    }
  while( strncmp( line_buffer, ".\r\n", 3 ));

  fflush( spool_file );
  fclose( spool_file );

  article_line = fl_get_browser_topline( fd_peruser->message_browser);

  update_read_file();

  article_count = 0;
  refresh_newsgroups();

  fl_set_browser_topline( fd_peruser->message_browser, article_line );
  
  return;
}

void references_callback( FL_OBJECT *obj, long caller )
{
  char message_id[ BUFFER_SIZE ], *reference_id;

  enum{ no, yes } last, changed, found;

  int index, limit;

  if( fl_get_browser( fd_peruser->group_browser ) > group_count )
    return;

  limit = fl_get_browser_maxline( fd_peruser->message_browser);
  last = changed = found = no;

  for( index = 1; index <= limit; ++index )
    {
      strcpy( line_buffer, fl_get_browser_line( fd_peruser->message_browser,
                                                index ));

      if ( !strncmp( line_buffer, "\r\n", 1 ))
	break;

      if ( last && strspn( line_buffer, " \t" ))
	strcat( second_line_buffer, 
		line_buffer + strspn( line_buffer, " \t" ));

      last = no;

      if ( !strncmp( line_buffer, "Message-ID: ", 12 ) ||
           !strncmp( line_buffer, "Message-Id: ", 12 ))
	strcpy( message_id, strtok( line_buffer + 12, "\r\n" ));

      if ( !strncmp( line_buffer, "References: ", 12 ))
	{
	  strcpy( second_line_buffer, line_buffer );
	  last = yes;
	  found = yes;
	}
    }

  if ( found )
    {
      reference_id = strtok( second_line_buffer + 12, " \t\r\n" ) +
                     strcspn( second_line_buffer + 12, "<" );

      while( reference_id )
	{
	  found = no;

	  if (( spool_file = fopen( spool_path_and_name, "r" )) == NULL )
	    {
	      perror( name );
	      exit( EXIT_FAILURE );
	    }

	  for( ; ; )
	    {
	      fgets( line_buffer, BUFFER_SIZE, spool_file );
	      if ( feof( spool_file ))
		break;

	      if ( line_buffer[ 0 ] == '\r' )
		do
		  fgets( line_buffer, BUFFER_SIZE, spool_file );
		while( strncmp( line_buffer, ".\r\n", 3 ));

	      if ( !strncmp( line_buffer, "Message-ID: ", 12 ) ||
                   !strncmp( line_buffer, "Message-Id: ", 12 ))
		if ( strstr( line_buffer + 12, reference_id ))
		  {
		    found = yes;
		    break;
		  }
	    }
	  fclose( spool_file );

	  if ( found )
	    {
	      reference_id = strtok( NULL, " \t\r\n" );
	      continue;
	    }

	  sprintf( requests_path, "%s/.peruser_spool/%s.requests",
		   home_directory, newsgroup );
	  if (( requests = fopen( requests_path, "r" )) == NULL )
	    {
	      perror( name );
	      exit( EXIT_FAILURE );
	    }

	  strcpy( temp_requests_path, requests_path );
	  strcat( temp_requests_path, ".temp" );

	  if (( temp_requests = fopen( temp_requests_path, "w" )) == NULL )
	    {
	      perror( name );
	      exit( EXIT_FAILURE );
	    }

	  found = no;

	  for( ; ; )
	    {
	      fgets( line_buffer, BUFFER_SIZE, requests );
	      if ( feof( requests ))
		break;

	      if ( strstr( line_buffer, reference_id ))
		{
		  found = yes;
		  break;
		}

	      fputs( line_buffer, temp_requests );
	    }

	  if ( found )
	    {
	      if ( !( strstr( strstr( line_buffer, ":" ) + 1, 
			      message_id )))
		fputs( line_buffer, temp_requests );
	      else
		changed = yes;

	      for( ; ; )
		{
		  fgets( line_buffer, BUFFER_SIZE, requests );
		  if( feof( requests ))
		    break;
		  fputs( line_buffer, temp_requests );
		}
	    }
	  else
	    {
	      changed = yes;
	      sprintf( line_buffer, "%s:%s\n", 
		       reference_id, message_id );
	      fputs( line_buffer, temp_requests );
	    }
	  
	  fflush( temp_requests );
	  fclose( temp_requests );

	  fclose( requests );
	  remove( requests_path );

	  rename( temp_requests_path, requests_path );

	  reference_id = strtok( NULL, " \t\r\n" );
	}

      if ( changed )
	toggle_italic( fl_get_browser( fd_peruser->subject_browser ));

    }

  return;
}

int check_for_running_netscape()
{
    FILE *ps_stream;
    char local_buffer[ BUFFER_SIZE ];
    char *local_pointer;
    int launched;

    launched = 0;
            
    if ( ! ( ps_stream = popen( "ps", "r" )))
    {
        perror( name );
        exit( EXIT_FAILURE );
    }

    for( ; ; )
    {
        fgets( local_buffer, BUFFER_SIZE, ps_stream );
        if ( feof( ps_stream ))
            break;

        strtok( local_buffer, " " );
        strtok( NULL, " " );
        strtok( NULL, " " );
        strtok( NULL, " " );
        local_pointer = strtok( NULL, " " );
        
        if ( !strncmp( local_pointer, "netscape", 8 ))
        {
            launched = 1;
            break;
        }
    }
    pclose( ps_stream );

    return launched;
}

void launch_url( FL_OBJECT *obj, long caller )
{
    int length, i;
    char local_buffer[ BUFFER_SIZE ];
    static char delimiters[ 12 ] = " \",>)]};\t\r\n";
    
        
    strcpy( line_buffer,
            fl_get_browser_line( fd_peruser->message_browser,
                                 fl_get_browser( fd_peruser->message_browser )));
    
    if ( line_buffer[ 0 ] != '@' )
        return;

    length = strlen( line_buffer );
    for( i = 0; i < length; ++i )
        local_buffer[ i ] = tolower( line_buffer[ i ] );

    local_buffer[ i ] = '\0';
    
    if ( ( line_pointer = strstr( local_buffer, "http://" )) != NULL )
    {
        if ( web_browser[ 0 ] == '\0' )
        {
            fl_show_alert( "Error:", "You have not entered the name of your",
                           "web browser in the configuration form.", 0 );
            return;
        }


        
        if ( !strncmp( web_browser, "netscape", 8 ) &&
             check_for_running_netscape() )
            {
                sprintf( second_line_buffer,
                         "%s -remote 'openURL(%s)' -raise &", web_browser,
                         strtok( line_pointer, delimiters ));
            }
            else
            {
                sprintf( second_line_buffer,
                         "%s %s &", web_browser,
                         strtok( line_pointer, delimiters ));
            }

            length = strlen( second_line_buffer ) - 1;
            if ( second_line_buffer[ length ] == '.' )
                second_line_buffer[ length ] = '\0';

            system( second_line_buffer );
            fl_show_message( "", "Launching URL...", "" );
    }
    else
        if ( ( line_pointer = strstr( line_buffer, "ftp://" )) != NULL )
        {
            if ( ftp_client[ 0 ] == '\0' )
            {
                fl_show_alert( "Error:",
                               "You have not entered the name of your",
                               "ftp client in the configuration form.", 0 );
                return;
            }

            if ( !strncmp( ftp_client, "netscape", 8 ) &&
                 check_for_running_netscape() )
            {
                sprintf( second_line_buffer,
                         "%s -remote 'openURL(%s)' -raise &", ftp_client,
                         strtok( line_pointer, delimiters ));
            }
            else
            {
                sprintf( second_line_buffer,
                         "%s %s &", ftp_client,
                         strtok( line_pointer, delimiters ));
            }

            length = strlen( second_line_buffer ) - 1;
            if ( second_line_buffer[ length ] == '.' )
                second_line_buffer[ length ] = '\0';

            system( second_line_buffer );
            fl_show_message( "", "Launching URL...", "" );
            
        }
            
    return;
}

void edit_callback( FL_OBJECT *obj, long caller )
{
  FILE *temp_read_file;
  char temp_read_path_and_name[ BUFFER_SIZE ];
  int index, maximum, subject_line, article_line;
  enum { no, yes } inbox;


  article_line = 0;
  /*
   Open spool and temporary spool.
*/

  if ( !article_count )
    return;

  if ( !caller )
      if ( !fl_show_question( "Delete message? \nOnce it's gone, it's gone forever!", 0 ))
          return;
      else
          article_line = fl_get_browser_topline( fd_peruser->message_browser );
      
  if ( regex_onscreen )
    fl_clear_browser( fd_regex->regex_browser );

  update_read_file();

  inbox = no;

  index = fl_get_browser( fd_peruser->group_browser ) - group_count;
  switch( index )
    {
    case 1:
      sprintf( follow_ups_spool_path_and_name,
	       "%s/.peruser_spool/INBOX", home_directory );
      sprintf( read_file_path_and_name, "%s.read", 
	       follow_ups_spool_path_and_name );
      inbox = yes;
      break;
    case 2:
      sprintf( follow_ups_spool_path_and_name, "%s/.peruser_spool/OUTBOX", 
	       home_directory );
      break;
    case 3:
      sprintf( follow_ups_spool_path_and_name,
	       "%s/.peruser_spool/FOLLOW-UPS", home_directory );
      break;
    case 4:
      sprintf( follow_ups_spool_path_and_name,
	       "%s/.peruser_spool/SENT-MAIL", home_directory );
      break;
    case 5:
      sprintf( follow_ups_spool_path_and_name,
	       "%s/.peruser_spool/POSTED", home_directory );
      break;
    case 6:
      sprintf( follow_ups_spool_path_and_name,
	       "%s/.peruser_spool/SAVED", home_directory );
      break;

        default:
            strcpy( line_buffer,
                    fl_get_browser_line( fd_peruser->group_browser,
                           fl_get_browser( fd_peruser->group_browser )));

            line_pointer = line_buffer;
            if ( line_pointer[ 0 ] == '@' )
                {
                    line_pointer++;
                    line_pointer++;
                }
            
            sprintf( follow_ups_spool_path_and_name,
                     "%s/%s",
                     spool_path,
                     line_pointer );
            
            break;
    }

  if ( !( follow_ups_spool_file = 
	  fopen( follow_ups_spool_path_and_name, "r" )))
    {
      perror( name );
      exit( EXIT_FAILURE );
    }

  temp_read_file = NULL;

  if ( ( temp_file = fopen( temp_path_and_name, "w" )) == NULL )
    {
      perror( name );
      exit( EXIT_FAILURE );
    }

  if ( inbox )
    {
      if ( !( read_file = fopen( read_file_path_and_name, "r" )))
	{
	  perror( name );
	  exit( EXIT_FAILURE );
	}
      sprintf( temp_read_path_and_name, "%s.temp", read_file_path_and_name );

      if ( ( temp_read_file = fopen( temp_read_path_and_name, "w" )) == NULL )
	{
	  perror( name );
	  exit( EXIT_FAILURE );
	}
    }

/*
   Advance spool.
*/

  
  subject_line = fl_get_browser( fd_peruser->subject_browser );

  index = subject_line;
  index--;
  while( index > 0 )
    {
      do
	{
	  fgets( line_buffer, BUFFER_SIZE, follow_ups_spool_file );
	  fputs( line_buffer, temp_file );
	}
      while( strncmp( line_buffer, ".\r\n", 3 ));

      if ( inbox )
	{
	  fgets( read_status, sizeof read_status, read_file );
	  fputs( read_status, temp_read_file );
	}

      index--;
    }

/*
   If editing follow-up, dump article into ~/.peruser_spool/.peruser_follow_up,
   and launch editor.
*/

  if ( caller )
    {
      if ( !( follow_up_file = fopen( follow_up_path_and_name, "w" )))
	{
	  perror( name );
	  exit( EXIT_FAILURE );
	}

      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, follow_ups_spool_file );
	  if ( !strncmp( line_buffer, ".\r\n", 3 ))
	    {
	      break;
	    }

/* 
   strtok( line_buffer, "\r" ) does not work as expected if
   line_buffer contains "\r\n". It seems to return a pointer to the
   newline. I use an inner loop here to remove "\r\n" from article
   lines. 
*/

	  maximum = strcspn( line_buffer, "\r" );
	  if ( maximum )
	    {
	      for( index = 0; index < maximum; index++ )
		{
		  fputc( line_buffer[ index ], follow_up_file );
		}
	    }
	  fputs( "\n", follow_up_file );
	}
      fflush( follow_up_file );
      fclose( follow_up_file );

      sprintf( line_buffer, "%s %s", editor_command, follow_up_path_and_name );

      system( line_buffer );

/* 
   Spool returned article, checking for a last line that was
   not terminated with a newline by the user.
*/

      if( !( follow_up_file = fopen( follow_up_path_and_name, "r" )))
	{
	  perror( name );
	  exit( EXIT_FAILURE );
	}

      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, follow_up_file );
	  if ( feof( follow_up_file ))
	    {
	      break;
	    }
	  line_pointer = strtok( line_buffer, "\n" );
	  if ( line_pointer )
	    {
	      fprintf( temp_file, "%s\r\n", line_pointer );
	      memcpy( second_line_buffer, line_buffer, BUFFER_SIZE );
	    }
	  else
	    {
	      fputs( "\r\n", temp_file );
	    }
	}
      if ( strncmp( line_buffer, second_line_buffer, BUFFER_SIZE ))
	{
	  fprintf( temp_file, "%s\r\n", line_buffer );
	}
      fputs( ".\r\n", temp_file );
      fclose( follow_up_file );
    }

/*
   If deleting follow-up, skip past the article: don't it copy over to 
   temporary spool.
*/

  else
  {
      do
      {
          fgets( line_buffer, BUFFER_SIZE, follow_ups_spool_file );
      }
      while( strncmp( line_buffer, ".\r\n", 3 ));

      if ( inbox )
          fgets( read_status, sizeof read_status, read_file );
  }

/* 
   Copy over the rest of the spool.
*/

  for( ; ; )
    {
      fgets( line_buffer, BUFFER_SIZE, follow_ups_spool_file );
      if ( feof( follow_ups_spool_file ))
	break;
      fputs( line_buffer, temp_file );
    }

  fflush( temp_file );
  fclose( temp_file );

  fclose( follow_ups_spool_file );
  remove( follow_ups_spool_path_and_name );
  rename( temp_path_and_name, follow_ups_spool_path_and_name );

  if ( inbox )
    {
      for( ; ; )
	{
	  fgets( read_status, sizeof read_status, read_file );
	  if ( feof( read_file ))
	    break;
	  fputs( read_status, temp_read_file );
	}

      fflush( temp_read_file );
      fclose( temp_read_file );
    
      fclose( read_file );
      remove( read_file_path_and_name );
      rename( temp_read_path_and_name, read_file_path_and_name );
    }

/*
   Refresh fd_peruser->subject_browser.
*/

  refresh_newsgroups();

  if ( !caller )
    {
      if ( article_count ) 
	{
	  if ( subject_line > article_count )
	  {
	    for( ; subject_line > article_count; subject_line-- )
	      {
	      }
	  }
	}
    }
  fl_select_browser_line( fd_peruser->subject_browser, subject_line );
  fl_set_browser_topline( fd_peruser->subject_browser, subject_line );
  fl_call_object_callback( fd_peruser->subject_browser );

  if ( caller )
    fl_set_browser_topline( fd_peruser->message_browser, article_line );

  return;
}

void delete_callback( FL_OBJECT *obj, long caller )
{
    int index;

    index = fl_get_menu( fd_peruser->delete_menu );

    switch( index )
    {
        case 1:
            edit_callback( obj, 0 );
            break;
        case 2:
            if ( !fl_show_question( "Delete ALL messages? \nOnce they're gone, they're gone forever!", 0 ))
                return;
            
            update_read_file();

            sprintf( line_buffer, "%s/%s", spool_path, newsgroup );
            truncate( line_buffer, 0 );
            strcat( line_buffer, ".read" );
            truncate( line_buffer, 0 );

            refresh_newsgroups();
            break;
    }

    return;
}
