 /*
    News Peruser Copyright (c) 1996, 1997 James Bailie
    ==================================================================

    News Peruser is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2, or (at
    your option) any later version.

    News Peruser is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    Although News Peruser is licensed under the Free Software
    Foundation's GNU General Public License, Peruser is not produced
    by, nor is it endorsed by the Free Software Foundation. The Free
    Software Foundation is not responsible for developing,
    distributing, or supporting Peruser in any way. Anyone may place
    software they own the copyright to, under the GNU General Public
    License.

    The GNU General Public License is included in the News Peruser 
    distribution archive in a file called COPYING. If you do
    not have a copy of the license, you can download one from
    ftp://prep.ai.mit.edu, or you can write to the Free Software
    Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    =====================================================================
 */

#include<signal.h>
#include<ctype.h>
#include<sys/types.h>
#include<sys/wait.h>
#include<sys/stat.h>

#include "peruser_includes.h"
#include "regex.h"
#include "uudeview.h"

void update_read_file();
void refresh_newsgroups();

typedef struct
{
    char *sort;
    char *true;
} address_array;

unsigned search_file( struct re_pattern_buffer, unsigned index );

void sync_form( void )
 {
   FL_Coord x,y;
   unsigned int keymask;
   fl_get_mouse(&x,&y,&keymask);
   return;
 }

 void toggle_italic( int line )
 {
   int index;

   strcpy( line_buffer, 
           fl_get_browser_line( fd_peruser->subject_browser,
                                line ));

   strncpy( second_line_buffer, line_buffer, 
            index = strcspn( line_buffer, " " ));
   second_line_buffer[ index ] = '\0';      

   if ( ( line_pointer = strstr( second_line_buffer, "@i" )) != NULL )
     strcpy( line_pointer, strstr( line_buffer, " " ));
   else
     {
       strcat( second_line_buffer, "@i" );
       strcat( second_line_buffer, strstr( line_buffer, " " ));
     }

   fl_replace_browser_line( fd_peruser->subject_browser, line,
                            second_line_buffer );

   sync_form();
   return;
 }

 void regex_callback( FL_OBJECT *obj, long caller )
 {
   static const char *regex = "";
   static struct re_pattern_buffer buffer;

   const char *result;
   char *local_pointer;
   char local_buffer[ BUFFER_SIZE ];

   unsigned int index, index2, index3;
   unsigned int matches, max;

   static enum{ current, all } group;

   switch( caller )
     {

 /*
   called by fd_regex->done_button
 */

     case 0:
       fl_hide_form( fd_regex->regex );
       regex_onscreen = no;
       fl_clear_browser( fd_regex->regex_browser );
       fl_clear_browser( fd_regex->feedback_browser );
       break;

 /*
   called by fd_regex->regex_input
 */

     case 1:
       regex = fl_get_input( fd_regex->regex_input );
       break;

 /*
         called by fd_regex->regex_browser.
 */

     case 2:
       strcpy( local_buffer,
               fl_get_browser_line( fd_regex->regex_browser,
                                    fl_get_browser( fd_regex->regex_browser )));

       local_pointer = local_buffer;
       if ( local_buffer[ 0 ] == '@' )
         {
           local_pointer++;
           local_pointer++;
           fl_replace_browser_line( fd_regex->regex_browser, 
                                    fl_get_browser( fd_regex->regex_browser ),
                                    local_pointer );
         }	  

 /*
   Parse the string completely before calling any callbacks, as
   they also call strtok. 
 */

       strtok( local_pointer, "(" );

       index = atoi( strtok( NULL, ":" ));
       index2 = atoi( strtok( NULL, ":" ));
       index3 = atoi( strtok( NULL, ")" ));

       fl_select_browser_line( fd_peruser->group_browser, index );
       fl_set_browser_topline( fd_peruser->group_browser, index );
       fl_call_object_callback( fd_peruser->group_browser );

       fl_select_browser_line( fd_peruser->subject_browser, index2 );
       fl_set_browser_topline( fd_peruser->subject_browser, index2 );
       fl_call_object_callback( fd_peruser->subject_browser );

       strcpy( line_buffer, 
               fl_get_browser_line( fd_peruser->message_browser, 2 ));


       if ( !strncmp( line_buffer, "@-", 2 ) && index3 >= 2 )
         index3++;

       fl_set_browser_topline( fd_peruser->message_browser, index3 );
       fl_select_browser_line( fd_peruser->message_browser, index3 );

       break;

 /*
   called by fd_regex->all_button
 */
     case 3:
       group = all;
       break;

 /*
   called by fd_peruser->regex_button
 */

     case 4:
       group = all;

       fl_set_button( fd_regex->current_button, 0 );
       fl_set_button( fd_regex->all_button, 1 );
       fl_set_input_return( fd_regex->regex_input, FL_RETURN_ALWAYS );

       fl_show_form( fd_regex->regex, FL_PLACE_MOUSE, FL_FULLBORDER, 
                     "Regex Search" );
       regex_onscreen = yes;
       break;

 /*
   called by fd_regex->search_button
 */

     case 5:
       fl_trigger_object( fd_regex->regex_input );
       if ( regex[ 0 ] == '\0' ) 
         break;

       re_syntax_options = RE_NO_BK_VBAR | RE_NO_BK_PARENS;
       
       buffer.translate = NULL;
       buffer.fastmap = ( char *)1;
       buffer.buffer = NULL;
       buffer.allocated = 0;

       if ( ( result = re_compile_pattern( regex, strlen( regex ), &buffer )) )
         {
           fl_show_alert( "Error:", result, "", 0 );
           break;
         }

       fl_clear_browser( fd_regex->regex_browser );
       fl_clear_browser( fd_regex->feedback_browser );
       fl_add_browser_line( fd_regex->feedback_browser, "@b@iSeaching..." );
       sync_form();
       fl_freeze_form( fd_regex->regex );

       matches = 0;
       if ( group )
         {
             max = fl_get_browser_maxline( fd_peruser->group_browser );
             for( index = 1; index <= max; index++ )
             {
               strcpy( line_buffer, 
                       fl_get_browser_line( fd_peruser->group_browser, index ));

               line_pointer = line_buffer;
               if ( line_buffer[ 0 ] == '@' )
                 {
                   line_pointer++;
                   line_pointer++;
                 }

               strcpy( second_line_buffer, line_pointer );

               matches += search_file( buffer, index );
             }
         }
       else
         {
           if ( !fl_get_browser( fd_peruser->group_browser ))
             {
               fl_clear_browser( fd_regex->feedback_browser );
               fl_show_message( "", "No newsgroup is selected", "" );
               fl_unfreeze_form( fd_regex->regex );
               break;
             }

           strcpy( second_line_buffer, newsgroup );
           matches = search_file( buffer, 
                                  fl_get_browser( fd_peruser->group_browser ));
         }
       sprintf( line_buffer, "@b%u Matches:", matches );
       fl_replace_browser_line( fd_regex->feedback_browser, 1, line_buffer );
       fl_unfreeze_form( fd_regex->regex );
       break;

 /*
   called by fd_regex->current_button
 */

     case 6:
       group = current;
       break;
     case 7:
       strcpy( line_buffer, user_email_address );
       regex = strtok( line_buffer, "@" );
       fl_call_object_callback( fd_regex->search_button );
       break;
     }

   return;
 }

 unsigned search_file( struct re_pattern_buffer buffer, unsigned index )
 {
   FILE *current;

   char result_string[ BUFFER_SIZE ];

   int return_code;

   unsigned message_count;
   unsigned line_count;
   unsigned matches;


   sprintf( line_buffer, "%s/.peruser_spool/%s", home_directory, 
            second_line_buffer );

   if (( current = fopen( line_buffer, "r" )) == NULL )
     if ( errno == 2 ) return( 0 );
     else
       {
         perror( name );
         exit( EXIT_FAILURE );
       }

   matches = 0;
   line_count = 0;
   message_count = 1;
   for( ; ; )
     {
       fgets( line_buffer, BUFFER_SIZE, current );

       if ( feof( current )) break;

       ++line_count;

       if ( !strncmp( line_buffer, ".\r\n", 3 ))
         {
           ++message_count;
           line_count = 0;
           continue;
         }

       return_code = re_search( &buffer, line_buffer, strlen( line_buffer ),
                                0, strlen( line_buffer ), NULL );


       if ( return_code == -1 )
         continue;
       else
         if ( return_code == -2 )
           {
             fl_show_alert( "An internal error occurred", 
                            "in the regex library", "", 0 );
             break;
           }

       matches++;

       sprintf( result_string, "@b%s (%u:%u:%u):  %s", 
                second_line_buffer, index, message_count, line_count, 
                ( line_buffer[ 0 ] != '\r' ) ? strtok( line_buffer, "\r"  ) : "" );

       fl_add_browser_line( fd_regex->regex_browser, 
                            result_string );
     }

   fclose( current );
   return( matches );
 }

int compare( const void *a, const void *b )
 {
   return( strcmp( *( char **)a, *( char **)b ));
 }
     

int address_array_compare( const void *a, const void *b )
{
    char *a_pointer, *b_pointer;

    a_pointer = (( address_array *)a)->sort;
    b_pointer = (( address_array *)b)->sort;
    
    if ( a_pointer[ 0 ] == '"' )
        a_pointer = a_pointer + 1;

    if ( b_pointer[ 0 ] == '"' )
        b_pointer = b_pointer + 1;
        
    return strcmp( a_pointer, b_pointer );
}



void address_callback( FL_OBJECT *obj, long caller )
 {
     address_array *array;
         
     int index, second_index, maxline, length;
   
   FILE *address_file;

   static int changed;

   switch( caller )
   {
       case 0:
           break;
       case 1:
           fl_add_browser_line( fd_address->address_browser,
                                fl_get_input( fd_address->address_input ));
           changed = 1;
           break;
       case 2:
           fl_set_input( fd_address->address_input,
              fl_get_browser_line( fd_address->address_browser,
                             fl_get_browser( fd_address->address_browser )));
           break;
     case 3:
       fl_delete_browser_line( fd_address->address_browser,
                               fl_get_browser( fd_address->address_browser ));
       fl_set_input( fd_address->address_input, "" );
       changed = 1;
       break;
     case 4:
       fl_hide_form( fd_address->address );
       break;
     case 5:
     case 6:
       if ( caller == 5 && ( fl_get_browser_maxline( fd_address->recipient_browser ) == 0 ))
         {
           fl_show_alert( "You have not selected an address!", "", "", 1 );
           break;
         }

       if ( changed )
         {
           sprintf( line_buffer, "%s/.peruser-addresses", home_directory );
           if (( address_file = fopen( line_buffer, "w" )) == NULL )
             {
               perror( name );
               exit( EXIT_FAILURE );
             }

           maxline = fl_get_browser_maxline( fd_address->address_browser );
           index = 1;
           while( index <= maxline )
             {
               fputs( fl_get_browser_line( fd_address->address_browser, index ),
                      address_file );
               fputc( '\n', address_file );
               ++index;
             }

           fflush( address_file );
           fclose( address_file );
         }

       if ( caller == 5 )
         {
           address_buffer[ 0 ] = '\0';
           maxline = fl_get_browser_maxline( fd_address->recipient_browser );

           for( index = 1; index <= maxline; index++ )
             {
               strcat( address_buffer, 
                       fl_get_browser_line( fd_address->recipient_browser, index ));

               if ( index < maxline ) 
                 strcat( address_buffer, ", " );
             }

           fl_hide_form( fd_address->address );
           run_editor( obj, 23 );
         }
       else
         fl_hide_form( fd_address->address );

       break;
     case 7:
         case 8:
         case 20:
       sprintf( line_buffer, "%s/.peruser-addresses", home_directory );
       fl_clear_browser( fd_address->address_browser );
       fl_load_browser( fd_address->address_browser, line_buffer );
       changed = 0;

       if ( caller == 20 )
         {
             fl_set_choices_shortcut( "y", "n", "c" );
             index = fl_show_choice( "Get address from address list?", "", "", 3, 
                                   "yes", "no",  "cancel", 2 );

           if ( index == 2 )
             {
               run_editor( obj, caller );
               break;
             }
           else
             if ( index == 3 )
                 break;
         }

       fl_set_input( fd_address->address_input, "" );
       fl_clear_browser( fd_address->recipient_browser );
       fl_show_form( fd_address->address, FL_PLACE_GEOMETRY, FL_FULLBORDER,
                     "Address List" );

       break;
     case 30:
       maxline = fl_get_browser_maxline( fd_address->address_browser );

       array = calloc( maxline, sizeof *array );
       
       for( index = 0; index < maxline; index++ )
         {
           strcpy( line_buffer, fl_get_browser_line( fd_address->address_browser,
                                                     index + 1 ));

           array[ index ].true = malloc( length = strlen( line_buffer ) + 1 );
           array[ index ].sort = malloc( length );
           
           strcpy( array[ index ].true, line_buffer );
         }

       for( index = 0; index < maxline; index++ )
       {
           length = strlen( array[ index ].true );
           
           for( second_index = 0; second_index < length; second_index++ )
               array[ index ].sort[ second_index ] =
                   tolower( array[ index ].true[ second_index ] );
       }
               
       qsort( array, maxline, sizeof *array, address_array_compare );

       for( index = 0; index < maxline; index++ )
         {
           fl_replace_browser_line( fd_address->address_browser, index + 1, 
                                    array[ index ].true );
           free( array[ index ].true );
           free( array[ index ].sort );
         }

       free( array );
       changed = 1;
       break;
     case 40:
       fl_delete_browser_line( fd_address->recipient_browser,
                         fl_get_browser( fd_address->recipient_browser ));
       break;
       case 50:
           fl_add_browser_line( fd_address->recipient_browser,
                  fl_get_browser_line( fd_address->address_browser,
                             fl_get_browser( fd_address->address_browser )));
           break;
       case 60:
           fl_replace_browser_line( fd_address->address_browser,
                                    fl_get_browser( fd_address->address_browser ),
                                    fl_get_input( fd_address->address_input ));
           break;
   }

   return;
 }

 void decode_callback( FL_OBJECT *obj, long caller )
 {
   uulist *item;
   int response, index, maxline;

   if ( article_count == 0 )
     return;

   if ( caller == 120 )
     {
       fl_hide_form( fd_decode->decode );
       return;
     }

   if (( response = UUInitialize()) != UURET_OK )
     {
       fl_show_alert( "Error initializing uulib:", line_buffer, 
                      second_line_buffer, 1 );
       return;
     }

   sprintf( line_buffer, "%s/", home_directory );
   UUSetOption( UUOPT_SAVEPATH, 0, line_buffer );

   if ( caller == 200 )
     {
       if (( temp_file = fopen( temp_path_and_name, "w" )) == NULL )
         {
           perror( name );
           exit( EXIT_FAILURE );
         }

       maxline = fl_get_browser_maxline( fd_peruser->message_browser );
       for( index = 1; index <= maxline; index++ )
         {
           fputs( fl_get_browser_line( fd_peruser->message_browser, index ),
                  temp_file );
           fputs( "\r\n", temp_file );
         }

       fflush( temp_file );
       fclose( temp_file );
       strcpy( line_buffer, temp_path_and_name );
     }
   else
     sprintf( line_buffer, "%s/%s", spool_path, newsgroup );

   if (( response = UULoadFile( line_buffer, NULL, 0 )) != UURET_OK )
     {
       strcpy( second_line_buffer, ( response == UURET_IOERR ) ?
               strerror( UUGetOption( UUOPT_ERRNO, NULL, NULL, 0 )) :
               UUstrerror( response ));
       fl_show_alert( "Could not load file:", line_buffer, second_line_buffer,
                      1 );
     }
   else
     {
       fl_clear_browser( fd_decode->message_browser );
       fl_show_form( fd_decode->decode, FL_PLACE_MOUSE, FL_FULLBORDER, 
                     "decode" );

       item = NULL;
       for( index = 0; item != NULL; index++ )
         { 
           if (( item = UUGetFileListItem( index )) == NULL )
             fl_add_browser_line( fd_decode->message_browser,
                                  "@C4No more encoded data found." );
           else
             if ( !( item->state & UUFILE_OK ))
               {
                 sprintf( second_line_buffer, "@C1ERROR: %s",
                          ( item->filename ) ? item->filename : 
                          "(no filename)" );

                 fl_add_browser_line( fd_decode->message_browser,
                                      second_line_buffer );

                 if ( item->state & UUFILE_MISPART )
                   { 
                     strcpy( second_line_buffer, 
                             "\tEncoded data is incomplete." );
                     fl_add_browser_line( fd_decode->message_browser,
                                          second_line_buffer);
                   }

                 if ( item->state & UUFILE_NOBEGIN )
                   {
                     strcpy( second_line_buffer, 
                             "\tNo \"begin\" line detected." );
                     fl_add_browser_line( fd_decode->message_browser,
                                          second_line_buffer);
                   }

                 if ( item->state & UUFILE_NOEND )
                   {
                     strcpy( second_line_buffer, 
                             "\tNo \"end\" line detected." );
                     fl_add_browser_line( fd_decode->message_browser,
                                          second_line_buffer);
                   }

                 if ( item->state & UUFILE_NODATA )
                   {
                     strcpy( second_line_buffer, 
                             "\tNo valid encoded data." );
                     fl_add_browser_line( fd_decode->message_browser,
                                          second_line_buffer);
                   }

                 if ( item->state & UUFILE_ERROR )
                   {
                     strcpy( second_line_buffer, 
                             "\tAn I/O error occurred." );
                     fl_add_browser_line( fd_decode->message_browser,
                                          second_line_buffer);
                   }

                 if ( item->state & UUFILE_DECODED )
                   {
                     strcpy( second_line_buffer, 
                             "\tFile has already been decoded." );
                     fl_add_browser_line( fd_decode->message_browser,
                                          second_line_buffer);
                   }
               }
             else
               if (( response = UUDecodeFile( item, NULL )) != UURET_OK )
                 {
                   sprintf( line_buffer, "\t%s", ( response == UURET_IOERR ) ?
                            strerror( UUGetOption( UUOPT_ERRNO, NULL, 
                                                   NULL, 0 )) :
                            UUstrerror( response ));

                   sprintf( second_line_buffer, "@C1ERROR: %s",
                            ( item->filename == NULL ) ?
                           "oops!" : item->filename );

                   fl_add_browser_line( fd_decode->message_browser, 
                                        second_line_buffer );
                   fl_add_browser_line( fd_decode->message_browser,
                                        line_buffer );
                 }
               else
                 {
                   sprintf( line_buffer, "@C2Successfully decoded: %s",
                            item->filename );
                   fl_add_browser_line( fd_decode->message_browser,
                                        line_buffer );
                 }
         }
     }

   UUCleanUp();
   return;
 }

 void execute_callback( FL_OBJECT *obj, long caller )
 {
   strcpy( line_buffer, fl_get_browser_line( fd_decode->message_browser,
                               fl_get_browser( fd_decode->message_browser )));

   if( strncmp( line_buffer, "@C2Successfully decoded: ", 25 ))
     return;

   sprintf( second_line_buffer, "viewer '%s'", strchr( line_buffer, ':' ) + 2 );
   system( second_line_buffer );

   return;
 }

 char *wrap_header( char *header, char *delimiter, int padding )
 {
   int end, span, pad;
   char *local_pointer;

   second_line_buffer[ 0 ] = '\0';
   local_pointer = header;
   end = 0;

   while( *local_pointer )
     {
       span = strcspn( local_pointer, delimiter );
       strncat( second_line_buffer, local_pointer, ++span );

       second_line_buffer[ span + end ] = '\n';
       second_line_buffer[ span + end + 1 ] = '\0';

       pad = padding;
       while( pad-- )
           strcat( second_line_buffer, " " );

       end = strlen( second_line_buffer );

       local_pointer = local_pointer + span;
       local_pointer = local_pointer + strspn( local_pointer, " \t\r\n" );
     }

   second_line_buffer[ strlen( second_line_buffer ) - ++padding ] = '\0';

   return( second_line_buffer );
 }

 void get_data( int fd, void *data )
 {
     line_buffer[ 0 ] = '\0';
     fgets( line_buffer, BUFFER_SIZE, client );
     if ( feof( client ) || ( line_buffer[ 0 ] == '\0' ))
     {
         fl_trigger_object( fd_client->client_browser );
         return;
     }

     if (( line_buffer[ 0 ] == '\n' ) || ( line_buffer[ 0 ] == '\r' ))
         fl_addto_browser( fd_client->client_browser, " " );
     else
         fl_addto_browser( fd_client->client_browser,
                           strtok( line_buffer, "\r\n" ));

     if ( fl_get_browser_maxline( fd_client->client_browser ) > 500 )
         fl_delete_browser_line( fd_client->client_browser, 1 );

     sync_form();

     return;
 }

 void run_client_program()
 {
     int index;
     enum{ no, yes }kill_request;
     pid_t pid;
     char **args[ 10 ];
     
     sprintf( line_buffer, "%s/.peruser_spool/client_pipe",
              home_directory );
     if ( mkfifo( line_buffer, 0700 ))
     {
         if ( errno != EEXIST )
             {
                 perror( name );
                 exit( EXIT_FAILURE );
             }
     }

     switch( pid = fork() )
     {
         case -1:
             perror( name );
             exit( EXIT_FAILURE );
             break;
         case 0:
             line_pointer = strtok( second_line_buffer, " " );
             if (( args[ 0 ] = malloc( strlen( line_pointer ) + 1 )) == NULL )
             {
                 perror( name );
                 exit( EXIT_FAILURE );
             }
             strcpy( ( char *)args[ 0 ], line_pointer );

             index = 1;            
             while(( line_pointer = strtok( NULL, " " )) != NULL )
             {
                 if (( args[ index ] = malloc( strlen( line_pointer ) + 1 )) ==
                     NULL )
                     {
                         perror( name );
                         exit( EXIT_FAILURE );
                     }

                 strcpy( ( char *)args[ index++ ], line_pointer );
             }

             args[ index ] = NULL;

             execvp( ( char *)args[ 0 ], ( char **)args );

             perror( name );
             exit( EXIT_FAILURE );
             break;
         default:
             if ( !client_onscreen )
             {
                 fl_show_form( fd_client->client, FL_PLACE_FREE, FL_FULLBORDER,
                               "Client Output" );
                 client_onscreen = yes;
             }
             else
                 fl_raise_form( fd_client->client );

             fl_addto_browser( fd_client->client_browser, "STARTING CLIENT..." );

             sync_form();

             sprintf( line_buffer, "%s/.peruser_spool/client_pipe",
                      home_directory );
             if (( client_fd = open( line_buffer, O_RDONLY | O_NONBLOCK ))
                 == -1 )
             {
                 perror( name );
                 fl_addto_browser( fd_client->client_browser,
                                   "ERROR: Unable to open client pipe." );
                 return;
             }

             if (( client = fdopen( client_fd, "r" )) == NULL )
             {
                 perror( name );
                 exit( EXIT_FAILURE );
             }

             
             kill_request = no;

             fl_add_io_callback( client_fd, FL_READ, get_data, NULL );
             fl_hide_object( fd_client->done_button );
             fl_deactivate_form( fd_peruser->peruser );

             if( fl_do_forms() != fd_client->client_browser )
                 kill_request = yes;

             if ( kill_request )
             {
                 if ( kill( pid, SIGTERM ))
                     if ( kill( pid, SIGKILL ))
                     {
                         perror( name );
                         fl_addto_browser( fd_client->client_browser,
                                           "ERROR: Unable to kill client." );
                         return;
                     }

                 fl_addto_browser( fd_client->client_browser,
                                   "CLIENT KILLED." );
             }
             else
                 fl_addto_browser( fd_client->client_browser,
                                   "CLIENT FINISHED." );
             
             fl_remove_io_callback( client_fd, FL_READ, get_data );
             fclose( client );
             close( client_fd );

             waitpid( pid, NULL, WNOHANG | WUNTRACED );
             fl_show_object( fd_client->done_button );
             fl_activate_form( fd_peruser->peruser );
     }

     return;
 }

 void client_callback( FL_OBJECT *obj, long caller )
 {
     if ( caller )
     {
         fl_show_form( fd_client->client, FL_PLACE_MOUSE,
                       FL_FULLBORDER, "Client Output" );
         client_onscreen = yes;
     }
     else
     {
         fl_hide_form( fd_client->client );
         client_onscreen = no;
     }

     return;
 }

 void highlight_url( char *original_line )
 {
     int length, i;
     char local_buffer[ BUFFER_SIZE ];

     length = strlen( original_line );
     
     for( i = 0; i < length; ++i )
         local_buffer[ i ] = tolower( original_line[ i ] );
     
     if ( strstr( local_buffer, "http://" ) ||
          strstr( local_buffer, "ftp://" ))
     {
         sprintf( second_line_buffer, "@C10%s", original_line );
         strcpy( original_line, second_line_buffer );
     }

     return;
 }

 void save_font_state()
 {
     sprintf( line_buffer, "%s/.peruser_spool/.fonts", home_directory );
     if (( temp_file = fopen( line_buffer, "w" )) == NULL )
     {
         perror( name );
         exit( EXIT_FAILURE );
     }

     fprintf( temp_file, "%d\n%d\n", font_size, font_style );
     fclose( temp_file );

     return;
 }

 void get_font_state()
 {
     sprintf( line_buffer, "%s/.peruser_spool/.fonts", home_directory );
     if (( temp_file = fopen( line_buffer, "r" )) == NULL )
     {
         if ( errno != 2 )
         {
             perror( name );
             exit( EXIT_FAILURE );
         }
         font_size = 16;
         font_style = FL_BOLD_STYLE;
     }
     else
     {
         fgets( line_buffer, BUFFER_SIZE, temp_file );
         font_size = atoi( line_buffer );
         fgets( line_buffer, BUFFER_SIZE, temp_file );
         font_style = atoi( line_buffer );

         fclose( temp_file );
     }

     switch( font_style )
     {
         case FL_BOLD_STYLE:
         case FL_FIXEDBOLD_STYLE:
             fl_set_button( fd_peruser->bold_button, 1 );
             break;
             
         case FL_NORMAL_STYLE:
         case FL_FIXED_STYLE:
             fl_set_button( fd_peruser->bold_button, 0 );
             break;
     }
                  
     font_callback( fd_peruser->font_size_menu, 100 );

     return;
 }

void folder_callback( FL_OBJECT *obj, long caller )
{
    char folder[ BUFFER_SIZE ], regex[ BUFFER_SIZE ];
    const char *input_line;
    int index, line, choice;
        
    switch( choice = fl_get_menu( fd_peruser->folder_menu ))
    {
        case 1:
            input_line = fl_show_input( "Folder Name:", "" );
            if ( input_line == NULL || input_line[ 0 ] == '\0' )
                {
                    fl_show_alert( "ERROR:",
                                   "name not entered.",
                                   "Folder not created", 0 );
                    return;
                }
            
            strcpy( line_buffer, input_line );

            line = strlen( line_buffer );
            for( index = 0; index < line; index++ )
                if ( isspace( line_buffer[ index ] ) ||
                     ispunct( line_buffer[ index ] ))
                {
                    fl_show_alert( "ERROR:",
                                   "You are limited to alphanumerics for folder names",
                                   "Folder not created.",
                                   0 );
                    return;
                }

            input_line = fl_show_input( "Regexp filter:", "" );
            if ( input_line == NULL || input_line[ 0 ] == '\0' )
            {
                fl_show_alert( "ERROR:",
                               "No regexp entered.",
                               "Folder not created", 0 );
                return;
            }
            
            sprintf( second_line_buffer, "%s/.user-folders",
                     spool_path );
            if (( spool_file = fopen( second_line_buffer, "a" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }

            fprintf( spool_file, "%s;%s\n", line_buffer, input_line );
            fclose( spool_file );
            
            update_read_file();
            refresh_newsgroups();
            break;

        case 2:
            if ( !( line = fl_get_browser( fd_peruser->group_browser )) ||
                 line < ( group_count + 7 ))
            {
                fl_show_alert( "ERROR:",
                               "A user-defined folder is not currently selected",
                               "", 0 );
                return;
            }
            
            strcpy( line_buffer, fl_get_browser_line( fd_peruser->group_browser,
                                                      line ));

            line_pointer = line_buffer;
            if ( line_buffer[ 0 ] == '@' )
            {
                line_pointer++;
                line_pointer++;
            }
            
            sprintf( second_line_buffer, "%s/.user-folders",
                     spool_path );
            if (( newsrc = fopen( second_line_buffer, "r" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }

            strcat( second_line_buffer, ".tmp" );
            if (( temp_file = fopen( second_line_buffer, "w" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }
            
            for( ; ; )
            {
                fgets( second_line_buffer, BUFFER_SIZE, newsrc );
                if ( feof( newsrc ))
                    break;

                if ( !strncmp( line_pointer, second_line_buffer,
                               strcspn( second_line_buffer, ";" )))
                    continue;
                
                fputs( second_line_buffer, temp_file );
            }
            fclose( newsrc );
            fclose( temp_file );

            sprintf( second_line_buffer, "rm -f %s/%s*",
                     spool_path, line_buffer );
            system( second_line_buffer );

            sprintf( line_buffer, "%s/.user-folders",
                     spool_path );
            sprintf( second_line_buffer, "%s.tmp", line_buffer );
            rename( second_line_buffer, line_buffer );

            update_read_file();
            refresh_newsgroups();
            
            break;
        case 3:
        case 4:
            if ( !( line = fl_get_browser( fd_peruser->group_browser )) ||
                 line < ( group_count + 7 ))
            {
                fl_show_alert( "ERROR:",
                               "A user-defined folder is not currently selected",
                               "", 0 );
                return;
            }
                        
            strcpy( line_buffer, fl_get_browser_line( fd_peruser->group_browser,
                                                      line ));

            line_pointer = line_buffer;
            if ( line_buffer[ 0 ] == '@' )
            {
                line_pointer++;
                line_pointer++;
            }

            strcpy( folder, line_buffer );

            sprintf( second_line_buffer, "%s/.user-folders", spool_path );
            if (( newsrc = fopen( second_line_buffer, "r" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }

            for( ; ; )
            {
                fgets( second_line_buffer, BUFFER_SIZE, newsrc );
                if ( feof( newsrc ))
                    break;

                if ( !strncmp( line_pointer, second_line_buffer,
                               strcspn( second_line_buffer, ";" )))
                {
                    if ( choice == 3 )
                    {
                        strcpy( line_buffer, strtok( second_line_buffer, ";" ));
                        strcpy( regex, strtok( NULL, "\n" ));
                        break;
                    }
                    else
                    {
                        strcpy( line_buffer,
                                strstr( second_line_buffer, ";" ) + 1 );
                        break;
                    }
                    
                }
            }
            fclose( newsrc );

            if ( choice == 3 )
            {
                input_line = fl_show_input( "Folder name:", line_buffer );
                if ( input_line == NULL || input_line[ 0 ] == '\0' )
                {
                    fl_show_alert( "ERROR:", "Name not entered.",
                                   "Folder name not changed.", 0 );
                    return;
                }
                strcpy( line_buffer, input_line );

                line = strlen( line_buffer );
                for( index = 0; index < line; index++ )
                    if ( isspace( line_buffer[ index ] ) ||
                         ispunct( line_buffer[ index ] ))
                    {
                        fl_show_alert( "ERROR:",
                                       "You are limited to alphanumerics for folder names",
                                       "Folder name not changed.",
                                       0 );
                        return;
                    }
            }
            else
            {
                
                input_line = fl_show_input( "Regexp filter:", line_buffer );
                if ( input_line == NULL || input_line[ 0 ] == '\0' )
                {
                    fl_show_alert( "ERROR:", "Regexp not entered",
                                   "Filter not changed.", 0 );
                    return;
                }
            }
            
            sprintf( second_line_buffer, "%s/.user-folders",
                     spool_path );
            if (( newsrc = fopen( second_line_buffer, "r" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }

            strcat( second_line_buffer, ".tmp" );
            if (( temp_file = fopen( second_line_buffer, "w" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }
            
            for( ; ; )
            {
                fgets( second_line_buffer, BUFFER_SIZE, newsrc );
                if ( feof( newsrc ))
                    break;

                if ( !strncmp( folder, second_line_buffer,
                               strcspn( second_line_buffer, ";" )))
                {
                    if ( choice == 3 )
                        {
                            sprintf( second_line_buffer, "%s;%s\n", input_line,
                                     regex );
                            fputs( second_line_buffer, temp_file );
                        }
                    else
                        {
                            sprintf( line_buffer, "%s;%s\n",
                                     folder, input_line );
                            fputs( line_buffer, temp_file );
                        }
                    
                    continue;
                }

                fputs( second_line_buffer, temp_file );
            }
            fclose( newsrc );
            fclose( temp_file );

            sprintf( line_buffer, "%s/.user-folders",
                     spool_path );
            sprintf( second_line_buffer, "%s.tmp", line_buffer );
            rename( second_line_buffer, line_buffer );

            update_read_file();
            refresh_newsgroups();

            break;
    }

    return;
}

void get_geometry( int args, char **arg_array[] )
{
    int count;
    enum{ no, yes }no_size;
    
    FL_COORD x, y, w, h;

    w = h = 0;
    
    while( --args )
        if ( !strncmp( ( char *)arg_array[ args ], "-geom", 5 ))
            break;

    if ( args )
    {
        strcpy( line_buffer, ( char *)arg_array[ args + 1 ] );
        strcat( line_buffer, "\n" );
        
        if ( ( line_pointer = strchr( line_buffer, 'x' )) != NULL )
            {
                count = strcspn( line_buffer, "x" );
                strncpy( second_line_buffer, line_buffer, count );
                second_line_buffer[ count ] = '\0';

                w = atoi( second_line_buffer );

                count = strcspn( ++line_pointer, "+-\n" );
                strncpy( second_line_buffer, line_pointer, count );
                second_line_buffer[ count ] = '\0';

                h = atoi( second_line_buffer );

                line_pointer = line_pointer + count;
                no_size = no;
                
                if ( *line_pointer == '\n' )
                    {
                        fl_set_form_size( fd_peruser->peruser,
                                          w, h );
                        return;
                    }
            }
        else
        {
            no_size = yes;
            line_pointer = line_buffer;
        }
            
    count = strcspn( line_pointer + 1, "-+" ) + 1;
    strncpy( second_line_buffer, line_pointer, count );
    second_line_buffer[ count ] = '\0';
            
    x = atoi( second_line_buffer );

    line_pointer = line_pointer + count;
    count = strcspn( line_pointer, "\n" );
    strncpy( second_line_buffer, line_pointer, count );
            
    y = atoi( second_line_buffer );
        
    if ( no_size )
        fl_set_form_position( fd_peruser->peruser, x, y );
    else
        fl_set_form_geometry( fd_peruser->peruser, x, y, w, h );
    }

    return;
}


