/* 
   Packer, Copyright (C) 1995 James Bailie
   ==================================================================

   Packer is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Packer is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   Although Packer is licensed under the Free Software Foundation's
   GNU General Public License, Hear is not produced by, nor is it
   endorsed by the Free Software Foundation. The Free Software
   Foundation is not responsible for developing, distributing, or
   supporting Packer in any way. Anyone may place software they own
   the copyright to, under the GNU General Public License.

   The GNU General Public License is included in the News Peruser
   distribution archive in a file called COPYING. If you do not have a
   copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   =====================================================================
*/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<signal.h>
#include<unistd.h>
#include<sys/wait.h>
#include<sys/types.h>
#include<sys/stat.h>

#define BUFFER_SIZE 256

pid_t pid;

char spool_path_and_name[ BUFFER_SIZE ];
char temp_path_and_name[ BUFFER_SIZE ];
char read_path_and_name[ BUFFER_SIZE ];
char temp_read_path_and_name[ BUFFER_SIZE ];

FILE *newsrc;
FILE *temporary_file;
FILE *spool_file;
FILE *read_file;
FILE *temp_read_file;
FILE *output, *client;


void signal_handler( int signal )
{
    if ( spool_file )
        fclose( spool_file );
    
    if ( read_file )
        fclose( read_file );

    if ( temporary_file )
        {
            fclose( temporary_file );
            remove( temp_path_and_name );
        }
    
    if ( temp_read_file )
    {
        fclose( temp_read_file );
        remove( temp_read_path_and_name );
    }

    if ( newsrc )
        fclose( newsrc );

    if ( client )
        fclose( client );
            
    if ( pid )
        {
            if ( kill( pid, SIGINT ))
                kill( pid, SIGKILL );

            waitpid( pid, NULL, WNOHANG | WUNTRACED );
        }

    fclose( output );
    
    exit( signal );
}

int main( int argc, char **argv[] )
{
  
    char *name = "packer";

    char line_buffer[ BUFFER_SIZE ];
    char home_directory[ BUFFER_SIZE ];
    char newsgroup[ BUFFER_SIZE ];
    char read_status[ 3 ];

/*
  Open .newsrc
*/

    strcpy( home_directory, getenv( "HOME" ));

    if ( strstr( ( char *)argv[ 0 ], "cpacker" ))
    {
        sprintf( line_buffer, "%s/.peruser_spool/client_pipe",
                 home_directory );
        if (( output = fopen( line_buffer, "w" )) == NULL )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }
    }
    else
        output = stdout;

    sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );
    if (( newsrc = fopen( line_buffer, "r" )) == NULL )
    {
        fclose( output );
        perror( name );
        exit( EXIT_FAILURE );
    }

/*
  Get the name of a newsgroup.
*/

    for( ; ; )
    {
        fgets( line_buffer, sizeof line_buffer, newsrc );
        if ( feof( newsrc ))
	{
            break;
	}
        if ( !strncmp( line_buffer, "\n", 1 ))
	{
            continue;
	}
        strcpy( newsgroup, strtok( line_buffer, ":" ));

/*
  Open the group's spool file.
*/

        sprintf( spool_path_and_name, "%s/.peruser_spool/%s", home_directory,
                 newsgroup );
        if (( spool_file = fopen( spool_path_and_name, "r" )) == NULL )
	{
            perror( name );
            continue;
	}

/*
  Open the group's read file.
*/

        sprintf( read_path_and_name, "%s/.peruser_spool/%s.read", home_directory,
                 newsgroup );
        if (( read_file = fopen( read_path_and_name, "r" )) == NULL )
	{
            perror( name );
            fclose( spool_file );
            continue;
	}

/*
  Open the temporary file to hole the new read file.
*/

        sprintf( temp_read_path_and_name, "%s/.peruser_spool/read.temp", 
                 home_directory );
        if (( temp_read_file = fopen( temp_read_path_and_name, "w" )) == NULL )
	{
            fclose( output );
            perror( name );
            exit( EXIT_FAILURE );
	}

/*
  Open the temporary file to hold the packed spool file.
*/

        sprintf( temp_path_and_name, "%s/.peruser_spool/.temp", home_directory );
        if (( temporary_file = fopen( temp_path_and_name, "w" )) == NULL )
	{
            fclose( output );
            perror( name );
            exit( EXIT_FAILURE );
	}

        fprintf( output, "packer: packing %s...\n", newsgroup );

/*
  Get the read status of each article, if unread copy over to temporary
  file, otherwise advance spool file past article.
*/

        for ( ; ; )
	{
            fgets( read_status, sizeof line_buffer, read_file );
            if ( feof( read_file ))
	    {
                break;
	    }
            if ( !strncmp( read_status, "u", 1 ))
	    {
                do
		{
                    fgets( line_buffer, sizeof line_buffer, spool_file );
                    fputs( line_buffer, temporary_file );
		}
                while( strncmp( line_buffer, ".\r\n", 3 ));
                fputs( "u\n", temp_read_file );
	    }
            else
	    {
                do
		{
                    fgets( line_buffer, sizeof line_buffer, spool_file );
		}
                while( strncmp( line_buffer, ".\r\n", 3 ));
	    }
	}
        fflush( temporary_file );
        fclose( temporary_file );
        fflush( temp_read_file );
        fclose( temp_read_file );
        fclose( spool_file );
        fclose( read_file );
        remove( spool_path_and_name );
        rename( temp_path_and_name, spool_path_and_name );
        remove( read_path_and_name );
        rename( temp_read_path_and_name, read_path_and_name );
    }

    fclose( newsrc );
    fprintf( output, "packer: finished.\n" );
    fflush( output );
  
    if ( output == stdout )
        execlp( "threader", "threader", NULL );
    else
    {
        sprintf( line_buffer, "%s/.peruser_spool/threader_pipe",
                 getenv( "HOME" ));
        if ( mkfifo( line_buffer, 0700 ))
        {
            if ( errno != EEXIST )
            {
                fclose( output );
                perror( name );
                exit( EXIT_FAILURE );
            }
        }

        pid = 0;
        switch( pid = fork() )
        {
            case -1:
                fclose( output );
                perror( name );
                exit( EXIT_FAILURE );
                break;
            case 0:
                execlp( "cthreader", "cthreader", NULL );

                fclose( output );
                perror( name );
                exit( EXIT_FAILURE );
                break;
            default:
                if (( client = fopen( line_buffer, "r" )) == NULL )
                {
                    fclose( output );
                    perror( name );
                    exit( EXIT_FAILURE );
                }

                for( ; ; )
                {
                    fgets( line_buffer, BUFFER_SIZE, client );
                    if ( feof( client ))
                        break;
                  
                    fputs( line_buffer, output );
                    fflush( output );
                }

                fclose( client );
                fclose( output );

                waitpid( pid, NULL, WNOHANG ^ WUNTRACED );
                pid = 0;
                
                exit( EXIT_SUCCESS );
                break;
        }
    }
  
    fclose( output );
    perror( name );
    exit( EXIT_FAILURE );
}

