/* 
   News Peruser, Copyright (C) 1996, 1997 James Bailie
   ==================================================================

   
   Forms Library, Copyright (c) 1996-1997 T.C. Zhao and Mark Overmars.
   ALL RIGHTS RESERVED.
   
   News Peruser uses the Forms Library to create its GUI.
   Although News Peruser is freeware, the Forms Library IS NOT!
   
   The Forms Library is, however, freely distributable and freely
   usable in non-commercial projects. Commercial distribution of
   software that uses the Forms Library requires separate licensing
   from the authors.

   In order to compile the News Peruser package you will need to obtain
   the Forms Library distribution for your system from:
   
   ftp://einstein.phys.uwm.edu/pub/xforms

   For more information on the Forms Library, see:

   http://bragg.phys.uwm.edu/xforms

   IF YOU DISTRIBUTE NEWS PERUSER YOU MUST NOT INCLUDE ANY MATERIAL
   FROM THE FORMS LIBRARY, OR YOU WILL BE VIOLATING THE COPYRIGHT
   OF THE FORMS LIBRARY.

   ==================================================================
*/

#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<unistd.h>
#include<string.h>
#include<time.h>

#include "forms.h"

void launch_url( FL_OBJECT *obj, long caller );
void get_geometry( int argc, char **argv );
void update_read_file();
void save_font_state();
int refresh_newsgroups();

/* 
   Form structure-type definitions generated by fdesign, the graphical
   form creator that is part of the Forms Library package. Forms are
   accessed via structures of pointers, each element pointing to an
   object on the form, such as a button, a browser, etc.
*/

#include "peruser_form.h"
#include "quoter_form.h"
#include "configure_form.h"
#include "hear_form.h"
#include "regex_form.h"
#include "address_form.h"
#include "decode_form.h"
#include "newsrc_form.h"
#include "client_form.h"

/*
   Macros to define size of variables that hold lines of text.
*/

#define BUFFER_SIZE 1024

/*
   Macros used to give meaningful names to elements of the read array, which
   keeps track of whether each article in the currently selected newsgroup
   are "read" or "unread".
*/

#define UNREAD 0
#define READ 1

/*
  Used by perror.
*/

const char *name = "peruser";

/*
   Create instances of each form structure-type definition.
*/

FD_peruser *fd_peruser;
FD_quoter *fd_quoter;
FD_configure *fd_configure;
FD_hear *fd_hear;
FD_regex *fd_regex;
FD_address *fd_address;
FD_decode *fd_decode;
FD_newsrc *fd_newsrc;
FD_client *fd_client;


/*
   Various file pointers.
*/

FILE *signature_file;
FILE *spool_file;
FILE *config_file;
FILE *quote_file;
FILE *follow_up_file;
FILE *follow_ups_spool_file;
FILE *read_file;
FILE *temp_file;
FILE *users_newsrc;
FILE *newsrc;
FILE *requests;
FILE *temp_requests;
int client_fd;
FILE *client;

/*
   Variables that hold lines of text.
*/

char line_buffer[ BUFFER_SIZE ];
char second_line_buffer[ BUFFER_SIZE ];
char address_buffer[ 256 ];

/*
   A variable that holds a line read from a newsgroup's read file,
   which stores the read array for a particular newsgroup when it is
   not the currently selected newsgroup. The read file is a text file,
   each line of which consists of two characters, which can be either,
   "u\n", to indicate an unread article, or "r\n" to indicate a read
   article. The third character of read_status is filled by the NULL
   terminator, of course.
*/

char read_status[ 3 ];

/*
   Variables that hold path and filename information.
*/

char home_directory[ BUFFER_SIZE ];
char spool_path[ BUFFER_SIZE ];
char spool_path_and_name[ BUFFER_SIZE ];
char follow_up_path_and_name[ BUFFER_SIZE ];
char quote_path_and_name[ BUFFER_SIZE ];
char config_path_and_name[ BUFFER_SIZE ];
char follow_ups_spool_path_and_name[ BUFFER_SIZE ];
char read_file_path_and_name[ BUFFER_SIZE ];
char temp_path_and_name[ BUFFER_SIZE ];
char requests_path[ BUFFER_SIZE ];
char temp_requests_path[ BUFFER_SIZE ];

/*
   A variable that holds the name of the currently selected newsgroup.
*/

char newsgroup[ BUFFER_SIZE ];

/*
   Variables that hold the user's configuration information.
*/

char user_name[ BUFFER_SIZE ];
char user_email_address[ BUFFER_SIZE ];
char editor_command[ BUFFER_SIZE ];
char server_address[ BUFFER_SIZE ];
char pop_server_address[ BUFFER_SIZE ];
char pop_password[ BUFFER_SIZE ];
char smtp_server_address[ BUFFER_SIZE ];
char nntp_username[ BUFFER_SIZE ];
char nntp_password[ BUFFER_SIZE ];
char web_browser[ BUFFER_SIZE ];
char ftp_client[ BUFFER_SIZE ];

/*
   A pointer that is used in the parsing and manipulation of variables that
   hold lines of text.
*/

char *line_pointer;

/*
   Three integers that respectively keep track of: the number of
   articles in the currently selected newsgroup, the number of
   newsgroups listed in the user's .newsrc file, the number of
   articles that have been exported from the spool files into separate
   text files. This number is used to create the filename of the
   exported article.
*/

int article_count;
int group_count;
int exported;
int authenticating;

int font_size, font_style;


/* 
   regex_onscreen is used to keep track of whether form fd_regex->regex is
   visible or not. Used by edit_callback to decide if it needs to
   update fd_regex->feedback_browser, by calling it's callback, to
   reflect changes to the folders' spool files.
*/

enum { no, yes, both } configuration, reply_by_mail, folder, user_folder,
    deleted, regex_onscreen, client_onscreen, hear_onscreen;

/*
   A pointer which is later pointed to a malloc-ed space that will
   hold the currently selected newsgroup's read array. Each element
   can be either READ or UNREAD, as #define-ed above.
*/

char *read_array;

/* 
   fl_root Refers to root window of the forms library. Used only in
   main to set icon. 
*/

extern Window fl_root;

#include "peruser_icon.xpm"

void get_font_state();

int main( int argc, char **argv[] )
{
  FL_OBJECT *obj;
  Pixmap icon_pixmap, icon_pixmap_mask;
  unsigned width, height;
  
  strcpy( home_directory, getenv( "HOME" ));
  
/*
  Initialize forms
*/

  fl_initialize( &argc, (char **)argv, "Peruser", 0, 0 );

  fd_peruser   = create_form_peruser();
  fd_quoter    = create_form_quoter();
  fd_configure = create_form_configure();
  fd_hear      = create_form_hear();
  fd_regex     = create_form_regex();
  fd_address   = create_form_address();
  fd_decode    = create_form_decode();
  fd_newsrc    = create_form_newsrc();
  fd_client    = create_form_client();
  
  fl_set_goodies_font( FL_NORMAL_STYLE, FL_MEDIUM_SIZE );


/* 
   Set icons for forms.
*/

  icon_pixmap = fl_create_from_pixmapdata( fl_root, peruser_icon_xpm,
					   &width, &height,
					   &icon_pixmap_mask, 0, 0, 0 );
  fl_set_form_icon( fd_peruser->peruser, icon_pixmap, icon_pixmap_mask );
  fl_set_form_icon( fd_quoter->quoter, icon_pixmap, icon_pixmap_mask );
  fl_set_form_icon( fd_configure->configure, icon_pixmap, icon_pixmap_mask );
  fl_set_form_icon( fd_hear->hear, icon_pixmap, icon_pixmap_mask );
  fl_set_form_icon( fd_regex->regex, icon_pixmap, icon_pixmap_mask );
  fl_set_form_icon( fd_address->address, icon_pixmap, icon_pixmap_mask );
  fl_set_form_icon( fd_decode->decode, icon_pixmap, icon_pixmap_mask );
  fl_set_form_icon( fd_newsrc->newsrc, icon_pixmap, icon_pixmap_mask );
  fl_set_form_icon( fd_client->client, icon_pixmap, icon_pixmap_mask );

    
/*
  Initializes fd_regex->feedback_browser.
*/

  fl_set_browser_fontsize( fd_regex->feedback_browser, FL_SMALL_SIZE );  

/*
   This initializes the counter on form fd_hear.
*/

  fl_set_counter_step( fd_hear->article_counter, 1, 10 );
  fl_set_counter_bounds( fd_hear->article_counter, 0, 1000 );
  fl_set_counter_precision( fd_hear->article_counter, 0 );
  fl_set_counter_value( fd_hear->article_counter, 0 );

/*
   These are the browsers on form fd_peruser.
*/

  fl_set_browser_fontsize( fd_peruser->group_count_browser, FL_SMALL_SIZE );
  fl_set_browser_fontsize( fd_peruser->article_count_browser, FL_SMALL_SIZE );
  fl_set_browser_fontsize( fd_peruser->group_browser, FL_MEDIUM_SIZE );
  fl_set_browser_fontsize( fd_peruser->subject_browser, FL_MEDIUM_SIZE );

  get_font_state();

/*
  These are the objects on form fd_regex.
*/

  fl_set_browser_fontsize( fd_regex->regex_browser, FL_MEDIUM_SIZE );
  fl_set_browser_fontstyle( fd_regex->regex_browser, FL_NORMAL_STYLE );
  fl_set_button( fd_regex->current_button, 1 );

/*
   These are the browsers on form fd_quoter.
*/


  fl_set_browser_fontsize( fd_quoter->quoter_source_browser, FL_MEDIUM_SIZE );
  fl_set_browser_fontstyle( fd_quoter->quoter_source_browser,
			    FL_NORMAL_STYLE );
  fl_set_browser_fontsize( fd_quoter->quoter_destination_browser,
			   FL_MEDIUM_SIZE );
  fl_set_browser_fontstyle( fd_quoter->quoter_destination_browser,
			    FL_NORMAL_STYLE );

/*
  Objects on form fd_address.
*/

  fl_set_browser_fontsize( fd_address->address_browser, FL_MEDIUM_SIZE );
  fl_set_browser_fontstyle( fd_address->address_browser, FL_NORMAL_STYLE );

  fl_set_browser_fontsize( fd_address->recipient_browser, FL_MEDIUM_SIZE );
  fl_set_browser_fontstyle( fd_address->recipient_browser, FL_NORMAL_STYLE );

  fl_set_browser_dblclick_callback( fd_address->address_browser,
                                    address_callback, ( long)50 );
  
  
/*
  fd_decode
*/

  fl_set_browser_fontsize( fd_decode->message_browser, FL_SMALL_SIZE );
  fl_set_browser_fontstyle( fd_decode->message_browser, FL_NORMAL_STYLE );

/*
    fd_newsrc
*/

  fl_set_browser_fontsize( fd_newsrc->newsrc_browser, 12 );
  fl_set_browser_fontsize( fd_newsrc->list_browser, 12 );
  fl_set_browser_fontstyle( fd_newsrc->newsrc_browser, FL_NORMAL_STYLE );
  fl_set_browser_fontstyle( fd_newsrc->list_browser, FL_NORMAL_STYLE );

  fl_set_browser_dblclick_callback( fd_newsrc->newsrc_browser,
                                    edit_newsrc_callback,
                                    ( long )1 );
  fl_set_browser_dblclick_callback( fd_newsrc->list_browser,
                                    edit_newsrc_callback,
                                    ( long )2 );
  
/*
  fd_client
*/

  fl_set_browser_fontsize( fd_client->client_browser, FL_MEDIUM_SIZE );
  fl_set_browser_fontstyle( fd_client->client_browser, FL_BOLD_STYLE );

/*
   Assign values to global variables.
*/

  sprintf( spool_path, "%s/.peruser_spool", home_directory );

  sprintf( follow_up_path_and_name, "%s/.peruser_follow_up", spool_path );
  sprintf( quote_path_and_name, "%s/.peruser_quotes", spool_path );

  sprintf( follow_ups_spool_path_and_name, "%s/FOLLOW-UPS", spool_path );
  sprintf( temp_path_and_name, "%s/.temp", spool_path );

  sprintf( config_path_and_name, "%s/.peruser_config", home_directory );

  sprintf( temp_requests_path, "%s_temp", requests_path );

  user_name[ 0 ] = '\0';
  user_email_address[ 0 ] = '\0';
  editor_command[ 0 ] = '\0';
  server_address[ 0 ] = '\0';
    

/*
  Create popup menus
*/

  fl_setpup_fontsize( 12 );
  fl_setpup_fontstyle( FL_NORMAL_STYLE );

  fl_set_menu( fd_peruser->select_menu, "current|references|thread|all" );
  fl_set_menu_item_shortcut( fd_peruser->select_menu, 1, "c" );
  fl_set_menu_item_shortcut( fd_peruser->select_menu, 2, "r" );
  fl_set_menu_item_shortcut( fd_peruser->select_menu, 3, "t" );
  fl_set_menu_item_shortcut( fd_peruser->select_menu, 4, "a" );

  fl_set_menu( fd_peruser->mark_menu, "group|all" );
  fl_set_menu_item_shortcut( fd_peruser->mark_menu, 1, "g" );
  fl_set_menu_item_shortcut( fd_peruser->mark_menu, 2, "a" );

  fl_set_menu( fd_peruser->clear_menu, "group|all" );
  fl_set_menu_item_shortcut( fd_peruser->clear_menu, 1, "g" );
  fl_set_menu_item_shortcut( fd_peruser->clear_menu, 2, "a" );

  fl_set_menu( fd_peruser->delete_menu, "current|all" );
  fl_set_menu_item_shortcut( fd_peruser->delete_menu, 1, "c" );
  fl_set_menu_item_shortcut( fd_peruser->delete_menu, 2, "a" );

  fl_set_menu( fd_peruser->compose_menu, "article|mail" );
  fl_set_menu_item_shortcut( fd_peruser->compose_menu, 1, "a" );
  fl_set_menu_item_shortcut( fd_peruser->compose_menu, 2, "m" );

  fl_set_menu( fd_peruser->folder_menu, "create|delete|edit name|edit regex" );
  fl_set_menu_item_shortcut( fd_peruser->folder_menu, 1, "c" );
  fl_set_menu_item_shortcut( fd_peruser->folder_menu, 2, "d" );
  fl_set_menu_item_shortcut( fd_peruser->folder_menu, 3, "n" );
  fl_set_menu_item_shortcut( fd_peruser->folder_menu, 4, "r" );
  
  fl_set_menu( fd_newsrc->sort_menu, "newsrc|list" );

  fl_set_menu( fd_peruser->font_size_menu, "small|medium1|medium2|large" );
  fl_set_menu_item_shortcut( fd_peruser->font_size_menu, 1, "s" );
  fl_set_menu_item_shortcut( fd_peruser->font_size_menu, 2, "1" );
  fl_set_menu_item_shortcut( fd_peruser->font_size_menu, 3, "2" );
  fl_set_menu_item_shortcut( fd_peruser->font_size_menu, 4, "l" );

  fl_set_menu( fd_peruser->font_style_menu, "fixed|proportional" );
  fl_set_menu_item_shortcut( fd_peruser->font_style_menu, 1, "f" );
  fl_set_menu_item_shortcut( fd_peruser->font_style_menu, 2, "p" );
  
/*
   Show the main form.
*/

  get_geometry( argc, ( char **)argv );
  fl_show_form( fd_peruser->peruser, FL_PLACE_POSITION,
                FL_FULLBORDER, "Peruser" );
  fl_set_app_mainform( fd_peruser->peruser );

/*
  Set permissions for file creation.
*/

  umask( 0077 );

/*
  Make sure the spool directory exists.
*/

  if ( mkdir( spool_path, 0755 ))
    {
      if ( errno != EEXIST )
	{
	  perror( name );
	  exit( EXIT_FAILURE );
	}
    }

/*
   Retrieve the user's configuration info.
*/

  if (( config_file = fopen( config_path_and_name, "r" )) != NULL )
    {
      fgets( line_buffer, sizeof user_name, config_file );
      strcpy( user_name, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof user_email_address, config_file );
      strcpy( user_email_address, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof editor_command, config_file );
      strcpy( editor_command, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof server_address, config_file );
      strcpy( server_address, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof pop_server_address, config_file );
      strcpy( pop_server_address, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof pop_password, config_file );
      strcpy( pop_password, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof smtp_server_address, config_file );
      strcpy( smtp_server_address, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof line_buffer, config_file );
      authenticating = atoi( line_buffer );

      fgets( line_buffer, sizeof nntp_username, config_file );
      strcpy( nntp_username, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof nntp_password, config_file );
      strcpy( nntp_password, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof web_browser, config_file );
      strcpy( web_browser, strtok( line_buffer, "\n" ));

      fgets( line_buffer, sizeof ftp_client, config_file );
      strcpy( ftp_client, strtok( line_buffer, "\n" ));
      
      fclose( config_file );
    }
  else
    if ( errno == 2 )
      { 
	strcpy( editor_command, 
		"xterm -fg LightGray -bg DarkBlue -cr Yellow -fn 10x20 -geometry 80x32 -e vi" );

	strcpy( pop_server_address, "none" );
	strcpy( pop_password, "none" );
	strcpy( smtp_server_address, "none" );
	strcpy( nntp_username, "none" );
	strcpy( nntp_password, "none" );
        strcpy( web_browser, "netscape" );
        strcpy( ftp_client, "netscape" );
        
	fl_call_object_callback( fd_peruser->configure_button );

        configuration = no;
        while( !configuration )
            fl_check_forms();
      }
    else
      {
	perror( name );
	exit( EXIT_FAILURE );
      }
  
/*
   Load newsgroups from ~/.peruser-newsrc
*/

  if ( refresh_newsgroups() )
      {
          users_newsrc = NULL;
          fl_call_object_callback( fd_peruser->edit_newsrc_button );
        
          while( !users_newsrc )
              fl_check_forms();
      }

  
  exported = 0;
  folder = no;
  read_array = NULL;
  
/*
   Main Loop.
*/

  do
    {
      obj = fl_do_forms();
    }
  while( obj != fd_peruser->quit_button );

  update_read_file();
  save_font_state();
  fl_hide_form( fd_peruser->peruser );
  
  return( 0 );
}
