/*
   News Peruser Copyright (c) 1996, 1997 James Bailie
   ==================================================================
   
   News Peruser is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   News Peruser is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   Although News Peruser is licensed under the Free Software
   Foundation's GNU General Public License, Peruser is not produced
   by, nor is it endorsed by the Free Software Foundation. The Free
   Software Foundation is not responsible for developing,
   distributing, or supporting Peruser in any way. Anyone may place
   software they own the copyright to, under the GNU General Public
   License.

   The GNU General Public License is included in the News Peruser 
   distribution archive in a file called COPYING. If you do
   not have a copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   =====================================================================
*/

#include "peruser_includes.h"
#include "regex.h"

int compare( const void *a, const void *b );
void adjust_browser( FL_OBJECT *browser );
void toggle_italic( int line );
void run_client_program();
void references_callback( FL_OBJECT *obj, long caller );
void sync_form();

void update_read_file( void )
{
  int index;

/* 
   Make sure there are articles in the currently selected newsgroup,
   and that the current newsgroup is not a folder. 
*/

  if ( !article_count || folder || read_array == NULL )
    return;
      
/*
   Open the read file for writing, and truncate it to zero length.
*/

  if (( read_file = fopen( read_file_path_and_name, "w" )) == NULL )
    {
      perror( name );
      exit( EXIT_FAILURE );
    }

/*
   Write the read file.
*/

  for( index = 0; index < article_count; index++ )
    {
      if ( read_array[ index ] == READ )
	fputs( "r\n", read_file );
      else
	fputs( "u\n", read_file );
    }
  fflush( read_file );
  fclose( read_file );

/*
   This is the ONLY place where read arrays are freed.
*/

  if ( read_array == NULL )
    {
      fprintf( stderr, "peruser: attempt to free NULL read_array\n" );
      exit( EXIT_FAILURE );
    }
  free( read_array );
  read_array = NULL;

  return;
}

int refresh_newsgroups( void )
{
  int  newsrc_line, group_line, subject_line, index;
  char folders[ 10 ][ 13 ] = { "OUTBOX", "@bOUTBOX", "FOLLOW-UPS", 
			      "@bFOLLOW-UPS", "SENT-MAIL", "@bSENT-MAIL",
			      "POSTED", "@bPOSTED", "SAVED", "@bSAVED" };

  fl_freeze_form( fd_peruser->peruser );

  group_line = fl_get_browser( fd_peruser->group_browser );
  subject_line = fl_get_browser( fd_peruser->subject_browser );

/*
   Find and open the user's .peruser-newsrc.
*/

  sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );
  if (( users_newsrc = fopen( line_buffer, "r+" )) == NULL )
    {
      if ( errno == 2 )
      {
          fl_unfreeze_form( fd_peruser->peruser );
          return( 1 );
      }
      
      else perror( name );
      exit( EXIT_FAILURE );
    }
  
  second_line_buffer[ 0 ] = getc( users_newsrc );
  if ( feof( users_newsrc ))
  {
      fl_unfreeze_form( fd_peruser->peruser );
      return( 1 );
  }
  else
    ungetc( second_line_buffer[ 0 ], users_newsrc );


    strcat( line_buffer, "-working" );
    if (( newsrc = fopen( line_buffer, "w+" )) == NULL )
      {
	perror( name );
	exit( EXIT_FAILURE );
      }

    for( ; ; )
      {
	fgets( line_buffer, BUFFER_SIZE, users_newsrc );
	if ( feof( users_newsrc ))
	  break;
	if ( line_buffer[ 0 ] == '\n' )
	  continue;
	fputs( line_buffer, newsrc );
      }
    fclose( users_newsrc );
    fputs( "INBOX:\n", newsrc );
    rewind( newsrc );

/*
  Check for any lines in ~/.peruser-newsrc without colons.
*/

  for( newsrc_line = 1; ; newsrc_line++ )
    {
      fgets( line_buffer, BUFFER_SIZE, newsrc );
      if ( feof( newsrc ))
	break;
      if ( !strchr( line_buffer, ':' ))
	{
	  sprintf( second_line_buffer, 
		   "Line %d in ~/.peruser-newsrc does not contain a colon\n",
		   newsrc_line );
	  fl_show_alert( second_line_buffer, "",
			 "Peruser cannot parse the newsgroup name\n", 1 );
	  fl_select_browser_line( fd_newsrc->newsrc_browser, newsrc_line );
	  fl_call_object_callback( fd_peruser->edit_button );
	}
    }
  rewind( newsrc );

/*
   Loop through the user's .peruser-newsrc.
*/

  fl_clear_browser( fd_peruser->group_browser );
  group_count = 0;

  if ( newsrc )
    {
      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, newsrc );
	  if( feof( newsrc ))
	    break;
	  if ( !strncmp( line_buffer, "\n", 1 ))
	    continue;

/* 
   Open the just-read newsgroup's read file.
*/
    
	  strcpy( newsgroup, strtok( line_buffer, ":" ));
	  sprintf( read_file_path_and_name, "%s/%s.read", 
		   spool_path, newsgroup );
	  strcpy( line_buffer, newsgroup );

	  if (( read_file = fopen( read_file_path_and_name, "r" )) == NULL )
	    {
	      if ( errno != 2 )
		{
		  perror( name );
		  exit( EXIT_FAILURE );
		}
	    }

/*
   Prefix a boldface attribute to newsgroup's name if the newsgroup's
   read file has at least one instance of UNREAD. That is to say, if
   there's at least one unread article in the newsgroup, its name
   will be displayed in boldface in fd_peruser->group_browser. 
*/

	  else
	    {
	      for( ; ; )
		{
		  fgets( read_status, sizeof read_status, read_file );
		  if ( feof( read_file ))
		    break;
		  if ( !strcmp( read_status, "u\n" ))
		    {
		      sprintf( line_buffer, "@b%s", newsgroup );
		      break;
		    }
		}
	      fclose( read_file );
	    }

/*
   Add the newsgroup's name to fd_peruser->group_browser, increment group
   count.
*/

	  fl_add_browser_line( fd_peruser->group_browser, line_buffer );
	  group_count++;
	}
    }
  fclose( newsrc );
  group_count--;
  
/*
   Check for articles in standard folders. If there are, add folders in 
   boldface, otherwise use normal face style.
*/

  for( index = 0; index < 10; index += 2 )
    {
      sprintf( line_buffer, "%s/%s", spool_path, folders[ index ] );
      if (( spool_file = fopen( line_buffer, "r" )) == NULL )
	if ( errno == 2 )
	  fl_add_browser_line( fd_peruser->group_browser, folders[ index ] );
	else
	  {
	    perror( name );
	    exit( EXIT_FAILURE );
	  }
      else
	{
	  fgetc( spool_file );

	  if ( feof( spool_file ))
	    fl_add_browser_line( fd_peruser->group_browser, folders[ index ] );
	  else
	    fl_add_browser_line( fd_peruser->group_browser, 
				 folders[ index + 1 ] );
	  
	  fclose( spool_file );
	}
    }

      /*
        Check for user-defined folders.
      */

  sprintf( line_buffer, "%s/.user-folders", spool_path );
  if (( newsrc = fopen( line_buffer, "r" )) == NULL )
  {
      if ( errno != 2 )
      {
          perror( name );
          exit( EXIT_FAILURE );
      }
  }
  else
  {
      for( ; ; )
      {
          fgets( line_buffer, BUFFER_SIZE, newsrc );
          if ( feof( newsrc ))
              break;
          
          sprintf( second_line_buffer, "%s/%s.read", spool_path,
                   strcpy( newsgroup, strtok( line_buffer, ";" )));
          if (( read_file = fopen( second_line_buffer, "r" )) == NULL )
              if ( errno == 2 )
                  fl_add_browser_line( fd_peruser->group_browser,
                                       newsgroup );
              else
              {
                  perror( name );
                  exit( EXIT_FAILURE );
              }
          else
          {
              strcpy( line_buffer, newsgroup );
              
	      for( ; ; )
		{
		  fgets( read_status, sizeof read_status, read_file );
		  if ( feof( read_file ))
		    break;
		  if ( !strcmp( read_status, "u\n" ))
		    {
		      sprintf( line_buffer, "@b%s", newsgroup );
		      break;
		    }
		}
	      fclose( read_file );

              fl_add_browser_line( fd_peruser->group_browser,
                                   line_buffer );
          }
      }
  }
  
  
/*
   Add group count to the form and show it.
*/

  fl_clear_browser( fd_peruser->group_count_browser );
  sprintf( line_buffer, "%d groups", group_count );
  fl_add_browser_line( fd_peruser->group_count_browser, line_buffer );

  article_count = 0;

  if ( group_line &&
       group_line <= fl_get_browser_maxline( fd_peruser->group_browser ))
      {
          fl_select_browser_line( fd_peruser->group_browser, group_line );
          adjust_browser( fd_peruser->group_browser );
          fl_call_object_callback( fd_peruser->group_browser );

          if ( subject_line )
          {
              fl_select_browser_line( fd_peruser->subject_browser, subject_line );
              adjust_browser( fd_peruser->subject_browser );
              fl_call_object_callback( fd_peruser->subject_browser );
          }
      }
  
  fl_unfreeze_form( fd_peruser->peruser );
  return( 0 );
}

void group_callback( FL_OBJECT *obj, long caller )
{
  int line;
  int index;

  char message_id[ BUFFER_SIZE ];

  enum { no, yes } readable, header, requested, found;

  readable = header = requested = found = user_folder = no;

/*
   Update read file for newsgroup last chosen, if there was one. 
*/

  update_read_file();

/*
   Determine which group was selected.
*/

  strcpy( newsgroup, 
	  fl_get_browser_line( fd_peruser->group_browser,
			       fl_get_browser( fd_peruser->group_browser )));

  if ( newsgroup[ 0 ] == '@' )
    {
      line_pointer = newsgroup;
      line_pointer++;
      line_pointer++;
      strcpy( line_buffer, line_pointer );
      strcpy( newsgroup, line_buffer );
    }

  if ( (  line = fl_get_browser( fd_peruser->group_browser )) >
       ( group_count + 1 ) &&
       line < ( group_count + 7 ))
    folder = yes;
  else
    folder = no;

  if ( line > ( group_count + 6 ))
      user_folder = yes;
  else
      user_folder = no;

  
/*
   Prepare the browsers.
*/

  fl_clear_browser( fd_peruser->subject_browser );
  fl_clear_browser( fd_peruser->article_count_browser );

/*
    open the spool and read files.
*/

  sprintf( line_buffer, "%s/%s", spool_path, newsgroup );
  memcpy( spool_path_and_name, line_buffer, BUFFER_SIZE );
  if (( spool_file = fopen( spool_path_and_name, "r" )) == NULL )
    if ( errno != 2 )
      {
	perror( name );
	exit( EXIT_FAILURE );
      }

  if ( !folder )
    {
      readable = yes;
      sprintf( read_file_path_and_name, "%s.read", line_buffer );
      if (( read_file = fopen( read_file_path_and_name, "r" )) == NULL )
	if ( errno != 2 )
	  {
	    perror( name );
	    exit( EXIT_FAILURE );
	  }
    }

/*
   Load fd_peruser->subject_browser with "Subject: ..." lines from spool.
*/

  fl_freeze_form( fd_peruser->peruser );
  article_count = 0;

  if ( spool_file )
    {
      sprintf( requests_path, "%s/.peruser_spool/%s.requests",
	       home_directory, newsgroup );
      if (( requests = fopen( requests_path, "a+" )) == NULL )
	if ( errno != 2 )
	  {
	    perror( name );
	    exit( EXIT_FAILURE );
	  }

      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, spool_file );
	  if ( feof( spool_file ))
	    break;

	  if ( !strncmp( line_buffer, "@C22Header of article", 21 ))
	    {
	      header = yes;
	      rewind( requests );
	      requested = no;

	      for( ; ; )
		{
		  fgets( second_line_buffer, BUFFER_SIZE, requests );
		  if ( feof( requests ))
		    break;
		  if ( !strcmp( line_buffer, second_line_buffer ))
		    {
		      requested = yes;
		      break;
		    }
		}
	      continue;
	    }
	  
	  if ( !strncmp( newsgroup, "INBOX", 5 ))
	    {
	      if ( !strncmp( line_buffer, "From: ", 6 ) ||
		   !strncmp( line_buffer, "FROM: ", 6 ))
		{
		  found = yes;
		  article_count++;
		  do
		    {
		      fgets( second_line_buffer, BUFFER_SIZE, spool_file );
		    }
		  while( strncmp( second_line_buffer, ".\r\n", 3 ));
	      
		  fgets( read_status, sizeof read_status, read_file );
		  if ( !strncmp( read_status, "u\n", 2 ))
		    sprintf( second_line_buffer, "@b%s", line_buffer );
		  else 
		    strcpy( second_line_buffer, line_buffer );

		  fl_add_browser_line( fd_peruser->subject_browser,
				       strtok( second_line_buffer, "\r\n" ));
		}
	      else
		if ( !strncmp( line_buffer, ".\r\n", 3 ) && !found )
		  {
		    article_count++;
		    fgets( read_status, sizeof read_status, read_file );
		    if ( !strncmp( read_status, "u\n", 2 ))
		      strcpy( second_line_buffer, "@b" );
		    else
		      second_line_buffer[ 0 ] = '\0';
		    strcat( second_line_buffer, "(no From: header line)" );
		    fl_add_browser_line( fd_peruser->subject_browser,
					 strtok( second_line_buffer, "\r\n" ));
		    found = no;
		  }
	    }
	  else
	    if ( !strncmp( newsgroup, "OUTBOX", 6 ) ||
		 !strncmp( newsgroup, "SENT-MAIL", 9 ))
	      {
		if ( !strncmp( line_buffer, "To: ", 4 ))
		  {
		    article_count++;

                    while( line_buffer[ strlen( line_buffer ) - 3 ] == ',' )
                    {
                        strcpy( second_line_buffer, strtok( line_buffer, "\r" ));
                        fgets( line_buffer, BUFFER_SIZE, spool_file );
                        strcat( second_line_buffer, " " );
                        strcat( second_line_buffer, line_buffer +
                                strspn( line_buffer, " \t" ));
                        strcpy( line_buffer, second_line_buffer );
                    }

                    do
		      {
			fgets( second_line_buffer, BUFFER_SIZE, spool_file );
		      }
		    while( strncmp( second_line_buffer, ".\r\n", 3 ));
	      
		    fl_add_browser_line( fd_peruser->subject_browser,
					 strtok( line_buffer, "\r\n" ));

		  }
	      }
            else
              if ( !strncmp( line_buffer, "Subject: ", 9 ) ||
		   !strncmp( line_buffer, "SUBJECT: ", 9 ))
		{
                    line_pointer = strstr( line_buffer, " " );
		  if ( readable )
		    {
		      fgets( read_status, sizeof read_status, read_file );
		      if ( !strncmp( read_status, "u\n", 2 ))
			{
			  if ( header )
			    {
			      header = no;
			      if ( requested )
				sprintf( line_buffer, "@b@C18%s", 
					 line_pointer );
			      else
				sprintf( line_buffer, "@b@C22%s", 
					 line_pointer );
			    }
			  else
			    sprintf( line_buffer, "@b %s", line_pointer );
			}
		      else
			if ( header )
			  {
			    header = no;
			    if ( requested )
			      sprintf( line_buffer, "@C18%s", line_pointer );
			    else
			      sprintf( line_buffer, "@C22%s", line_pointer );
			  }
			else
			  strcpy( line_buffer, line_pointer );
		      line_pointer = line_buffer;
		    }
		
                  fl_add_browser_line( fd_peruser->subject_browser, 
                                       strtok( line_pointer, "\r\n" ));
		  article_count++;
      
                      /*
		    Skip past the rest of the article, in case the string 
		    "Subject: " appears in the body of the article.
		   */

		  do
		    {
		      fgets( line_buffer, BUFFER_SIZE, spool_file );
		    }
		  while( strncmp( line_buffer, ".\r\n", 3 ));
		}

          if ( !strncmp( line_buffer, "\r\n", 2 ))
          {
              fl_add_browser_line( fd_peruser->subject_browser,
                                   "(Message has no subject header.)" );

              ++article_count;
              
              do
              {
                  fgets( line_buffer, BUFFER_SIZE, spool_file );
              }
              while( strncmp( line_buffer, ".\r\n", 3 ));
          }
	}
      
      rewind( spool_file );

      for( line = 1; line <= article_count; line++ )
	{
	  do
	    {
	      fgets( second_line_buffer, BUFFER_SIZE, spool_file );

	      if ( !strncmp( second_line_buffer, "Message-ID: ", 12 ) ||
                   !strncmp( second_line_buffer, "Message-Id: ", 12 ))
		strcpy( message_id, strtok( second_line_buffer + 12,
					    "\r\n" ));
	    }
	  while( strncmp( second_line_buffer, ".\r\n", 3 ));

	  rewind( requests );

	  for( ; ; )
	    {
	      fgets( line_buffer, BUFFER_SIZE, requests );
	      if ( feof( requests ))
		break;
	  
	      if ( line_buffer[ 0 ] == '@' )
		continue;

	      if ( strstr( strstr( line_buffer, ":" ) + 1, message_id ))
		{
		  toggle_italic( line );
		  break;
		}
	    }
	}

      fclose( requests );
      fclose( spool_file );
      if ( readable )
	fclose( read_file );
    }

/*
   Create read array.
*/

  if ( article_count && readable )
    {
      if ( read_array != NULL )
	{ 
	  fprintf( stderr, "peruser: read_array non-NULL\n" );
	  exit( EXIT_FAILURE );
	}
      if ( !( read_array = malloc( article_count )))
	{
	  fputs( "peruser: malloc failed in group_callback().\n", stderr );
	  exit( EXIT_FAILURE );
	}
      
      for( index = 0; index < article_count; index++ )
	{
	  line = index;
	  line++;
	  strcpy( line_buffer,
		  fl_get_browser_line( fd_peruser->subject_browser, line ));
	  if ( !strncmp( line_buffer, "@b", 2 ))
	    {
	      read_array[ index ] = UNREAD;
	    }
	  else
	    {
	      read_array[ index ] = READ;
	    }
	}
    }

/*
   Update article_count_browser.
*/

  sprintf( line_buffer, "%d article(s)", article_count );
  fl_add_browser_line( fd_peruser->article_count_browser, line_buffer );

  if ( article_count && !caller )
    {
      fl_select_browser_line( fd_peruser->subject_browser, 1 );
      fl_call_object_callback( fd_peruser->subject_browser );
    }
  else
      fl_clear_browser( fd_peruser->message_browser );

  fl_unfreeze_form( fd_peruser->peruser );
  return;
}

extern void highlight_url( char *original_line );

void subject_callback( FL_OBJECT *obj, long caller )
{
  int line, line_copy, index, templine;

  enum{ no, yes } header, match, sent_mail, found;

  if ( !( line = fl_get_browser( fd_peruser->subject_browser )))
    return;

  header = match = sent_mail = found = no;


/*
   Choose the form to use.
*/

  if ( !strncmp( newsgroup, "SENT-MAIL", 9 ))
    sent_mail = yes;

  if ( folder )
  {
      fl_hide_object( fd_peruser->select_menu );
      fl_hide_object( fd_peruser->follow_up_button );
      fl_hide_object( fd_peruser->reply_button );
      fl_hide_object( fd_peruser->both_button );

      fl_show_object( fd_peruser->delete_menu );
  }
  else
  {
      fl_show_object( fd_peruser->select_menu );
      fl_show_object( fd_peruser->reply_button );

      if ( strncmp( newsgroup, "INBOX", 5 ) &&
           !user_folder )
      {
          fl_hide_object( fd_peruser->delete_menu );
          fl_show_object( fd_peruser->follow_up_button );
          fl_show_object( fd_peruser->both_button );
      }
      else
      {
          fl_hide_object( fd_peruser->follow_up_button );
          fl_hide_object( fd_peruser->both_button );
          fl_show_object( fd_peruser->delete_menu );
      }
  }

  if ( strncmp( newsgroup, "OUTBOX", 6 ) &&
       strncmp( newsgroup, "FOLLOW-UPS", 10 ))
      fl_hide_object( fd_peruser->edit_button );
  else
      fl_show_object( fd_peruser->edit_button );
      
      /*
        Open spool file.
      */

  if (( spool_file = fopen( spool_path_and_name, "r" )) == NULL )
    {
      perror( name );
      exit( EXIT_FAILURE );
    }

/*
   Find article's position in spool.
*/

  switch( caller )
    {

/*
   fd_peruser->subject_browser.
*/

    case 0:
      line_copy = line;
      line--;
      break;

/*
   fd_article->previous, fd_follow-up->previous.
*/

    case 1:
      templine = line;
      line--;
      if ( !line )
	return;
      line_copy = line;
      fl_deselect_browser_line( fd_peruser->subject_browser, templine );
      fl_select_browser_line( fd_peruser->subject_browser, line );
      adjust_browser( fd_peruser->subject_browser );
      line--;
      break;

/*
   fd_article->next_button, fd_follow_up->next_button.
*/

    case 2:
      if ( line == article_count )
	return;
      templine = line;
      templine++;
      line_copy = templine;
      fl_deselect_browser_line( fd_peruser->subject_browser, line );
      fl_select_browser_line( fd_peruser->subject_browser, templine );
      adjust_browser( fd_peruser->subject_browser );
      break;

    default:
      line_copy = 0;
    }

/*
   Remove boldface attribute prefix.
*/

  if ( !folder )
    {    
      strcpy( line_buffer, fl_get_browser_line( fd_peruser->subject_browser,
						line_copy ));

      if ( !strncmp( line_buffer, "@b", 2 ))
	{
	  line_pointer = line_buffer;
	  line_pointer++;
	  line_pointer++;
	  fl_replace_browser_line( fd_peruser->subject_browser, line_copy,
				   line_pointer );

/*
   Update read array.
*/

	  read_array[ line ] = READ;

/* 
   Search read array for remaining unread articles. If none are found,
   remove the boldface attribute prefix from the newsgroup name in
   fd_peruser->group_browser. 
*/

	  found = no;
	  for( index = 0; index < article_count; index++ )
	    {
	      if ( !read_array[ index ] )
		{
		  found = yes;
		  break;
		}
	    }
	  
	  if ( !found )
	    {
	      templine = fl_get_browser( fd_peruser->group_browser );
	      strcpy( line_buffer,
		      fl_get_browser_line( fd_peruser->group_browser, 
					   templine ));
	      line_pointer = line_buffer;
	      line_pointer++;
	      line_pointer++;
	      fl_replace_browser_line( fd_peruser->group_browser, templine,
				       line_pointer );
	    }
	}
    }

/*
   Advance spool to desired article.
*/

  while( line > 0 )
    {
      do
	{
	  fgets( line_buffer, BUFFER_SIZE, spool_file );
	}
      while( strncmp( line_buffer, ".\r\n", 3 ));
      line--;
    }

/*
   Prepare the browser.
*/

  fl_freeze_form( fd_peruser->peruser );
  fl_clear_browser( fd_peruser->message_browser );

/*
   Load article.
*/

  for( ; ; )
    {
      fgets( line_buffer, BUFFER_SIZE, spool_file );
      if ( !strncmp( line_buffer, ".\r\n", 3 ))
	break;

/* 
   If a line starts with a double period, remove one period, unless
   the article is from group FOLLOW-UPS. Tell will double initial
   periods in outgoing messages.
*/

      if ( !strncmp( line_buffer, "..", 2) && 
	   !strncmp( newsgroup, "FOLLOW-UPS", 10 ))
	{
	  memcpy( second_line_buffer, line_buffer, BUFFER_SIZE );
	  line_pointer = second_line_buffer + 1;
	  strcpy( line_buffer, line_pointer );
	}

/* 
   If the article contains only a header, then the first line
   will consist of "@C22Header of article number x:", where x is the
   article number on the configured server.
*/

      if ( !strncmp( line_buffer, "@C22Header of article", 21 ))
	{
	  header = yes;

	  sprintf( requests_path, "%s/.peruser_spool/%s.requests",
		   home_directory, newsgroup );
	  if (( requests = fopen( requests_path, "a+" )) == NULL )
	    {
	      perror( name );
	      exit( EXIT_FAILURE );
	    }
	  
	  rewind( requests );
	  for( ; ; )
	    {
	      fgets( second_line_buffer, BUFFER_SIZE, requests );
	      if ( feof( requests ))
		break;
	      if ( !strcmp( line_buffer, second_line_buffer ))
		{
		  sprintf( second_line_buffer, 
			   "@C18%s: requested for downloading", 
			   strtok( line_buffer + 4, ":" ));
		  match = yes;
		  break;
		}
	    }
	  fclose( requests );
	  if ( !match )
	    {
	      sprintf( second_line_buffer, "%s: not requested for downloading",
		       strtok( line_buffer, ":" ));
	    }
	  sprintf( line_buffer, "%s\n@-", second_line_buffer );
	}

/* 
   "Subject: " header lines are added in red.
*/

      if ( !strncmp( line_buffer, "Subject: ", 9 ))
	{
	  sprintf( second_line_buffer, "@C1Subject: %s", 
		   line_buffer + 9 + strspn( line_buffer + 9, " \t" ));
	  strcpy( line_buffer, second_line_buffer );
	}

/*
  "From: " header lines are added in blue.
*/

      if ( !strncmp( line_buffer, "From: ", 6 ))
	{
	  sprintf( second_line_buffer, "@C4%s", line_buffer );
	  strcpy( line_buffer, second_line_buffer );
	}

/*
  Quotes in .
*/

      if ( line_buffer[ strspn( line_buffer, " \t" ) ] == '>' )
	{
	  sprintf( second_line_buffer, "@i%s", line_buffer );
	  strcpy( line_buffer, second_line_buffer );
	}

/*
   Add line to appropriate form.
*/

      highlight_url( line_buffer );
      fl_add_browser_line( fd_peruser->message_browser,
                           ( line_buffer[ 0 ] != '\r' && line_buffer[ 0 ] != '\n' )
                           ? strtok( line_buffer, "\r\n" ): "\r" );
    }
  fclose( spool_file );

/*
   Show the form.
*/

  fl_unfreeze_form( fd_peruser->peruser );

  return;
}

void mark_group( void )
{
  int index;
  int line;

/*
   If there's no articles, or the currently selected newsgroup is FOLLOW-UPS,
   or if the newsgroup is already marked fully read (the group name is not in
   boldface), then return.
*/

  index = fl_get_browser( fd_peruser->group_browser );

  if ( !index || !article_count || 
       strncmp( fl_get_browser_line( fd_peruser->group_browser, index ),
	        "@b", 2 ))
    return;

/*
   Remove boldface attribute prefixes from lines that have it in
   fd_peruser->subject_browser, and set the read array element corresponding
   to those articles to READ.
*/
   
  fl_freeze_form( fd_peruser->peruser );
  for( index = 0; index < article_count; index++ )
    {
      strcpy( line_buffer,
	      fl_get_browser_line( fd_peruser->subject_browser, index + 1 )); 
      
      if ( !strncmp( line_buffer, "@b", 2 ))
	{
	  fl_replace_browser_line( fd_peruser->subject_browser, index + 1,
		 		   line_buffer + 2 );
	  
	  read_array[ index ] = READ;
	}
    }
  fl_unfreeze_form( fd_peruser->peruser );

/*
   Remove the boldface attribute from the newsgroup name in
   fd_peruser->group_browser.
*/

  line = fl_get_browser( fd_peruser->group_browser );
  strcpy( line_buffer,
	  fl_get_browser_line( fd_peruser->group_browser, line ));
  fl_replace_browser_line( fd_peruser->group_browser, line, line_buffer + 2 );

  return;
}

void pack_callback( FL_OBJECT *obj, long caller )
{
  int index, group, subject;

  fl_freeze_form( fd_peruser->peruser );

  switch( caller )
    {
      
/*
   fd_peruser->pack_button.
*/

    case 0:
      fl_clear_browser( fd_regex->regex_browser );

/* We must call update_read_file() before packer to make sure the
   currently selected newsgroup's (if there is one) read file is
   updated. We set article_count to zero before calling group_callback
   because group_callback will also call update_read_file, which
   returns if article_count is zero, preventing read_array from being
   freed twice.
*/

      strcpy( second_line_buffer, "cpacker" );

      update_read_file();

      run_client_program();

      article_count = 0;
      refresh_newsgroups();
      break;
      
/*
   fd_peruser->mark-button.
*/

    case 1:
      if ( fl_get_browser( fd_peruser->group_browser ) > group_count )
	return;
      mark_group();
      break;

/*
  fd_peruser->mark_all_button.
*/

    case 2:
        group = fl_get_browser( fd_peruser->group_browser );
        subject = fl_get_browser( fd_peruser->subject_browser );

        for( index = 1; index <= group_count ; index++ )
	{
	  fl_select_browser_line( fd_peruser->group_browser, index );
	  fl_call_object_callback( fd_peruser->group_browser );
	  mark_group();
	}

        fl_select_browser_line( fd_peruser->group_browser, group );
        fl_call_object_callback( fd_peruser->group_browser );
        adjust_browser( fd_peruser->group_browser );

        fl_select_browser_line( fd_peruser->subject_browser, subject );
        fl_call_object_callback( fd_peruser->subject_browser );
        adjust_browser( fd_peruser->subject_browser );
    }
  
  fl_unfreeze_form( fd_peruser->peruser );
  return;
}

void tell_callback( FL_OBJECT *obj, long caller )
{
  fl_clear_browser( fd_regex->regex_browser );

  sprintf( second_line_buffer, "cteller %s", server_address );
  update_read_file();

  run_client_program();
  
  article_count = 0;
  refresh_newsgroups();

  return;
}

void about_callback( FL_OBJECT *obj, long caller )
{
    fl_show_message( "News Peruser", "copyright James Bailie 1996, 1997",
                     "jbailie@wwdc.com" );

    return;
}

void configure_callback( FL_OBJECT *obj, long caller )
{
    switch( caller )
    {
      
/*
   Called by fd_peruser->configure_button.
*/

        case 0:
            fl_set_input( fd_configure->name_input, user_name );
            fl_set_input_return( fd_configure->name_input, FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->email_input, user_email_address );
            fl_set_input_return( fd_configure->email_input, FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->editor_input, editor_command );
            fl_set_input_return( fd_configure->editor_input, FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->server_input, server_address );
            fl_set_input_return( fd_configure->server_input, FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->pop_server_input, pop_server_address );
            fl_set_input_return( fd_configure->pop_server_input, FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->pop_password_input, pop_password );
            fl_set_input_return( fd_configure->pop_password_input, 
                                 FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->smtp_server_input, smtp_server_address );
            fl_set_input_return( fd_configure->smtp_server_input,
                                 FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->nntp_username_input, nntp_username );
            fl_set_input_return( fd_configure->nntp_username_input,
                                 FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->nntp_password_input, nntp_password );
            fl_set_input_return( fd_configure->nntp_password_input,
                                 FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->web_input, web_browser );
            fl_set_input_return( fd_configure->web_input,
                                 FL_RETURN_ALWAYS );

            fl_set_input( fd_configure->ftp_input, ftp_client );
            fl_set_input_return( fd_configure->ftp_input,
                                 FL_RETURN_ALWAYS );
      
            fl_set_button( fd_configure->authenticating_button, authenticating );
            fl_call_object_callback( fd_configure->authenticating_button );

            fl_show_form( fd_configure->configure, FL_PLACE_MOUSE, 
                          FL_FULLBORDER, "Configure" );

            fl_deactivate_form( fd_peruser->peruser );
            break;

/*
   Called by fd_configure->name_input.
*/

        case 1:
            strcpy( user_name, fl_get_input( fd_configure->name_input ));
            break;

/*
   Called by fd_configure->email_input.
*/

        case 2:
            strcpy( user_email_address, fl_get_input( fd_configure->email_input ));
            break;

/*
   Callled by fd_configure->editor_input.
*/
      
        case 3:
            strcpy( editor_command, fl_get_input( fd_configure->editor_input ));
            break;

/*
   Called by fd_configure->configure_done_button
*/

        case 4:
            if ( !strlen( user_name ) || !strlen( user_email_address ) ||
                 !strlen( editor_command ) || !strlen( web_browser ) ||
                 !strlen( server_address ) || !strlen( pop_server_address ) ||
                 !strlen( pop_password ) || !strlen( ftp_client ) ||
                 ( authenticating && 
                   ( !strlen( nntp_username) || !strlen( nntp_password ))))
            {
                fl_show_alert( "Configuration information incomplete!", 
                               "You must fill out each item in the configure form",
                               "before you may use peruser.", 1 );

                break;
            }
            else
                configuration = yes;
            
            if (( config_file = fopen( config_path_and_name, "w" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }

            fprintf( config_file,
                     "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%d\n%s\n%s\n%s\n%s\n",
                     user_name, user_email_address, editor_command,
                     server_address, pop_server_address, pop_password,
                     smtp_server_address, authenticating, nntp_username, 
                     nntp_password, web_browser, ftp_client );

            fflush( config_file );
            fclose( config_file );
            fl_hide_form( fd_configure->configure );
            fl_activate_form( fd_peruser->peruser );
            break;

/*
   Called by fd_configure->server_input.
*/

        case 6:
            strcpy( server_address, fl_get_input( fd_configure->server_input ));
            break;

/*
  fd_configure->pop_server_input
*/

        case 10:
            strcpy( pop_server_address, 
                    fl_get_input( fd_configure->pop_server_input ));
            break;

/*
  fd_configure->pop_password_input
*/

        case 20:
            strcpy( pop_password, fl_get_input( fd_configure->pop_password_input ));
            break;

/*
  fd_configure->smtp_server_input
*/
        case 30:
            strcpy( smtp_server_address, 
                    fl_get_input( fd_configure->smtp_server_input ));
            break;

        case 40:
            authenticating = fl_get_button( fd_configure->authenticating_button );
            if ( authenticating )
            {
                fl_show_object( fd_configure->nntp_username_input );
                fl_show_object( fd_configure->nntp_password_input );
            }
            else
            {
                strcpy( nntp_username, "none" );
                strcpy( nntp_password, "none" );
                fl_hide_object( fd_configure->nntp_username_input );
                fl_hide_object( fd_configure->nntp_password_input );
            }
            break;

        case 50:
            strcpy( nntp_username, 
                    fl_get_input( fd_configure->nntp_username_input ));
            break;

        case 60:
            strcpy( nntp_password, 
                    fl_get_input( fd_configure->nntp_password_input ));
            break;
        case 100:
            strcpy( web_browser,
                    fl_get_input( fd_configure->web_input ));
            break;
        case 110:
            strcpy( ftp_client,
                    fl_get_input( fd_configure->ftp_input ));
            break;
    }
  return;
}

void select_all_callback( FL_OBJECT *obj, long caller )
{
  int group_line, subject_line, subject_topline, article_line, max_line, index;

  group_line = fl_get_browser( fd_peruser->group_browser );
  subject_line = fl_get_browser( fd_peruser->subject_browser );
  subject_topline = fl_get_browser_topline( fd_peruser->subject_browser );
  article_line = fl_get_browser_topline( fd_peruser->message_browser );

  if ( !article_count || ( group_line > group_count ))
    return;

  fl_freeze_form( fd_peruser->peruser );

  max_line = fl_get_browser_maxline( fd_peruser->subject_browser );

  for( index = 0; index < max_line; index++ )
    {
      fl_select_browser_line( fd_peruser->subject_browser, index + 1 );
      fl_set_browser_topline( fd_peruser->subject_browser, index + 1 );
      strcpy( line_buffer, 
	      fl_get_browser_line( fd_peruser->subject_browser, index + 1 ));

      if ( strstr( line_buffer, "@C18" ) || strstr( line_buffer, "@C22" ))
	fl_call_object_callback( fd_peruser->subject_browser );
      else
	continue;

      if ( *fl_get_browser_line( fd_peruser->message_browser, 1 )  == '@' )
	select_callback( obj, 2 );
    }

  fl_set_browser_topline( fd_peruser->subject_browser, subject_topline );
  fl_select_browser_line( fd_peruser->subject_browser, subject_line );
  fl_call_object_callback( fd_peruser->subject_browser );

  fl_set_browser_topline( fd_peruser->message_browser, article_line );

  fl_unfreeze_form( fd_peruser->peruser );
  return;
}

void select_thread_callback( FL_OBJECT *obj, long caller );

void select_callback( FL_OBJECT *obj, long caller )
{
  int line, choice;

  if ( !article_count )
    return;

  switch( caller )
    {
    case 0:
      choice = fl_get_menu( fd_peruser->select_menu );
      break;
    case 2:
      choice = 1;
      break;
      
    default:
      choice = 0;
    }

  switch( choice )
    {
    case 1:
      strcpy( line_buffer, fl_get_browser_line( fd_peruser->message_browser, 1 ));
      if ( !strstr( line_buffer, "Header of article" ) && 
	   !strstr( line_buffer, "selected for downloading" ))
	return;

      sprintf( requests_path, "%s/.peruser_spool/%s.requests",
	       home_directory, newsgroup );
      if (( requests = fopen( requests_path, "a+" )) == NULL )
	{
	  perror( name );
	  exit( EXIT_FAILURE );
	}

      line_pointer = strrchr( line_buffer, ':' ) + 2;
  
      if ( !strncmp( line_pointer, "not", 3 ))
	{
	  strcpy( second_line_buffer, strtok( line_buffer, ":" ));
	  fprintf( requests, "%s:\r\n", second_line_buffer );
	  fflush( requests );
	  fclose( requests );
	  sprintf( line_buffer, "@C18%s: requested for downloading", 
		   second_line_buffer + 4 );
	  fl_replace_browser_line( fd_peruser->message_browser, 1,
				   line_buffer );

	  line = fl_get_browser( fd_peruser->subject_browser );
	  strcpy( line_buffer, 
		  fl_get_browser_line( fd_peruser->subject_browser, line ));

	  sprintf( second_line_buffer, "@C18%s", 
		   strstr( line_buffer, "@i" ) ? 
		   strstr( line_buffer, "@i" ) : 
		   strstr( line_buffer, " " ));

	  fl_replace_browser_line( fd_peruser->subject_browser, line,
				   second_line_buffer );

	}
      else
	{
	  if (( temp_requests = fopen( temp_requests_path, "w" )) == NULL )
	    {
	      perror( name );
	      exit( EXIT_FAILURE );
	    }

	  rewind( requests );
	  for( ; ; )
	    {
	      fgets( second_line_buffer, BUFFER_SIZE, requests );
	      if ( feof( requests ))
		break;
	      if ( !strncmp( line_buffer + 4, second_line_buffer + 4, 
			     strlen( second_line_buffer ) - 7 ))
		continue;
	      fputs( second_line_buffer, temp_requests );
	    }
	  fflush( temp_requests );
	  fclose( temp_requests );
	  fclose( requests );
	  rename( temp_requests_path, requests_path );
	  sprintf( second_line_buffer, "@C22%s: not requested for downloading",
		   strtok( line_buffer + 4, ":" ));
	  fl_replace_browser_line( fd_peruser->message_browser, 1,
				   second_line_buffer );

	  line = fl_get_browser( fd_peruser->subject_browser );
	  strcpy( line_buffer, 
		  fl_get_browser_line( fd_peruser->subject_browser, line ));

	  sprintf( second_line_buffer, "@C22%s", 
		   strstr( line_buffer, "@i" ) ? 
		   strstr( line_buffer, "@i" ) : 
		   strstr( line_buffer, " " ));
	  fl_replace_browser_line( fd_peruser->subject_browser, line,
				   second_line_buffer );
	}
      break;
    case 2:
      references_callback( obj, caller );
      break;
    case 3:
      select_thread_callback( obj, caller );
      break;
    case 4:
      select_all_callback( obj, caller );
      break;
    }

  return;
}

void select_thread_callback( FL_OBJECT *obj, long caller )
{
  int current, line, maxline, first_tabs, second_tabs;
  char *first, *second;


  fl_freeze_form( fd_peruser->peruser );

  current = line = fl_get_browser( fd_peruser->subject_browser );
  maxline = fl_get_browser_maxline( fd_peruser->subject_browser );

  select_callback( obj, 2 );
  references_callback( obj, caller );
  
  while( line > 1 )
    {
      first = strchr( fl_get_browser_line( fd_peruser->subject_browser, line-- ), ' ' ) + 1;
      first_tabs = strspn( first, "\t" );
      first = first + first_tabs;

      second = strchr( fl_get_browser_line( fd_peruser->subject_browser, line ), ' ' ) + 1;
      second_tabs = strspn( second, "\t" );
      second = second + second_tabs;

      if ( !strcmp( first, second ) || strstr( first, second ) ||
	   ( first_tabs && second_tabs ))
	{
	  fl_select_browser_line( fd_peruser->subject_browser, line );
	  fl_call_object_callback( fd_peruser->subject_browser );
	  select_callback( obj, 2 );
          references_callback( obj, caller );
	}
      else
	break;
    }

  line = current;

  while( line < maxline )
    {
      first = strchr( fl_get_browser_line( fd_peruser->subject_browser, line++ ), ' ' ) + 1;
      first_tabs = strspn( first, "\t" );
      first = first + first_tabs;

      second = strchr( fl_get_browser_line( fd_peruser->subject_browser, line ), ' ' ) + 1;
      second_tabs = strspn( second, "\t" );
      second = second + second_tabs;

      if ( !strcmp( first, second ) || strstr( second, first ) ||
	   ( first_tabs && second_tabs ))
	{
	  fl_select_browser_line( fd_peruser->subject_browser, line );
	  fl_call_object_callback( fd_peruser->subject_browser );
	  select_callback( obj, 2 );
          references_callback( obj, caller );
	}
      else
	break;
    }

  fl_select_browser_line( fd_peruser->subject_browser, current );
  fl_call_object_callback( fd_peruser->subject_browser );

  fl_unfreeze_form( fd_peruser->peruser );

  return;
}

void write_timestamp()
{
    time_t unix_time;
    struct tm *gmt;
    
    unix_time = time( NULL );
    gmt = gmtime( &unix_time );

    sprintf( line_buffer, "%2.2u%2.2u%2.2u %2.2u%2.2u%2.2u\n",
             gmt->tm_year, gmt->tm_mon + 1, gmt->tm_mday,
             gmt->tm_hour, gmt->tm_min, gmt->tm_sec );

    sprintf( second_line_buffer, "%s/.timestamp", spool_path );
    if (( temp_file = fopen( second_line_buffer, "w" )) == NULL )
    {
        perror( name );
        exit( EXIT_FAILURE );
    }
        
    fputs( line_buffer, temp_file );
    fflush( temp_file );
    fclose( temp_file );

    return;
}

void edit_newsrc_callback( FL_OBJECT *obj, long caller )
{
  static struct re_pattern_buffer buffer;
  struct re_registers registers;
  const char *result;
  int return_code;

  FL_OBJECT *browser;
  
  char group[ BUFFER_SIZE ];
  int maxline, index;
  static int changed;
  char **array;
  
  switch( caller )
    {
    case 1:
      strcpy( group,
              fl_get_browser_line( fd_newsrc->newsrc_browser, 
                                   index = fl_get_browser( 
                                              fd_newsrc->newsrc_browser )));

      sprintf( line_buffer, "%s/%s", spool_path, strtok( group, ":" ));
      remove( line_buffer );
      sprintf( second_line_buffer, "%s.read", line_buffer );
      remove( second_line_buffer );
      sprintf( second_line_buffer, "%s.requests", line_buffer );
      remove( second_line_buffer );

      fl_delete_browser_line( fd_newsrc->newsrc_browser, index );

      changed = 1;
      break;
    case 2:
      strcpy( group, fl_get_browser_line( fd_newsrc->list_browser,
					  fl_get_browser( 
					     fd_newsrc->list_browser )));
      strcat( group, ":" );
      fl_add_browser_line( fd_newsrc->newsrc_browser, group );

      changed = 1;
      break;
    case 3:
        sprintf( line_buffer, "%s/.timestamp", spool_path );
        sprintf( second_line_buffer, "%s/.timestamp.bak", spool_path );
        rename( line_buffer, second_line_buffer );
        
        sprintf( line_buffer, "%s/LIST", spool_path );
        sprintf( second_line_buffer, "%s/LIST.bak", spool_path );
        rename( line_buffer, second_line_buffer );
        
        sprintf( line_buffer, "chearer %s %s %s %s list",
	       server_address,
	       authenticating ? "-a" : " ",
	       authenticating ? nntp_username : " ",
	       authenticating ? nntp_password : " " );

        strcpy( second_line_buffer, line_buffer );
        run_client_program();

        sprintf( line_buffer, "%s/LIST", spool_path );
        if (( spool_file = fopen( line_buffer, "r" )) == NULL )
        {
            if ( errno == 2 )
            {
                sprintf( second_line_buffer, "%s/LIST.bak", spool_path );
                rename( second_line_buffer, line_buffer );

                sprintf( line_buffer, "%s/.timestamp", spool_path );
                sprintf( second_line_buffer, "%s/.timestamp.bak", spool_path );
                rename( second_line_buffer, line_buffer );
            }
            else
            {
                perror( name );
                exit( EXIT_FAILURE );
            }
        }
        else
            {
                fclose( spool_file );
                write_timestamp();
            }
      
        fl_clear_browser( fd_newsrc->list_browser );
        sprintf( line_buffer, "%s/LIST", spool_path );

        strcpy( second_line_buffer, "@cLoading LIST..." );
        fl_replace_browser_line( fd_newsrc->count_browser, 1,
                                 second_line_buffer );
        sync_form();

        fl_freeze_form( fd_newsrc->newsrc );
        fl_load_browser( fd_newsrc->list_browser, line_buffer );
        sprintf( line_buffer, "@c%d\rGROUPS",
                 fl_get_browser_maxline( fd_newsrc->list_browser ));
        fl_replace_browser_line( fd_newsrc->count_browser, 1,
                                 line_buffer );
        fl_unfreeze_form( fd_newsrc->newsrc );
        break;
        case 4:
      if ( ! ( maxline = fl_get_browser_maxline( fd_newsrc->newsrc_browser )))
	{
	  fl_show_alert( "Error:",
			 "You must enter at least one newsgroup in",
			 "your ~/.peruser-newsrc.", 0 );
	  break;
	}

      if ( changed )
	{
	  sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );
	  if (( newsrc = fopen( line_buffer, "w" )) == NULL )
	    { 
	      perror( name );
	      exit( EXIT_FAILURE );
	    }

	  if ( maxline )
	    {
	      for( index = 1; index <= maxline; index++ )
		{
		  strcpy( line_buffer, 
			  ( fl_get_browser_line( fd_newsrc->newsrc_browser, 
						 index )));
		  strcat( line_buffer, "\n" );
		  fputs( line_buffer, newsrc );
		}

	      fflush( newsrc );
	      fclose( newsrc );
	    }
	}
 
      fl_hide_form( fd_newsrc->newsrc );
      update_read_file();
      refresh_newsgroups();
      fl_activate_form( fd_peruser->peruser );

      break;
    case 5:
        fl_deactivate_form( fd_peruser->peruser );
        
        changed = 0;

      fl_clear_browser( fd_newsrc->newsrc_browser );
      fl_clear_browser( fd_newsrc->list_browser );
      fl_clear_browser( fd_newsrc->count_browser );
      
      fl_set_input_return( fd_newsrc->search_input, FL_RETURN_CHANGED );

      fl_show_form( fd_newsrc->newsrc, FL_PLACE_MOUSE | FL_PLACE_FREE, 
		    FL_FULLBORDER,
		    "edit ~/.peruser-newsrc" );

      sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );
      fl_load_browser( fd_newsrc->newsrc_browser, line_buffer );

      strcpy( second_line_buffer, "@cLoading LIST..." );
      fl_addto_browser( fd_newsrc->count_browser, second_line_buffer );
      sync_form();

      fl_freeze_form( fd_newsrc->newsrc );
      sprintf( line_buffer, "%s/LIST", spool_path );      
      fl_load_browser( fd_newsrc->list_browser, line_buffer );

      sprintf( line_buffer, "@c%d\rGROUPS",
               fl_get_browser_maxline( fd_newsrc->list_browser ));
      fl_replace_browser_line( fd_newsrc->count_browser, 1,
                               line_buffer );
      fl_unfreeze_form( fd_newsrc->newsrc );
      break;
    case 6:
      strcpy( line_buffer, fl_get_input( fd_newsrc->newsrc_input ));
      if ( line_buffer[ 0 ] == '\0' )
	break;

      if ( line_buffer[ strlen( line_buffer ) - 1 ] != ':' )
	strcat( line_buffer, ":" );

      fl_add_browser_line( fd_newsrc->newsrc_browser, line_buffer );
      changed = 1;
      break;
    case 7:
        switch( fl_get_menu( fd_newsrc->sort_menu ))
        {
            case 1:
                browser = fd_newsrc->newsrc_browser;
                break;

            default:
                browser = fd_newsrc->list_browser;
                break;
        }
        
        fl_freeze_form( fd_newsrc->newsrc );

        maxline = fl_get_browser_maxline( browser );

        if ( !( array = calloc( maxline, sizeof array )))
        {
            perror( name );
            exit( EXIT_FAILURE );
        }
        

        for( index = 0; index < maxline; index++ )
	{
            strcpy( line_buffer, fl_get_browser_line( browser,
                                                      index + 1 ));
            if ( ! ( array[ index ] = malloc( strlen( line_buffer ) + 1 )))
            {
                perror( name );
                exit( EXIT_FAILURE );
            }
            
            strcpy( array[ index ], line_buffer );
	}

        qsort( array, maxline, sizeof array, compare );

        for( index = 0; index < maxline; index++ )
	{
            fl_replace_browser_line( browser, index + 1, 
                                     array[ index ] );
            free( array[ index ] );
	}

        free( array );
        if ( browser == fd_newsrc->newsrc_browser )
            changed = 1;
        else
        {
            sprintf( line_buffer, "%s/LIST", spool_path );
            if (( spool_file = fopen( line_buffer, "w" )) == NULL )
            {
                perror( name );
                exit( EXIT_FAILURE );
            }

            for( index = 1; index <= maxline; index++ )
            {
                strcpy( line_buffer,
                        fl_get_browser_line( fd_newsrc->list_browser,
                                             index ));
                strcat( line_buffer, "\n" );
                fputs( line_buffer, spool_file );
            }

            fclose( spool_file );
        }

        fl_unfreeze_form( fd_newsrc->newsrc );
        break;
    case 10:
    case 12:
      strcpy( group, fl_get_input( fd_newsrc->search_input ));
      if ( group[ 0 ] == '\0' )
	break;

      re_syntax_options = RE_NO_BK_VBAR | RE_NO_BK_PARENS;

      buffer.translate = NULL;
      buffer.fastmap = ( char *)1;
      buffer.buffer = NULL;
      buffer.allocated = 0;

      maxline = fl_get_browser_maxline( fd_newsrc->list_browser );

      if ( caller == 12 )
	{
	  if ( !( index = fl_get_browser( fd_newsrc->list_browser )) ||
	       ++index > maxline )
	    index = 1;
	}
      else
	index = 1;

      for( ; index <= maxline; index++ )
	{
	  strcpy( line_buffer, fl_get_browser_line( fd_newsrc->list_browser,
						    index ));
	  

	  if ( ( result
		 = re_compile_pattern( group, strlen( group ), &buffer )) )
	    {
	      fl_show_alert( "Error:", result, "", 0 );
	      break;
	    }

	  return_code = re_search( &buffer, line_buffer, strlen( line_buffer ),
				   0, strlen( line_buffer ), &registers );

	  if ( return_code == -1 )
	    continue;

	  if ( return_code == -2 )
	    {
	      fl_show_alert( "An internal error occurred",
			     "in the regex library", "", 0 );
	      break;
	    }

	  fl_select_browser_line( fd_newsrc->list_browser, index );
	  fl_set_browser_topline( fd_newsrc->list_browser, index );
	  break;
	}
      break;
        case 50:
            sprintf( second_line_buffer, "%s/.timestamp", spool_path );
            if (( temp_file = fopen( second_line_buffer, "r" )) == NULL )
            {
                if ( errno != 2 )
                    {
                        perror( name );
                        exit( EXIT_FAILURE );
                    }
                else
                {
                    fl_show_message( "I can't issue a NEWGROUPS command",
                                     "because I don't have a timestamp",
                                     "on file. Retrieve a LIST to create one." );
                    return;
                }
            }

            fclose( temp_file );

            sprintf( second_line_buffer, "chearer %s %s %s %s list",
                     server_address,
                     authenticating ? "-a" : " ",
                     authenticating ? nntp_username : " ",
                     authenticating ? nntp_password : " " );

            fl_deactivate_form( fd_newsrc->newsrc );
            run_client_program();
            fl_activate_form( fd_newsrc->newsrc );
            
            write_timestamp();
            
            strcpy( second_line_buffer, "@cLoading LIST..." );
            fl_replace_browser_line( fd_newsrc->count_browser, 1,
                                     second_line_buffer );
            sync_form();
            
            fl_clear_browser( fd_newsrc->list_browser );
            sprintf( line_buffer, "%s/LIST", spool_path );
            fl_load_browser( fd_newsrc->list_browser, line_buffer );

            sprintf( line_buffer, "@c%d\rGROUPS",
                     fl_get_browser_maxline( fd_newsrc->list_browser ));
            fl_replace_browser_line( fd_newsrc->count_browser, 1,
                                     line_buffer );

            break;
            
    }

  return;
}

void clear_callback( FL_OBJECT *obj, long caller )
{
  int subject_line, subject_topline, article_line;

  subject_line = fl_get_browser( fd_peruser->subject_browser );
  subject_topline = fl_get_browser_topline( fd_peruser->subject_browser );
  article_line = fl_get_browser( fd_peruser->message_browser );

  switch( caller )
    {
    case 0:
      sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );
      if (( newsrc = fopen( line_buffer, "r" )) == NULL )
	{
	  perror( name );
	  exit( EXIT_FAILURE );
	}

      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, newsrc );
	  if ( feof( newsrc ))
	    break;
	  if ( line_buffer[ 0 ] == '\n' )
	    continue;
	  strcpy( second_line_buffer, strtok( line_buffer, ":" ));
      
	  sprintf( line_buffer, "%s/.peruser_spool/%s.requests", 
		   home_directory, second_line_buffer );
	  if ( truncate( line_buffer, 0 ))
	    if ( errno != ENOENT )
	      {
		perror( name );
		exit( EXIT_FAILURE );
	      }
	}
      fclose( newsrc );

      break;
    case 1:
      sprintf( line_buffer, "%s/.peruser_spool/%s.requests",
	       home_directory, newsgroup );
      if ( truncate( line_buffer, 0 ))
	if ( errno != ENOENT )
	  {
	    perror( name );
	    exit( EXIT_FAILURE );
	  }
      break;
    }

  if ( article_count )
    {
      fl_call_object_callback( fd_peruser->group_browser );
      
      fl_select_browser_line( fd_peruser->subject_browser, subject_line );
      fl_set_browser_topline( fd_peruser->subject_browser, subject_topline );
      fl_call_object_callback( fd_peruser->subject_browser );

      fl_set_browser_topline( fd_peruser->message_browser, article_line );
    }

  return;
}

void mail_callback( FL_OBJECT *obj, long caller )
{
  if ( caller )
    {
      if ( !strncmp( pop_server_address, "none", 4 ))
	{
	  fl_show_alert( "", "You must set your POP3 server's address", 
			 "in the configure form.", 1 );
	  return;
	}
      strcpy( second_line_buffer, user_email_address );
      sprintf( line_buffer, "cfetcher %s %s %s", 
	       pop_server_address, strtok( second_line_buffer, "@" ),
	       pop_password );
    }
  else
    {
      if ( !strncmp( smtp_server_address, "none", 4 ))
	{
	  fl_show_alert( "", "You must set your SMTP server's address", 
			 "in the configure form.", 1 );
	  return;
	}
      sprintf( line_buffer, "csender %s", smtp_server_address );
    }

  update_read_file();

  strcpy( second_line_buffer, line_buffer );
  run_client_program();

  article_count = 0;
  refresh_newsgroups();

  return;
}

void mark_callback( FL_OBJECT *obj, long caller )
{
    int choice;

    choice = fl_get_menu( fd_peruser->mark_menu );
    
    switch( choice )
    {
        case 1:
            pack_callback( obj, 1 );
            break;
        case 2:
            pack_callback( obj, 2 );
            break;
    }

    return;
}

void clear_requests_callback( FL_OBJECT *obj, long caller )
{
    int choice;

    choice = fl_get_menu( fd_peruser->clear_menu );

    switch( choice )
    {
        case 1:
            clear_callback( obj, 1 );
            break;
        case 2:
            clear_callback( obj, 0 );
            break;
    }

    return;
}

void compose_callback( FL_OBJECT *obj, long caller )
{
    int index;

    index = fl_get_menu( fd_peruser->compose_menu );

    switch( index )
    {
        case 1:
            run_editor( obj, 0 );
            break;
        case 2:
            address_callback( obj, 20 );
            break;
    }

    return;
}

void next_unread_callback( FL_OBJECT *obj, long caller )
{
    FILE *current_read;
    int current_group, current_subject, max_subject, old_group, old_subject,
        max_group, index, found;

    old_group = fl_get_browser( fd_peruser->group_browser );
    old_subject = fl_get_browser( fd_peruser->subject_browser );
    max_group = fl_get_browser_maxline( fd_peruser->group_browser );

    if ( !old_group )
         {
             if ( caller )
             {
                 fl_select_browser_line( fd_peruser->group_browser, 1 );
                 adjust_browser( fd_peruser->group_browser );
                 group_callback( obj, 1 );
             }
             else
                 {
                     fl_select_browser_line( fd_peruser->group_browser,
                        fl_get_browser_maxline( fd_peruser->group_browser ));
                     adjust_browser( fd_peruser->group_browser );
                     group_callback( obj, 1 );
                 }
         }
    
    current_group = fl_get_browser( fd_peruser->group_browser );
    current_subject = fl_get_browser( fd_peruser->subject_browser );
    max_subject = fl_get_browser_maxline( fd_peruser->subject_browser );

    while( current_group > group_count + 1 &&
           current_group < group_count + 7 )
        if ( caller )
            ++current_group;
        else
            --current_group;

    while( current_subject > 0 &&
           current_subject <= max_subject &&
           strncmp( fl_get_browser_line( fd_peruser->subject_browser,
                                         current_subject ),
                    "@b", 2 ))
    {
        if ( caller )
            ++current_subject;
        else
            --current_subject;
    }
    
    if ( current_subject > 0 &&
         current_subject <= max_subject )
    {
        fl_select_browser_line( fd_peruser->subject_browser,
                                current_subject );
        adjust_browser( fd_peruser->subject_browser );
        fl_call_object_callback( fd_peruser->subject_browser );
        return;
    }

    if ( caller )
        ++current_group;
    else
        --current_group;
            
    while( current_group > 0  &&
           current_group <= max_group &&
           fl_get_browser_line( fd_peruser->group_browser, current_group )
           [ 0 ] != '@' )
    {
        if ( caller )
            ++current_group;
        else
            --current_group;

        while( current_group > group_count + 1 &&
               current_group < group_count + 7 )
            if ( caller )
                ++current_group;
            else
                --current_group;
    }


    if ( current_group < 1 ||
         current_group > max_group )
    {
        fl_select_browser_line( fd_peruser->group_browser, old_group );
        adjust_browser( fd_peruser->group_browser );
        group_callback( obj, 1 );
        
        fl_select_browser_line( fd_peruser->subject_browser, old_subject );
        adjust_browser( fd_peruser->subject_browser );
        fl_call_object_callback( fd_peruser->subject_browser );

        return;
    }
    
    sprintf( line_buffer, "%s/.peruser_spool/%s.read", home_directory,
             fl_get_browser_line( fd_peruser->group_browser,
                                  current_group ) + 2 );
            
    if ( !( current_read = fopen( line_buffer, "r" )))
    {
        if ( errno != 2 )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }
    }

    if ( current_read != NULL )
    {
        found = 0;
                        
        for( index = 1; ; index++ )
        {
            fgets( line_buffer, BUFFER_SIZE, current_read );
            if ( feof( current_read ))
                break;

            if ( line_buffer[ 0 ] == 'r' )
                continue;

            found = index;

            if ( caller )
                break;
        }
        fclose( current_read );
    
        if ( found )
        {
            fl_select_browser_line( fd_peruser->group_browser,
                                    current_group );
            adjust_browser( fd_peruser->group_browser );
            fl_call_object_callback( fd_peruser->group_browser );

            fl_select_browser_line( fd_peruser->subject_browser, found );
            adjust_browser( fd_peruser->subject_browser );
            fl_call_object_callback( fd_peruser->subject_browser );
        }
    }
    
    return;
}

void grab_address_callback( FL_OBJECT *obj, long caller )
{
    char local_buffer[ BUFFER_SIZE ];
    
    int index, maxline;
    enum{ no, yes }found;

    found = no;
    maxline = fl_get_browser_maxline( fd_peruser->message_browser );

    for( index = 1; index <= maxline; index++ )
    {
        strcpy( local_buffer, fl_get_browser_line( fd_peruser->message_browser,
                                                  index ));

        if ( local_buffer[ 0 ] == '\n' )
            break;
        
        if ( !strncmp( local_buffer, "@C4From: ", 9 ) ||
             !strncmp( local_buffer, "@C4FROM: ", 9 ))
        {
            found = yes;
            break;
        }
    }
    
    if ( !found )
        {
            fl_show_alert( "ERROR:", "No From: header line found.", "", 0 );
            return;
        }
    
    address_callback( obj, 8 );
    fl_set_input( fd_address->address_input, local_buffer + 9 );
    
    return;
}

    
