
#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<fcntl.h>
#include<unistd.h>
#include<string.h>
#include<time.h>

/*
   Forms Library prototypes.
*/

#include "forms.h"

/* 
   Form structure-type definitions generated by fdesign, the graphical
   form creator that is part of the Forms Library package. Forms are
   accessed via structures of pointers, each element pointing to an
   object on the form, such as a button, a browser, etc.
*/

#include "peruser_form.h"
#include "quoter_form.h"
#include "configure_form.h"
#include "hear_form.h"
#include "regex_form.h"
#include "address_form.h"
#include "decode_form.h"
#include "newsrc_form.h"
#include "client_form.h"


/*
   Macros to define size of variables that hold lines of text.
*/

#define BUFFER_SIZE 1024

/*
   Macros used to give meaningful names to elements of the read array, which
   keeps track of whether each article in the currently selected newsgroup
   are "read" or "unread".
*/

#define UNREAD 0
#define READ 1

/*
  Used by perror.
*/

extern char name[ 7 ];

/*
   Create instances of each form structure-type definition.
*/

extern FD_peruser *fd_peruser;
extern FD_quoter *fd_quoter;
extern FD_configure *fd_configure;
extern FD_hear *fd_hear;
extern FD_regex *fd_regex;
extern FD_address *fd_address;
extern FD_decode *fd_decode;
extern FD_newsrc *fd_newsrc;
extern FD_client *fd_client;

/*
   Various file pointers.
*/

extern FILE *signature_file;
extern FILE *spool_file;
extern FILE *config_file;
extern FILE *quote_file;
extern FILE *follow_up_file;
extern FILE *follow_ups_spool_file;
extern FILE *read_file;
extern FILE *temp_file;
extern FILE *users_newsrc;
extern FILE *newsrc;
extern FILE *requests;
extern FILE *temp_requests;
extern int client_fd;
extern FILE *client;

/*
   Variables that hold lines of text.
*/

extern char line_buffer[ BUFFER_SIZE ];
extern char second_line_buffer[ BUFFER_SIZE ];
extern char address_buffer[ 256 ];

/*
   A variable that holds a line read from a newsgroup's read file,
   which stores the read array for a particular newsgroup when it is
   not the currently selected newsgroup. The read file is a text file,
   each line of which consists of two characters, which can be either,
   "u\n", to indicate an unread article, or "r\n" to indicate a read
   article. The third character of read_status is filled by the NULL
   terminator, of course.
*/

extern char read_status[ 3 ];

/*
   Variables that hold path and filename information.
*/

extern char home_directory[ BUFFER_SIZE ];
extern char spool_path[ BUFFER_SIZE ];
extern char spool_path_and_name[ BUFFER_SIZE ];
extern char follow_up_path_and_name[ BUFFER_SIZE ];
extern char quote_path_and_name[ BUFFER_SIZE ];
extern char config_path_and_name[ BUFFER_SIZE ];
extern char follow_ups_spool_path_and_name[ BUFFER_SIZE ];
extern char read_file_path_and_name[ BUFFER_SIZE ];
extern char temp_path_and_name[ BUFFER_SIZE ];
extern char requests_path[ BUFFER_SIZE ];
extern char temp_requests_path[ BUFFER_SIZE ];

/*
   A variable that holds the name of the currently selected newsgroup.
*/

extern char newsgroup[ BUFFER_SIZE ];

/*
   Variables that hold the user's configuration information.
*/

extern char user_name[ BUFFER_SIZE ];
extern char user_email_address[ BUFFER_SIZE ];
extern char editor_command[ BUFFER_SIZE ];
extern char shell_command[ BUFFER_SIZE ];
extern char server_address[ BUFFER_SIZE ];
extern char pop_server_address[ BUFFER_SIZE ];
extern char pop_password[ BUFFER_SIZE ];
extern char smtp_server_address[ BUFFER_SIZE ];
extern char nntp_username[ BUFFER_SIZE ];
extern char nntp_password[ BUFFER_SIZE ];
extern char web_browser[ BUFFER_SIZE ];
extern char ftp_client[ BUFFER_SIZE ];

/*
   A pointer that is used in the parsing and manipulation of variables that
   hold lines of text.
*/

extern char *line_pointer;

/*
   Three integers that respectively keep track of: the number of
   articles in the currently selected newsgroup, the number of
   newsgroups listed in the user's .newsrc file, the number of
   articles that have been exported from the spool files into separate
   text files. This number is used to create the filename of the
   exported article.
*/

extern int article_count;
extern int group_count;
extern int exported;
extern int authenticating;

extern int font_size, font_style;

/* 
   regex_onscreen is used to keep track of whether form fd_regex->regex is
   visible or not. Used by edit_callback to decide if it needs to
   update fd_regex->feedback_browser, by calling it's callback, to
   reflect changes to the folders' spool files.
*/

extern enum { no, yes, both } reply_by_mail, folder, configuration, user_folder,
    deleted, regex_onscreen, client_onscreen, hear_onscreen;

/*
   A pointer which is later pointed to a malloc-ed space that will
   hold the currently selected newsgroup's read array. Each element
   can be either READ or UNREAD, as #define-ed above. 
*/

extern char *read_array;

void run_editor( FL_OBJECT *obj, long caller );
