/* 
   Sender, Copyright (C) 1996, 1997 James Bailie
   ==================================================================

   Sender is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Sender is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
   License for more details.

   Although Sender is licensed under the Free Software Foundation's GNU
   General Public License, Sender is not produced by, nor is it endorsed
   by the Free Software Foundation. The Free Software Foundation is
   not responsible for developing, distributing, or supporting Sender
   in any way. Anyone may place software they own the copyright to,
   under the GNU General Public License.

   The GNU General Public License is included in the News Peruser
   distribution archive in a file called COPYING. If you do not have a
   copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   =====================================================================
*/


#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<signal.h>
#include<sys/socket.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<ctype.h>
#include<netinet/in.h>
#include<netdb.h>
#include<unistd.h>

#define BUFFER_SIZE 1024

int mail_socket, found, count;  

FILE *mail_server, *mail_file, *sent_mail, *header_file, *temp_file,
    *output;

struct protoent *protocol;
struct hostent *host;
struct servent *service;
struct sockaddr_in socket_address;

char my_host[ 128 ], line_buffer[ BUFFER_SIZE ], 
     to[ BUFFER_SIZE ], from[ BUFFER_SIZE ], header_path[ BUFFER_SIZE ];

char *line_pointer, *second_pointer, *delimiter, *server, *home_directory;
char *name = "sender";


void check_response( int expected_code  )
{
  int response_code = 0;
  
  line_buffer[ 3 ] = '\0';
  response_code = atoi( line_buffer );
  if ( response_code != expected_code )
  {
      fprintf( output, "sender: unexpected response: %d, from %s\n",
	       response_code, server );
      fclose( output );
      exit( EXIT_FAILURE );
  }
  return;
}

int main( int argc, char **argv[] )
{
    home_directory = getenv( "HOME" );

    if ( strstr( ( char *)argv[ 0 ], "csender" ))
    {
        sprintf( line_buffer, "%s/.peruser_spool/client_pipe",
                 home_directory );
        if (( output = fopen( line_buffer, "w" )) == NULL )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }
    }
    else
        output = stdout;


    if ( argc < 2 )
    {
      fprintf( output, "sender: You must supply the SMTP server's name,\n" );
      fprintf( output, "sender: as a command line argument to Sender.\n" );
      fclose( output );
      exit( EXIT_FAILURE );
    }

    server = *argv[ 1 ];

  /*
   * Open OUTBOX.
   */

  sprintf( line_buffer, "%s/.peruser_spool/OUTBOX", home_directory );
  if (( mail_file = fopen( line_buffer, "r" )) == NULL )
    {
      if ( errno != 2 )
	{
            fclose( output );
            perror( name );
            exit( EXIT_FAILURE );
	}
      fprintf( output, "sender: no outgoing mail in queue!\n" );
      fclose( output );
      exit( EXIT_FAILURE );
    }

  line_buffer[ 0 ] = fgetc( mail_file );

  if ( feof( mail_file ))
    {
      fprintf( output, "sender: no outgoing mail in queue!\n" );
      fclose( output );
      exit( EXIT_FAILURE );
    }
  
  ungetc( line_buffer[ 0 ], mail_file );

  /* 
   *Open SENT-MAIL.
   */ 

  sprintf( line_buffer, "%s/.peruser_spool/SENT-MAIL", home_directory );
  if (( sent_mail = fopen( line_buffer, "a+" )) == NULL )
      {
          fclose( output );
          perror( name );
          exit( EXIT_FAILURE );
      }

  /* 
   * Lookup the protocol number assigned to TCP. 
   */

  if ( !( protocol = getprotobyname( "tcp" )))
    {
      fprintf( output, "sender: Couldn't find a protocol number for \"tcp\"\n" );
      fprintf( output, "sender: in /etc/protocols.\n" );
      fclose( output );
      exit( EXIT_FAILURE );
    }

  /*
   * Lookup the service number assigned to SMTP.
   */

  if ( !( service = getservbyname( "smtp", "tcp" )))
    {
      fprintf( output, "sender: Couldn't find a service number for \"smtp\"\n" );
      fprintf( output, "sender: in /etc/services.\n" );
      fclose( output );
      exit( EXIT_FAILURE );
    }

  /*
   *  Lookup the address of the SMTP server named on the command line.
   */

  if ( !( host = gethostbyname( ( char *)argv[ 1 ] )))
     {
       herror( name );
       fclose( output );
       exit( EXIT_FAILURE );
     }

  /*
   * Create a socket for communication with the SMTP server.
   */

  if (( mail_socket = socket( AF_INET, SOCK_STREAM, protocol->p_proto ))
      == -1 )
    {
        fclose( output );
        perror( name );
        exit( EXIT_FAILURE );
    }

  /*
   *  Zero and initialize the socket_address structure.
   */

  memset( &socket_address, 0, sizeof socket_address );
  socket_address.sin_family = AF_INET;
  memmove( (char *)&socket_address.sin_addr, host->h_addr, host->h_length );
  socket_address.sin_port = service->s_port;

  /*
   *  Connect to the SMTP server through the socket.
   */

  if ( connect( mail_socket, ( struct sockaddr *)&socket_address, 
	        sizeof socket_address ) == -1 )
  {
      fclose( output );
      perror( name );
      exit( EXIT_FAILURE );
    }

  /*
   *  Create a stream and associate it with the socket.
   */

  if ( !( mail_server = fdopen( mail_socket, "r+" )))
  {
      fclose( output );
      perror( name );
      exit( EXIT_FAILURE );
    }

  /*
   *  Get the welcome message from the SMTP server.
   */

  fgets( line_buffer, BUFFER_SIZE, mail_server );
  fprintf( output, line_buffer );
  fflush( output );
  check_response( 220 );

  strcpy( line_buffer, "HELO localhost\r\n" );
  fputs( line_buffer, mail_server );
  fprintf( output, "sender:   %s", line_buffer );
  fflush( output );
    
  fgets( line_buffer, BUFFER_SIZE, mail_server );
  fprintf( output, "server: %s", line_buffer );
  fflush( output );
  
  check_response( 250 );

  /*
   * main loop.
   */

  sprintf( header_path, "%s/.peruser_spool/.header", home_directory );

  for ( ; ; )
    {

      /*
       * get header info.
       */

      if (( header_file = fopen( header_path, "w+" )) == NULL )
	{
            fclose( output );
            perror( name );
            exit( EXIT_FAILURE );
	}

      for( ; ; )
	{
	  sprintf( line_buffer, "%s/.peruser_spool/.temp", home_directory );
	  if (( temp_file = fopen( line_buffer, "w+" )) == NULL )
	    {
                fclose( output );
                perror( name );
                exit( EXIT_FAILURE );
	    }

	  fgets( line_buffer, BUFFER_SIZE, mail_file );
	  if ( !strncmp( line_buffer, "\r\n", 2 ) ||
	       feof( mail_file ))
	    break;
	  
	  fputs( line_buffer, header_file );

	  if ( !strncmp( line_buffer, "To: ", 4 ))
              {
                  strcpy( to, strtok( line_buffer, "\r" ));
                  while( to[ strlen( to ) - 1 ] == ',' )
                  {
                      fgets( line_buffer, BUFFER_SIZE, mail_file );
                      fputs( line_buffer, header_file );
                      strcat( to, strtok( line_buffer, "\r" ));
                  }
              }
          
	  if ( !strncmp( line_buffer, "From: ", 6 ))
              strcpy( from, strtok( line_buffer, "\r" ));
	}

      fflush( header_file );

      if ( feof( mail_file ))
	{
	  fclose( header_file );
	  break;
	}

      /*
       * Skip messages where the user has forgotten to fill in the 
       * receiver's email address.
       */

      if ( !strncmp( to, "To: (replace with address)", 26 )) 
	{
	  fprintf( output,
                   "sender: skipping message in queue without To: address\n" );
	  do
	    {
	      fgets( line_buffer, BUFFER_SIZE, mail_file );
	    }
	  while( strncmp( line_buffer, ".\r\n", 3 ));
	  continue;
	}

      /*
       * give the MAIL command to SMTP server.
       */

      if ( ( line_pointer = strrchr( from, '<' )) != NULL )
	delimiter = ">";
      else
	{
	  line_pointer = strchr( to, ' ' );
	  if ( strchr( to, '(' ))
	    delimiter = " (";
	  else
	    delimiter = "\r\n";
	}

      line_pointer++;

      sprintf( line_buffer, "MAIL FROM: <%s>\r\n", strtok( line_pointer, delimiter ));

      fputs( line_buffer, mail_server );
      fprintf( output, "sender: %s", line_buffer );
      fflush( output );
      
      fgets( line_buffer, BUFFER_SIZE, mail_server );
      fprintf( output, "server: %s", line_buffer );
      fflush( output );
      check_response( 250 );

      /*
       * give the RCPT commands to the server
       */

      strcat( to, "," );
      line_pointer = strtok( to + 4, "," );

      do
	{
	  if ( ( second_pointer = strchr( line_pointer, '<' )) != NULL )
	    {
	      second_pointer++;
	      delimiter = ">";
	    }
	  else
	    {
	      second_pointer = line_pointer + strspn( line_pointer, " \r\n" );

	      if ( strchr( line_pointer, '(' ))
		delimiter = "(";
	      else
		delimiter = NULL;
	    }

	  strcpy( line_buffer, "RCPT TO: <" );

	  if ( delimiter == NULL )
	    count = strlen( second_pointer );
	  else
	    count = strstr( second_pointer, delimiter ) - second_pointer;

	  strncat( line_buffer, second_pointer, count );
	  line_buffer[ 10 + count ] = '\0';
	  strcat( line_buffer, ">\r\n" );

	  fputs( line_buffer, mail_server );
	  fprintf( output, "sender: %s", line_buffer );
          fflush( output );

	  fgets( line_buffer, BUFFER_SIZE, mail_server );
	  fprintf( output, "server: %s", line_buffer );
          fflush( output );
	  check_response( 250 );
	}
      while( ( line_pointer = strtok( NULL, "," )) != NULL );

      /*
       * Send the message.
       */

      strcpy( line_buffer, "DATA\r\n" );
      fputs( line_buffer, mail_server );
      fprintf( output, "sender: %s", line_buffer );
      fflush( output );
      
      fgets( line_buffer, BUFFER_SIZE, mail_server );
      fprintf( output, "server: %s", line_buffer );
      fflush( output );
      check_response( 354 );

      fprintf( output, "        (sending message...)\n" );
      fflush( output );

      rewind( header_file );
      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, header_file );
	  if ( feof( header_file ))
	    {
	      fclose( header_file );
	      break;
	    }

	  fputs( line_buffer, mail_server );
	  fputs( line_buffer, temp_file );
	}

      fputs( "\r\n", mail_server );
      fputs( "\r\n", temp_file );

      do
	{
	  fgets( line_buffer, BUFFER_SIZE, mail_file );
	  fputs( line_buffer, mail_server );
	  fputs( line_buffer, temp_file );
	}
      while( strncmp( line_buffer, ".\r\n", 3 ));

      fgets( line_buffer, BUFFER_SIZE, mail_server );
      fprintf( output, "server: %s", line_buffer );
      fflush( output );
      check_response( 250 );

      rewind( temp_file );
      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, temp_file );
	  if ( feof( temp_file ))
	    break;
	  
	  fputs( line_buffer, sent_mail );
	}
      fclose( temp_file );

      fprintf( output, "\n" );
      fflush( output );
    }

  /*
   * Parting is such sweet sorrow...
   */

  fflush( sent_mail );
  fclose( sent_mail );
  fclose( mail_file );
  sprintf( line_buffer, "%s/.peruser_spool/OUTBOX", home_directory );
  truncate( line_buffer, 0 );
  sprintf( line_buffer, "%s/.peruser-spool/.temp", home_directory );
  remove( line_buffer );

  strcpy( line_buffer, "QUIT\r\n" );
  fputs( line_buffer, mail_server );
  fprintf( output, "sender: %s", line_buffer );
  fflush( output );
  
  fgets( line_buffer, BUFFER_SIZE, mail_server );
  fprintf( output, "server: %s", line_buffer );
  
  fclose( mail_server );
  shutdown( mail_socket, 2 );
  close( mail_socket );

  fflush( output );
  fclose( output );
  
  return( 0 );
}
