/* 
   Sorter, Copyright (C) 1995, 1996, 1997 James Bailie
   ==================================================================

   Sorter is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2, or (at your
   option) any later version.

   Sorter is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   Although Sorter is licensed under the Free Software Foundation's
   GNU General Public License, Sorter is not produced by, nor is it
   endorsed by the Free Software Foundation. The Free Software
   Foundation is not responsible for developing, distributing, or
   supporting Sorter in any way. Anyone may place software they own
   the copyright to, under the GNU General Public License.

   The GNU General Public License is included in the News Peruser
   distribution archive in a file called COPYING. If you do not have a
   copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   ===================================================================== */

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<signal.h>
#include<ctype.h>
#include<sys/wait.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<unistd.h>

#define BUFFER_SIZE 1024

char *name = "sorter";
FILE *output, *client;

pid_t pid;

char spool_path_and_name[ BUFFER_SIZE ];
char temp_spool_path_and_name[ BUFFER_SIZE ];
char read_path_and_name[ BUFFER_SIZE ];
char temp_read_path_and_name[ BUFFER_SIZE ];
char newsgroup[ BUFFER_SIZE ];

FILE *spool_file;
FILE *read_file;
FILE *temp_read_file;
FILE *temp_spool_file;
FILE *newsrc;

void signal_handler( int signal )
{
    if ( spool_file )
        fclose( spool_file );
    
    if ( read_file )
        fclose( read_file );

    if ( temp_spool_file )
    {
        fclose( temp_spool_file );
        remove( temp_spool_path_and_name );
    }

    if ( temp_read_file )
    {
        fclose( temp_read_file );
        remove( temp_read_path_and_name );
    }

    if ( newsrc )
        fclose( newsrc );

    if ( client )
        fclose( client );
    
    if ( pid )
        {
            if ( kill( pid, SIGINT ))
                kill( pid, SIGKILL );

            waitpid( pid, NULL, WNOHANG | WUNTRACED );
        }

    fclose( output );
    
    exit( signal );
}

int compare( const void *a, const void *b )
{
  return( strcmp( *(char **)a, *( char **)b ));
}

int main( int argc, char **argv[] )
{
    char line_buffer[ BUFFER_SIZE ];
    char home_directory[ BUFFER_SIZE ];

    char *line_pointer;

    struct subject_type
    {
        char *line;
        unsigned long int position;
    };

    struct subject_type *subject;

    unsigned long int article_count, index, second_index, length;


    signal( SIGTERM, signal_handler );
    signal( SIGINT, signal_handler );
    signal( SIGHUP, signal_handler );
    
    strcpy( home_directory, getenv( "HOME" ));

    if ( strstr( ( char *)argv[ 0 ], "csorter" ))
    {
        sprintf( line_buffer, "%s/.peruser_spool/sorter_pipe",
                 home_directory );
        if (( output = fopen( line_buffer, "w" )) == NULL )
        {
            perror( name );
            exit( EXIT_FAILURE );
        }
    }
    else
        output = stdout;
  
/*
  Find and open the user's .newsrc.
*/
  
    sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );

    if (( newsrc = fopen( line_buffer, "r" )) == NULL )
    {
        perror( name );
        fclose( output );
        exit( EXIT_FAILURE );
    }

/*
  Read a newsgroup from .newsrc, open appropriate read file.
*/

    for( ; ; )
    {
        fgets( line_buffer, BUFFER_SIZE, newsrc );
        if ( feof( newsrc ))
            break;

        if ( !strncmp( line_buffer, "\n", 1 ))
            continue;
      
        strcpy( newsgroup, strtok( line_buffer, ":" ));

        fprintf( output, "sorter:   sorting %s...\n", newsgroup );
        fflush( output );

        sprintf( read_path_and_name, "%s/.peruser_spool/%s.read", 
                 home_directory, newsgroup );

        if (( read_file = fopen( read_path_and_name, "r" )) == NULL )
	{
            if ( errno == 2 )
                continue;
            fclose( output );
            perror( name );
            exit( EXIT_FAILURE );
	}
      
/*
  Count articles in spool_file, by counting lines in corresponding read
  file.
*/

        for( article_count = -1; !feof( read_file ); article_count++ )
            fgets( line_buffer, BUFFER_SIZE, read_file );

        if ( !article_count )
	{
            fclose( read_file );
            continue;
	}
  
/*
  Open newsgroup's spool file.
*/

        sprintf( spool_path_and_name, "%s/.peruser_spool/%s", home_directory,
                 newsgroup );

        if (( spool_file = fopen( spool_path_and_name, "r" )) == NULL )
	{
            perror( name );
            fclose( output );
            exit( EXIT_FAILURE );
	}

/* 
   Create array of structures to hold pointers to subject lines and
   article's place in spool file.
*/

        if (( subject = calloc( article_count, sizeof *subject )) == NULL )
	{
            fprintf( output, "sorter: calloc failed.\n" );
            fclose( output );
            exit( EXIT_FAILURE );
	}

/*
  Load element of array from spool file.
*/

        for( index = 0; index < article_count; index++ )
	{
            do
	    {
                fgets( line_buffer, BUFFER_SIZE, spool_file );
	    }
            while( strncmp( line_buffer, "Subject: ", 9 ));
            subject[ index ].position = index;

/*
  Advance past the characters "Subject: ".
*/
	  
            if (( subject[ index ].line = 
                  malloc( strlen( line_buffer + 9 ) + 3 )) == NULL )
            {
                perror( name );
                fclose( output);
                exit( EXIT_FAILURE );
            }
            strcpy( subject[ index ].line, line_buffer + 9 );
	  
/*
  Advance spool file to end of article, in case the string "Subject: "
  occurs in the article body.
*/

            do
	    {
                fgets( line_buffer, BUFFER_SIZE, spool_file );
	    }
            while( strncmp( line_buffer, ".\r\n", 3 ));
	}

/* 
   Remove "Re: " from the beginning of any lines, concatenating a 'a' to
   the end of those lines, to ensure they will be sorted just behind the
   root messages, if present in the spool. This loop Also converts all
   capitals to lower-case.
*/

        for( index = 0; index < article_count; index++ )
	{
            if ( !strncmp( subject[ index ].line, "Re: ", 4 ))
	    {
                strcpy( line_buffer, subject[ index ].line + 4 );
                line_pointer = strrchr( line_buffer, '\r' );
                strcpy( line_pointer, " a\r\n" );
                strcpy( subject[ index ].line, line_buffer );
	    }

            length = strlen( subject[ index ].line );
            for( second_index = 0; second_index < length; second_index++ ) 
	    {
                subject[ index ].line[ second_index ] =
                    tolower( subject[ index ].line[ second_index ] );
	    }
	}

/*
  Sort the array. Compare will stop comparing at the NULL terminators
  of subject[ n ].line, making this the sort key. Passing element_size
  as the third argument will ensure entire structures are sorted.
*/

        qsort( subject, article_count, sizeof *subject, compare );

/* 
   Open temporary spool file.
*/

        sprintf( temp_spool_path_and_name, "%s/.peruser_spool/spool.temp", 
                 home_directory );

        if (( temp_spool_file = fopen( temp_spool_path_and_name, "w" )) == NULL )
	{
            perror( name );
            fclose( output );
            exit( EXIT_FAILURE );
	}

/*
  Open temporary read file.
*/


        sprintf( temp_read_path_and_name, "%s/.peruser_spool/read.temp",
                 home_directory );

        if (( temp_read_file = fopen( temp_read_path_and_name, "w" )) == NULL )
	{
            perror( name );
            fclose( output );
            exit( EXIT_FAILURE );
	}
     
/*
  Copy spool_file's articles over to temporary_file in sorted order.
  Copy read_file over to temp_read_file in sorted order.
*/

        for( index = 0; index < article_count; index++ )
	{
            rewind( spool_file );
            rewind( read_file );

            free( subject[ index ].line );

/*
  Advance spool file and read file to current article.
*/

            while( subject[ index ].position )
	    {
                do
		{
                    fgets( line_buffer, BUFFER_SIZE, spool_file );
		}
                while( strncmp( line_buffer, ".\r\n", 3 ));
                fgets( line_buffer, BUFFER_SIZE, read_file );
                subject[ index ].position--;
	    }

/*
  copy over current article.
*/
            do
	    {
                fgets( line_buffer, BUFFER_SIZE, spool_file );
                fputs( line_buffer, temp_spool_file );
	    }
            while( strncmp( line_buffer, ".\r\n", 3 ));
            fgets( line_buffer, BUFFER_SIZE, read_file );
            fputs( line_buffer, temp_read_file );
	}

/*
  Clean up.
*/

        free( subject );

        fflush( temp_spool_file );
        fclose( temp_spool_file );

        fflush( temp_read_file );
        fclose( temp_read_file );

        fclose( read_file );

        fclose( spool_file );

        remove( spool_path_and_name );
        rename( temp_spool_path_and_name, spool_path_and_name );

        remove( read_path_and_name );
        rename( temp_read_path_and_name, read_path_and_name );
    } 

    fclose( newsrc );
    fprintf( output, "sorter: finished.\n" );
    fflush( output );
  
    if ( output == stdout )
        execlp( "threader", "threader", NULL );
    else
    {
        sprintf( line_buffer, "%s/.peruser_spool/threader_pipe",
                 home_directory );
        if ( mkfifo( line_buffer, 0700 ))
        {
            if ( errno != EEXIST )
            {
                fclose( output );
                perror( name );
                exit( EXIT_FAILURE );
            }
        }

        pid = 0;
        switch( pid = fork() )
        {
            case -1:
                fclose( output );
                perror( name );
                exit( EXIT_FAILURE );
                break;
            case 0:
                execlp( "cthreader", "cthreader", NULL );

                perror( name );
                exit( EXIT_FAILURE );
                break;
            default:
                if (( client = fopen( line_buffer, "r" )) == NULL )
                {
                    fclose( output );
                    perror( name );
                    exit( EXIT_FAILURE );
                }

                for( ; ; )
                {
                    fgets( line_buffer, BUFFER_SIZE, client );
                    if ( feof( client ))
                        break;
                  
                    fputs( line_buffer, output );
                    fflush( output );
                }

                fclose( client );
                fclose( output );

                waitpid( pid, NULL, WNOHANG ^ WUNTRACED );
                pid = 0;
                exit( EXIT_SUCCESS );
                break;
        }
    }
  
    fclose( output );
    perror( name );
    exit( EXIT_FAILURE );
}

