/* 
   Threader, Copyright (C) 1997 James Bailie
   ==================================================================

   Threader is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Threader is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   Although Threader is licensed under the Free Software Foundation's
   GNU General Public License, Threader is not produced by, nor is it
   endorsed by the Free Software Foundation. The Free Software
   Foundation is not responsible for developing, distributing, or
   supporting Threader in any way. Anyone may place software they own
   the copyright to, under the GNU General Public License.

   The GNU General Public License is included in the News Peruser
   distribution archive in a file called COPYING. If you do not have a
   copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   =====================================================================
*/

/*
 * This is a C program thinly disguised as a C++ program.
 */

#include<stdlib.h>
#include<stdio.h>
#include<string.h>
#include<signal.h>

#define BUFFER_SIZE 1024


char line_buffer[ BUFFER_SIZE ], newsgroup[ BUFFER_SIZE ],
    spool_path[ BUFFER_SIZE ], temp_spool_path[ BUFFER_SIZE ],
    read_path[ BUFFER_SIZE ], temp_read_path[ BUFFER_SIZE ];

FILE *newsrc, *spool_file, *temp_file, *read_file, *temp_read_file, *output;

void signal_handler( int signal )
{
    if ( spool_file )
        fclose( spool_file );

    if ( read_file )
        fclose( read_file );

    if ( temp_file )
    {
        fclose( temp_file );
        remove( temp_spool_path );
    }

    if ( temp_read_file )
    {
        fclose( temp_read_file );
        remove( temp_read_path );
    }

    if ( newsrc )
        fclose( newsrc );

    fclose( output );
    
    exit( signal );
}

void allocation_failure()
{
  fprintf( stderr, "threader: new failed\n" );
  fclose( output );
  exit( EXIT_FAILURE );
}

typedef struct
{
  char *current_ref;
  void *next_ref;
  void *prev_ref;
} list;

class node
{
public:
  long unsigned int position;
  long unsigned int gap;

  char *message_id;
  list *references;

  bool queued;
  node *queue_head;
  node *queue_tail;
  node *queue_next;

  node *next_node;
  node *prev_node;

  void save_node( long unsigned int depth );

  node( long unsigned int my_pos, char *my_id, char *refs_line );
  ~node();

private:
  char *get_reference( char *string );
};

/*
 * Returns char pointer to one complete Message-ID from the References 
 * header line passed to it. Successive calls with NULL as argument return
 * successive Message-IDs. Returns NULL when References is exhausted.
 */

char *node::get_reference( char *string )
{
  char *test;
  static char *token, *remainder, *temp;
  size_t length;

  if ( string )
    {
      if (( token = new char[ strlen( string ) + 1 ] ) == NULL )
	allocation_failure();
      if (( remainder = new char[ strlen( string ) + 1 ] ) == NULL )
	allocation_failure();
      if (( temp = new char[ strlen( string ) + 1 ] ) == NULL )
	allocation_failure();

      test = strchr( string, '<' );
    }
  else
    test = strchr( remainder, '<' );

  if ( test )
    {
      strcpy( temp, test );
      length = strcspn( temp, " \t\r\n" );
      strncpy( token, temp, length );
      token[ length ] = '\0';
      strcpy( remainder, temp + length );
      return( token );	
    }
  else
    {
      delete [] token;
      delete [] remainder;
      delete [] temp;
      return( NULL );
    }
}

/*
 * Rewrite current nodes's corresponding article to new spool file, 
 * pseudo-recursing through any (sub)queues.
 */

void node::save_node( long unsigned int depth )
{
  node *local_pointer;
  char local_buffer[ BUFFER_SIZE ], subject[ BUFFER_SIZE ];
  long unsigned int index;

  rewind( spool_file );
  rewind( read_file );

  while( --position )
    {
      do
	{
	  fgets( local_buffer, BUFFER_SIZE, spool_file );
	}
      while( strncmp( local_buffer, ".\r\n", 3 ));

      fgets( local_buffer, BUFFER_SIZE, read_file );
    }

  for( ; ; )
    {
      fgets( local_buffer, BUFFER_SIZE, spool_file );

      if ( strncmp( local_buffer, "Subject: ", 9 ))
	fputs( local_buffer, temp_file );
      else
	break;
    }

  strcpy( subject, 
	  local_buffer + 9 + strspn( local_buffer + 9, " \t" ));

  local_buffer[ 9 ] = '\0';

  index = depth + gap;

  while( index-- )
    strcat( local_buffer, "\t" );

  strcat( local_buffer, subject );

  fputs( local_buffer, temp_file );

  do
    {
      fgets( local_buffer, BUFFER_SIZE, spool_file );
      fputs( local_buffer, temp_file );
    }
  while( strncmp( local_buffer, ".\r\n", 3 ));

  fgets( local_buffer, BUFFER_SIZE, read_file );
  fputs( local_buffer, temp_read_file );

  local_pointer = queue_head;

  if ( local_pointer )
    ++depth;

  while( local_pointer )
    {
      local_pointer->save_node( depth + gap );
      local_pointer = local_pointer->queue_next;
    }

  return;
}

node::node( long unsigned int my_pos, char *my_id, char *my_refs )
{
  char *one_ref;
  list *list_pointer, *previous;
  
  position = my_pos;

  gap = 0;

  queued  = false;
  queue_head = queue_tail = queue_next = NULL;

  if (( message_id = new char[ strlen( strtok( my_id, "\r\n" )) + 1 ] ) 
      == NULL )
    allocation_failure();
  strcpy( message_id, my_id );

  /*
   * Calls get_reference to create linked list
   * of references from my_refs.
   */

  if ( my_refs && ( one_ref = get_reference( my_refs )))
    {
      if (( references = new list ) == NULL )
	allocation_failure();

      if (( references->current_ref = new char[ strlen( one_ref ) + 1 ] ) == NULL )
	allocation_failure();

      strcpy( references->current_ref, one_ref );

      references->prev_ref = NULL;

      list_pointer = references;

      while( ( one_ref = get_reference( NULL )) != NULL )
	{
	  if (( list_pointer->next_ref = new list ) == NULL )
	    allocation_failure();

	  previous = list_pointer;
	  list_pointer = ( list *)list_pointer->next_ref;
	  list_pointer->prev_ref = previous;

	  if (( list_pointer->current_ref = new char[ strlen( one_ref ) + 1 ] )
	      == NULL )
	    allocation_failure();

	  strcpy( list_pointer->current_ref, one_ref );
	}

      list_pointer->next_ref = NULL;
    }
  else
    references = NULL;

  return;
}

node::~node()
{ 
  list *list_pointer, *previous;

  delete [] message_id;

  /*
   * Destroy the linked list of references, if it exists.
   */

  if ( references )
    {
      list_pointer = references;

      while( list_pointer->next_ref )
	list_pointer = ( list *)list_pointer->next_ref;

      do
	{
	  previous = ( list *)list_pointer->prev_ref;

	  delete [] list_pointer->current_ref;
	  delete list_pointer;

	  list_pointer = previous;
	}
      while( list_pointer );
    }

  return;
}

int main( char **argv[] )
{ 
  char *home_directory, *current_message_id, 
       *current_references, *references_buffer;

  const char *name = "threader";

  bool last_line_was_reference_header;

  long unsigned int counter, gap_count;

  list *references_pointer;

  node *article_list, *current_article, *previous_article, *other_article;


  home_directory = getenv( "HOME" );

  if ( strstr( ( char *)argv[ 0 ], "cthreader" ))
      {
          sprintf( line_buffer, "%s/.peruser_spool/threader_pipe",
                   home_directory );
          if (( output = fopen( line_buffer, "w" )) == NULL )
          {
              perror( name );
              exit( EXIT_FAILURE );
          }
      }
  else
      output = stdout;
  
  /*
   * Open ~/.peruser-newsrc, get a newsgroup name and open its spool_file file.
   */

  sprintf( line_buffer, "%s/.peruser-newsrc", home_directory );

  if (( newsrc = fopen( line_buffer, "r" )) == NULL )
    {
      perror( name );
      fclose( output );
      exit( EXIT_FAILURE );
    }

  for( ; ; )
    {
      fgets( line_buffer, BUFFER_SIZE, newsrc );
      if ( feof( newsrc ))
	break;
      if ( line_buffer[ 0 ] == '\n' )
	continue;

      sprintf( spool_path, "%s/.peruser_spool/%s", 
	       home_directory, strcpy( newsgroup, strtok( line_buffer, ":" )));

      fprintf( output, "threader: threading %s...\n", newsgroup );
      fflush( output );
      
      if (( spool_file = fopen( spool_path, "r" )) == NULL )
	{
	  if ( errno == 2 )
	    continue;
	  else
	    {
                fclose( output );
                perror( name );
                exit( EXIT_FAILURE );
	    }
	}

      /* Search for each article's Message-ID, and References header
       * lines, stopping at the blank line separating the header from
       * the article body, or when the end of the article is reached,
       * which will happen if the the current article is is a header,
       * and it has no references (There will be no blank line before
       * the body because there will be no body.). */

      article_list = NULL;
      counter = 0;
      current_article = previous_article = NULL;
      last_line_was_reference_header = false;
      
      do
	{
	  current_message_id = current_references = NULL;
	  
	  for( ; ; )
	    {
	      fgets( line_buffer, BUFFER_SIZE, spool_file );
	  	    
	      if ( feof( spool_file )
		   || !strncmp( line_buffer, "\r\n", 2 )
		   || !strncmp( line_buffer, ".\r\n", 3 ))
		break;

	      if ( !strncmp( line_buffer, "\t", 1 ) && 
		   last_line_was_reference_header )
		{
		  if (( references_buffer = 
			new char[ strlen( line_buffer ) +
				strlen( current_references ) + 1 ]) != NULL )
		    {

		      strcpy( references_buffer, 
			      strtok( current_references, "\r\n" ));
		      strcat( references_buffer, line_buffer );
		  
		      delete current_references;
		    }
		  else
		    allocation_failure();

		  if (( current_references = 
			new char[ strlen( references_buffer ) + 1 ]) != NULL )
		    {
		      strcpy( current_references, references_buffer );

		      delete references_buffer;
		      continue;
		    }
		  else
		    allocation_failure();
		}
	      
	      last_line_was_reference_header = false;

	      if ( !strncmp( line_buffer, "Message-ID: ", 12 ) ||
                   !strncmp( line_buffer, "Message-Id: ", 12 ))
		if (( current_message_id = 
		      new char[ strlen( line_buffer + 12 ) + 1 ]) != NULL )
		  {
		    strcpy( current_message_id, line_buffer + 12 );
		    continue;
		  } 
		else
		  allocation_failure();

	      if ( !strncmp( line_buffer, "References: ", 12 ))
		if (( current_references = 
		      new char[ strlen( line_buffer + 12 ) + 1 ]) != NULL )
		  {
		    strcpy( current_references, line_buffer + 12 );
		    last_line_was_reference_header = true;
		    continue;
		  }
		else
		  allocation_failure();
	    }

	  /*
	   * If we found an article, create node object for it, and add it
	   * to linked list.
	   */

	  if ( !feof( spool_file ))
	    {
	      /*
	       * If the current location in the spool file is at the blank line
	       * separating header from body, then advance until end of article
	       * is found.
	       */

	      while( strncmp( line_buffer, ".\r\n", 3 ))
		fgets( line_buffer, BUFFER_SIZE, spool_file );
	  
	      counter++;

	      if ( current_article )
		previous_article = current_article;

	      if (( current_article = new node( counter, 
						current_message_id, 
						current_references )) == NULL ) 
		allocation_failure();

	      if ( !article_list )
		article_list = current_article;

	      current_article->prev_node = previous_article;
	      current_article->next_node = NULL;

	      if ( previous_article )
		previous_article->next_node = current_article;

	      delete [] current_message_id;
	      delete [] current_references;
	    }
	}
      while( !feof( spool_file ));

      /*
       * If there's a list, thread it.
       */

      if ( !article_list )
	continue;

      for( current_article = article_list;
	   current_article;
	   current_article = current_article->next_node )
	{
	  if ( !( current_article->references ))
	    continue;

	  references_pointer = current_article->references;

	  while( references_pointer->next_ref )
	    references_pointer = ( list *)references_pointer->next_ref;
		 
	  for( gap_count = 0;
	       references_pointer;
	       references_pointer = ( list *)references_pointer->prev_ref )
	    {
	      for( other_article = article_list ;
		   other_article;
		   other_article = other_article->next_node )
		if ( other_article == current_article )
		  continue;
		else
		  if ( !strcmp( other_article->message_id,
				references_pointer->current_ref ))
		    {
		      if ( other_article->queue_head != NULL )
			if ( gap_count >
			     other_article->queue_head->gap )
			  {
			    previous_article = other_article->queue_head;
			    other_article->queue_head = current_article;
			    current_article->queue_next = previous_article;
			  }
			else
			  {
			    other_article->queue_tail->queue_next = 
			      current_article;
			    other_article->queue_tail = current_article;
			  }
		      else
			other_article->queue_head = 
			  other_article->queue_tail = current_article;

		      current_article->queued = true;
		      current_article->gap = gap_count;

		      break;
		    }
	      if ( current_article->queued )
		break;

	      ++gap_count;
	    }
	}

      /*
       * Rewrite spool
       */

      sprintf( temp_spool_path, "%s.temp", spool_path );
      if (( temp_file = fopen( temp_spool_path, "w" )) == NULL )
	{
            fclose( output );
            perror( name );
            exit( EXIT_FAILURE );
	}

      sprintf( read_path, "%s.read", spool_path );
      if (( read_file = fopen( read_path, "r" )) == NULL )
	{
            fclose( output );
            perror( name );
            exit( EXIT_FAILURE );
	}

      sprintf( temp_read_path, "%s.temp", read_path );
      if (( temp_read_file = fopen( temp_read_path, "w" )) == NULL )
	{
            fclose( output );
            perror( name );
            exit( EXIT_FAILURE );
	}

      for( current_article = article_list; 
	   current_article;
	   current_article = current_article->next_node )
	{
	  if ( current_article->queued )
	    continue;

	  current_article->save_node( 0 );
	}

      fflush( temp_file );
      fclose( temp_file );
      fclose( spool_file );

      remove( spool_path );
      rename( temp_spool_path, spool_path );

      fflush( temp_read_file );
      fclose( temp_read_file );
      fclose( read_file );

      remove( read_path );
      rename( temp_read_path, read_path );

      /*
       * Delete the list, working backwards. We could have deleted the
       * list a recursive function and elimated the previous_article
       * pointer, but we'd likely overflow the stack, if the user had,
       * say, a thousand articles in the spool.
       */

      if ( article_list )
	{
	  current_article = article_list;

	  while( current_article->next_node )
	    current_article = current_article->next_node;

	  do
	    {
	      previous_article = current_article->prev_node;

	      delete current_article;

	      current_article = previous_article;
	    }
	  while( current_article );
	}
    }

  fprintf( output, "threader: finished.\n" );
  fflush( output );
  if ( output != stdout )
      fclose( output );
    
  return( 0 );
}

