/* 
   Viewer, Copyright (C) 1997 James Bailie
   ==================================================================

   Viewer is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Viewer is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   Although Viewer is licensed under the Free Software Foundation's
   GNU General Public License, Viewer is not produced by, nor is it
   endorsed by the Free Software Foundation. The Free Software
   Foundation is not responsible for developing, distributing, or
   supporting Viewer in any way. Anyone may place software they own
   the copyright to, under the GNU General Public License.

   The GNU General Public License is included in the News Peruser
   distribution archive in a file called COPYING. If you do not have a
   copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   =====================================================================
*/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>

#define BUFFER_SIZE 1024

void malformed_line( char *line, char *file )
{
  fprintf( stderr, "Viewer: malformed line: %s\n", 
	   strtok( line, "\n" ) ? line : line );
  fprintf( stderr, "Viewer: in %s\n", file );

  exit( EXIT_FAILURE );
}

int main( int argc, char *argv[] )
{
  FILE *m_types, *m_cap;

  char line_buffer[ BUFFER_SIZE ], format [ BUFFER_SIZE ];
  char *home_directory, *copy_buffer, *token, *suffix_token, *position;

  char *name = "Viewer";

  int index, place;

  typedef struct
    {
      char *suffix;
      void *next;
      void *previous;
  } suffix_type;

  typedef struct
    {
      char *type;
      suffix_type *suffixes;
      void *next;
      void *previous;
  } mime_types;

  mime_types *current_type, *other_type, *types_pointer;

  suffix_type *current_suffix, *other_suffix;

  typedef struct
    {
      char *type;
      char *player;
      void *next;
      void *previous;
  } mailcap;

  mailcap *current_mailcap, *other_mailcap, *mailcap_pointer;

  bool no_user_types, no_system_types, no_user_mailcap, no_system_mailcap, 
    found;

  no_user_types = no_system_types = no_user_mailcap = no_system_mailcap = 
    found = false;

  current_type = NULL;
  current_suffix = NULL;
  current_mailcap = NULL;

  /*
   * Check for arguments.
   */

  if ( argc < 2 )
    {
      fprintf( stderr, "Viewer: no filenames passed as parameters\n" );
      exit( EXIT_FAILURE );
    }

  /*
   * Create linked list of structures pairing mime types with filename
   * suffixes according to /etc/mime.types and ~/.mime.types, with
   * ~/.mime.types overriding entries in /etc/mime.types.
   */

  home_directory = getenv( "HOME" );

  sprintf( line_buffer, "%s/.mime.types", home_directory );
  if (( m_types = fopen( line_buffer, "r" )) == NULL )
    if ( errno != 2 )
      {
	perror( name );
	exit( EXIT_FAILURE );
      }
    else
      no_user_types = true;

  types_pointer = NULL;

  if ( !no_user_types )
    {
      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, m_types );
	  if ( feof( m_types ))
	    break;

	  if ( line_buffer[ 0 ] == '\n' || line_buffer[ 0 ] == '#' )
	    continue;

	  copy_buffer = new char[ strlen( line_buffer ) + 2 ];
	  strcpy( copy_buffer, line_buffer );

	  if ( !( token = strtok( line_buffer, "\t " )))
	    malformed_line( copy_buffer, "~/.mime.types" );

	  if ( !types_pointer)
	    {
	      types_pointer = new mime_types;
	      current_type = types_pointer;
	      current_type->previous = NULL;
	    }
	  else
	    {
	      current_type->next = new mime_types;
	      (( mime_types *)current_type->next )->previous = current_type;
	      current_type = ( mime_types *)current_type->next;
	    }
      
	  current_type->type = new char[ strlen( token ) + 2 ];

	  strcpy( current_type->type, token );

	  if ( !( token = strtok( NULL, "\n" )))
	    malformed_line( copy_buffer, "~/.mime.types" );

	  strcat( token, "\n" );
	  
	  current_type->suffixes = NULL;

	  for( suffix_token = strtok( token, " \t\n" );
	       suffix_token;
	       suffix_token = strtok( NULL, " \t\n" ))
	    {
	      if ( !( current_type->suffixes ))
		{
		  current_type->suffixes = new suffix_type;
		  current_suffix = current_type->suffixes;
		  current_suffix->previous = NULL;
		}
	      else
		{
		  current_suffix->next = new suffix_type;
		  (( suffix_type *)current_suffix->next )->previous
		    = current_suffix;
		  current_suffix = ( suffix_type *)current_suffix->next;
		}

	      current_suffix->suffix = new char[ suffix_token + 2 ];
	      strcpy( current_suffix->suffix, suffix_token );
	      current_suffix->next = NULL;
	    }
	}
      fclose( m_types );
      current_type->next = NULL;
    }

  if (( m_types = fopen( "/etc/mime.types", "r" )) == NULL )
    if ( errno != 2 )
      { 
	perror( name );
	exit( EXIT_FAILURE );
      }
    else
      if ( no_user_types )
	{
	  fprintf( stderr, 
		   "Viewer: no /etc/mime.types or ~/.mime.types found\n" );
	  exit( EXIT_FAILURE );
	}
      else
	no_system_types = true;

  if ( !no_system_types )
    {
      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, m_types );
	  if ( feof( m_types ))
	    break;

	  if ( line_buffer[ 0 ] == '\n' || line_buffer[ 0 ] == '#' )
	    continue;

	  copy_buffer = new char[ strlen( line_buffer ) + 2 ];
	  strcpy( copy_buffer, line_buffer );

	  if ( !( token = strtok( line_buffer, "\t " )))
	    malformed_line( copy_buffer, "/etc/mime.types" );

	  if ( types_pointer )
	    {
	      for( other_type = types_pointer; 
		   other_type;
		   other_type = ( mime_types *)other_type->next )
		if ( !strcmp( other_type->type, token ))
		  {
		    found = true;
		    break;
		  }
	  
	      if ( found )
		continue;
	    }

	  if ( !types_pointer)
	    {
	      types_pointer = new mime_types;
	      current_type = types_pointer;
	      current_type->previous = NULL;
	    }
	  else
	    {
	      current_type->next = new mime_types;
	      (( mime_types *)current_type->next )->previous
		= current_type;
	      current_type = ( mime_types *)current_type->next;
	    }
      
	  current_type->type = new char[ strlen( token ) + 2 ];

	  strcpy( current_type->type, token );

	  if ( !( token = strtok( NULL, "\n" )))
	    malformed_line( copy_buffer, "etc/mime.types" );

	  strcat( token, "\n" );

	  current_type->suffixes = NULL;
	  
	  for( suffix_token = strtok( token, " \t\n" );
	       suffix_token;
	       suffix_token = strtok( NULL, " \t\n" ))
	    {
	      if ( !( current_type->suffixes ))
		{
		  current_type->suffixes = new suffix_type;
		  current_suffix = current_type->suffixes;
		  current_suffix->previous = NULL;
		}
	      else
		{
		  current_suffix->next = new suffix_type;
		  (( suffix_type *)current_suffix->next )->previous
		    = current_suffix;
		  current_suffix = ( suffix_type *)current_suffix->next;
		}

	      current_suffix->suffix = new char[ suffix_token + 2 ];
	      strcpy( current_suffix->suffix, suffix_token );
	      current_suffix->next = NULL;
	    }
	}

      fclose( m_types );
      current_type->next = NULL;
    }

  /*
   * Create linked list pairing suffixes with media players from
   * /etc/mailcap and ~/.mailcap, whith ~/.mailcap having precedence.
   */

  sprintf( line_buffer, "%s/.mailcap", home_directory );
  if (( m_cap = fopen( line_buffer, "r" )) == NULL )
    if ( errno != 2 )
      {
	perror( name );
	exit( EXIT_FAILURE );
      }
    else
      no_user_mailcap = true;

  mailcap_pointer = NULL;

  if ( !no_user_mailcap )
    {
      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, m_cap );
	  if ( feof( m_cap ))
	    break;

	  if ( line_buffer[ 0 ] == '\n' || line_buffer[ 0 ] == '#' )
	    continue;

	  copy_buffer = new char[ strlen( line_buffer ) + 2 ];
	  strcpy( copy_buffer, line_buffer );

	  if ( !( token = strtok( line_buffer, ";" )))
	    malformed_line( copy_buffer, "~/.mailcap" );

	  if ( !mailcap_pointer)
	    {
	      mailcap_pointer = new mailcap;
	      current_mailcap = mailcap_pointer;
	      current_mailcap->previous = NULL;
	    }
	  else
	    {
	      current_mailcap->next = new mailcap;
	      (( mailcap *)current_mailcap->next )->previous
		= current_mailcap;
	      current_mailcap = ( mailcap *)current_mailcap->next;
	    }
      
	  current_mailcap->type = new char[ strlen( token ) + 2 ];

	  strcpy( current_mailcap->type, token );

	  if ( !( token = strtok( NULL, "\n" )))
	    malformed_line( copy_buffer, "~/.mailcap" );

	  current_mailcap->player = new char[ strlen( token ) + 2 ];

	  strcpy( current_mailcap->player, token );
	  current_mailcap->next = NULL;
	}

      fclose( m_cap );
    }

  if (( m_cap = fopen( "/etc/mailcap", "r" )) == NULL )
    if ( errno != 2 )
      { 
	perror( name );
	exit( EXIT_FAILURE );
      }
    else
      if ( no_user_mailcap )
	{
	  fprintf( stderr, 
		   "Viewer: no /etc/mailcap or ~/.mailcap found\n" );
	  exit( EXIT_FAILURE );
	}
      else
	no_system_mailcap = true;

  if ( !no_system_mailcap )
    {
      for( ; ; )
	{
	  fgets( line_buffer, BUFFER_SIZE, m_cap );
	  if ( feof( m_cap ))
	    break;

	  if ( line_buffer[ 0 ] == '\n' || line_buffer[ 0 ] == '#' )
	    continue;

	  copy_buffer = new char[ strlen( line_buffer ) + 2 ];
	  strcpy( copy_buffer, line_buffer );

	  if ( !( token = strtok( line_buffer, ";" )))
	    malformed_line( copy_buffer, "/etc/mailcap" );

	  if ( mailcap_pointer )
	    {
	      for( other_mailcap = mailcap_pointer; 
		   other_mailcap;
		   other_mailcap = ( mailcap *)other_mailcap->next )
		if ( !strcmp( other_mailcap->type, token ))
		  {
		    found = true;
		    break;
		  }
	  
	      if ( found )
		continue;
	    }

	  if ( !mailcap_pointer)
	    {
	      mailcap_pointer = new mailcap;
	      current_mailcap = mailcap_pointer;
	      current_mailcap->previous = NULL;
	    }
	  else
	    {
	      current_mailcap->next = new mailcap;
	      (( mailcap *)current_mailcap->next )->previous
		= current_mailcap;
	      current_mailcap = ( mailcap *)current_mailcap->next;
	    }
      
	  current_mailcap->type = new char[ strlen( token ) + 2 ];

	  strcpy( current_mailcap->type, token );

	  if ( !( token = strtok( NULL, "\n" )))
	    malformed_line( copy_buffer, "etc/mailcap" );

	  current_mailcap->player = new char[ strlen( token ) + 2 ];

	  strcpy( current_mailcap->player, token );
	  current_mailcap->next = NULL;
	}

      fclose( m_cap );
    }
  
  /*
   * Fork appropriate viewer for each filename passed as an argument.
   */

  for( index = 1; index < argc; index++ )
    {
      if ( !( position = strrchr( argv[ index ], '.' )))
	{
	  fprintf( stderr, "Viewer: %s has no suffix\n", argv[ index ] );
	  continue;
	}
      
      position++;
      strcpy( line_buffer, position );

      found = false;

      for( current_type = types_pointer;
	   current_type;
	   current_type = ( mime_types *)current_type->next )
	{
	  for( current_suffix = current_type->suffixes;
	       current_suffix;
	       current_suffix = ( suffix_type *)current_suffix->next )
	    {
	      if ( !strcmp( current_suffix->suffix, line_buffer ))
		for( current_mailcap = mailcap_pointer;
		     current_mailcap;
		     current_mailcap = ( mailcap *)current_mailcap->next )
		  {
		    place = strcspn( current_mailcap->type, "/" );
		    if ( !strncmp( current_type->type, current_mailcap->type,
				   place ))
		      if ( !strcmp( current_type->type + place, 
				    current_mailcap->type + place )
			   || strstr( current_mailcap->type, "*" ))
			{
			  place = strcspn( current_mailcap->player, "%" );
			  strncpy( format, current_mailcap->player, place );
			  format[ place ] = '\0';
			  strcat( format, "'\%s'" );
			  strcat( format,
				  current_mailcap->player + place + 2 );

			  sprintf( line_buffer, format, argv[ index ] );
			  strcat( line_buffer, " &" );
			  system( line_buffer );
			  found = true;
			  break;
			} 
		  }
	      if ( found )
		break;
	    }
	  if ( found )
	    break;
	}

      if ( !found )
	fprintf( stderr, "viewer: no viewer found for %s\n", argv[ index ] );
    }
  
  /*
   * Delete the linked list of mime types.
   */

  if ( types_pointer )
    {
      current_type = types_pointer;
      while( current_type->next )
	current_type = ( mime_types *)current_type->next;

      do
	{
	  other_type = ( mime_types *)current_type->previous;

	  delete [] current_type->type;
	  
	  current_suffix = current_type->suffixes;
	  while( current_suffix->next )
	    current_suffix = ( suffix_type *)current_suffix->next;

	  do
	    {
	      other_suffix = ( suffix_type *)current_suffix->previous;

	      delete [] current_suffix->suffix;
	      delete current_suffix;

	      current_suffix = other_suffix;
	    }
	  while( current_suffix );  

	  delete current_type;

	  current_type = other_type;
	}
      while( current_type );
    }

  /*
   * Delete the linked list of mailcap entries.
   */

  if ( mailcap_pointer )
    {
      current_mailcap = mailcap_pointer;
      while( current_mailcap->next )
	current_mailcap = ( mailcap *)current_mailcap->next;

      do
	{
	  other_mailcap = ( mailcap *)current_mailcap->previous;

	  delete [] current_mailcap->type;
	  delete [] current_mailcap->player;
	  delete current_mailcap;

	  current_mailcap = other_mailcap;
	}
      while( current_mailcap );
    }
  
  return( 0 );
}
